/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   FiRES,  Germany
**/

// No Mouse support yet (baut functions are implemented)

#include <windows.h>
#include <dinput.h>

// DirectInput stuff (have to be called after DD_Init)
BOOL    DI_Init(void);
void    DI_End(void);
void    DI_GetMouseData(void);
void    DI_GetKeyboardData(void);
 

// Taken from main
extern HWND hGraphics;
extern HINSTANCE hInstanceGlobal;
extern int DebuggerStatus;

extern "C" {

	#include "h_input.h"

	extern int Paused;
// Mouse input
	int MouX, MouY, MouDeltaX, MouDeltaY, MouPress, MouRelease;

// Kbd input
	byte Keys[256], KeyPressed;

// VR stuff (placeholder)
	int VR_Yaw=0, VR_Pitch=0, VR_Roll=0;


	int Input_Init(int flags) {
		return (DI_Init());
	}

	void Input_Shut(void) {
		DI_End();
	}

	int Input_GetData(void) {
//		DI_GetMouseData();
		DI_GetKeyboardData();
		return(INPUT_KEYBOARD|INPUT_MOUSE);
	}

// Special func, called from engine msg processing
	void    DI_SyncAcquire(void);

}




// DirectInput stuff

LPDIRECTINPUT			pdi = NULL;
LPDIRECTINPUTDEVICE		pMouse = NULL;
LPDIRECTINPUTDEVICE		pKeyboard = NULL;




// Init all DirectInput-devices (mouse + keyboard). 

BOOL DI_Init(void)
{
	HRESULT hr;

// Already initialized
	if (pdi) return TRUE;

// Init DirectInput
	hr = DirectInputCreate(hInstanceGlobal,DIRECTINPUT_VERSION,&pdi,NULL);
	if (FAILED(hr)) return FALSE;

// Init mouse
	hr = pdi->CreateDevice(GUID_SysMouse,&pMouse,NULL);
	if (FAILED(hr)) return FALSE;

	hr = pMouse->SetDataFormat(&c_dfDIMouse);
	if (FAILED(hr)) return FALSE;

	hr = pMouse->SetCooperativeLevel(hGraphics,DISCL_EXCLUSIVE|DISCL_FOREGROUND);
	if (FAILED(hr)) return FALSE;

// Init keyboard
	hr = pdi->CreateDevice(GUID_SysKeyboard,&pKeyboard,NULL);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetDataFormat(&c_dfDIKeyboard);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetCooperativeLevel(hGraphics,DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	if (FAILED(hr)) return FALSE;

// acquire mouse and kbd
	DI_SyncAcquire();

	return TRUE;
}


void DI_End(void)
{

// Shut keyboard
	if (pKeyboard) {
		pKeyboard->Unacquire();
		pKeyboard->Release();
		pKeyboard = NULL;
	}

// Shut mouse
	if (pMouse) {
		pMouse->Unacquire();
		pMouse->Release();
		pMouse = NULL;
	}

// Shut DirectInput
	if (pdi) {
		pdi->Release();
		pdi = NULL;
	}
}


void DI_GetMouseData(void)
{
	DIMOUSESTATE dims;
	HRESULT hr;
	int done = 1;

	if (!pMouse) return;

	while(done)
	{
		done = 0;
		hr = pMouse->GetDeviceState(sizeof(DIMOUSESTATE), &dims);
		if (hr == DIERR_INPUTLOST) {
			hr = pMouse->Acquire();
				if (SUCCEEDED(hr)) done = 1;
		}
	}


	if (SUCCEEDED(hr)) {
		MouDeltaX = dims.lX;
		MouDeltaY = dims.lY;

		MouRelease=0;
		if (MouPress) MouRelease=MouPress;
		MouPress=(dims.rgbButtons[0]&0x80) ? 1:0;
		MouPress|=(dims.rgbButtons[1]&0x80) ? 2:0;
		if (MouPress) MouRelease=0;
	}
}


void DI_GetKeyboardData(void)
{
	HRESULT hr;
	int done = 1;

	if (!pKeyboard) return;

	while (done)
	{
		done = 0;
		hr = pKeyboard->GetDeviceState(sizeof(Keys), Keys);
		if (hr == DIERR_INPUTLOST) {
			hr = pKeyboard->Acquire();
			if (SUCCEEDED(hr)) done = 1;
		}
	}


	if (SUCCEEDED(hr)) {
	}
}



extern "C" {

void DI_SyncAcquire(void)
{
        if (!pMouse||!pKeyboard) return;
        if (Paused) {
                pMouse->Unacquire();
                pKeyboard->Unacquire();
        }
        else {
                pMouse->Acquire();
                pKeyboard->Acquire();
        }
}

}

extern "C" {

        static WORD		   keystate = 0;
        static signed char stick_x = 0;
        static signed char stick_y = 0;


int tr = 0;

void PausedKey(void)
{
//** Pause Button (mhh)
    if (Keys[kP])    
	{
		Paused = 1;
		DebuggerStatus = 1;
		Keys[kP] = 0;
	}
}

DWORD GetDInputData()
{
	int StickFlag = 0;
	Input_GetData();

//** KeyPress and KeyReleases 


//** Function-Buttons
  //** Start Button
    if (Keys[kQ])    
		keystate |= 0x1000;
	else
		keystate &= ~0x1000;

  //** A-Button
    if (Keys[kW])    
        keystate |= 0x8000;
	else
		keystate &= ~0x8000;

  //** B-Button
    if (Keys[kE])    
        keystate |= 0x4000;
	else
		keystate &= ~0x4000;

//** C-Buttons...
  //** C-Button UP
    if (Keys[kHOME])    
        keystate |= 0x0008;
	else
		keystate &= ~0x0008;

  //** C-Button DOWN
    if (Keys[kEND])    
        keystate |= 0x0004;
	else
		keystate &= ~0x0004;

  //** C-Button LEFT
    if (Keys[kDEL])    
        keystate |= 0x0002;
	else
		keystate &= ~0x0002;

  //** C-Button RIGHT
    if (Keys[kPGDN])    
        keystate |= 0x0001;
	else
		keystate &= ~0x0001;

//** D-Buttons...
  //** D-Button UP
    if (Keys[kUP])    
        keystate |= 0x0800;
	else
		keystate &= ~0x0800;

  //** D-Button DOWN
    if (Keys[kDOWN])    
        keystate |= 0x0400;
	else
		keystate &= ~0x0400;

  //** D-Button LEFT
    if (Keys[kLEFT])    
        keystate |= 0x0200;
	else
		keystate &= ~0x0200;

  //** D-Button RIGHT
    if (Keys[kRIGHT])    
        keystate |= 0x0100;
	else
		keystate &= ~0x0100;

//** d
  //** L-Button 
    if (Keys[kA])    
        keystate |= 0x0020;
	else
		keystate &= ~0x0020;

  //** R-Button
    if (Keys[kS])    
        keystate |= 0x0010;
	else
		keystate &= ~0x0010;

  //** Z-Button
    if (Keys[kD])    
        keystate |= 0x2000;
	else
		keystate &= ~0x2000;

//************************************

  //** Stick UP
	if (Keys[kK])
	{
		StickFlag = 1;
		if(stick_y < 120) stick_y += 0x08;
	}

  //** Stick DOWN
    if (Keys[kJ])
	{
		StickFlag = 1;
        if(stick_y > -120) stick_y -= 0x08;
	}

  //** Stick RIGHT  
	if (Keys[kL])
	{
		StickFlag = 1;
        if(stick_x < 120) stick_x += 0x08;
	}

  //** Stick LEFT
    if (Keys[kH])
	{
		StickFlag = 1;
        if(stick_x > -120) stick_x -= 0x08;
	}
			

//** Release Stick
	if (StickFlag == 0)
	{
		stick_x = 0;
		stick_y = 0;
	}

    return(((DWORD)(keystate << 16)) | ((DWORD)(WORD)(stick_x << 8)) | ((DWORD)(byte)stick_y));

}

} // end: extern "C"