/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   FiRES,  Germany
**/

#include <windows.h>
#include <stdio.h>
#include <shlwapi.h>

#include "resource.h"
#include "debug.h"
#include "debug_hnd.h"

HWND hDebugger;

extern "C"
{
	#include "n64/registers.h"
	#include "n64/rsp_registers.h"
	#include "file_loader_extern.h"
	#include "n64/interrupts.h"
	extern int cpu_step(void);
	extern __int64 counter;
}
// ******** Globals ******
/////////////////////////////////////////////////////////////////////////////////
extern HINSTANCE hInstanceGlobal;
PROPSHEETPAGE	Main_psp[1], Reg_psp[3];
PROPSHEETHEADER Main_psh,    Reg_psh;

//** Debugger Commands
bool  LogCPU = false;
uWORD LogCPUPos = 0;

int DebuggerStatus = 0;
int vec_num = 0;

bool	DStep = false;
int		DebuggerSteps = 0;

bool		  DStopOnAddress = false;
unsigned int  DStopAddress = 0;
//**
long PosListBox = 0xa4000040;

bool DRun = false;
/////////////////////////////////////////////////////////////////////////////////
// ******** All Functions *********
//** Regs
LRESULT CALLBACK RSP_GRegs(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK CP0Regs(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK RSP_VRegs(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
void UpdateRegs(void);
int CreateRegSheet(void);

//** Main
LRESULT CALLBACK Main_Deb(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
int CreateMainSheet(void);
void Update_All(void);
void Next_Cpu_Step(void);
int StartDebugger(void);
void UpdateListBox(void);
void LogCPUData(void);
// ******** End All Functions *********


#include "debug_cpu.h"
#include "debug_regs.h"



/////////////////////////////////////////////////////////////////////////////////
void Next_Cpu_Step(void)
{
	if (LogCPU) LogCPUData();
	cpu_step();
}

__int32 countit = 0;
void LogCPUData(void)
{
		if (LogCPUPos != reg.pc)
		{
			FILE *stream;
			countit++;
			stream = fopen("CPU.log", "at");

			which_code(reg.pc);
			fprintf(stream,"_%08x_ %x %s \n", countit, reg.pc, output);

			fclose(stream);
			LogCPUPos = reg.pc;
		}
}

void UpdateListBox(void)
{
	int ret;
	char test1[1024];
	char umb[3];
	
	test1[0] = 0;
	umb[0] = 0x0d;
	umb[1] = 0x0a;
	umb[2] = 00;

	for (int i = 0; i<12; i++)
	{
		sprintf(output, "%x: ",PosListBox + i*4);
		strcat(test1, output);
		which_code(PosListBox + i * 4);
		strcat(test1, output);
		strcat(test1, umb);
	}

	ret = SendMessage( MAIN_LIST_OPSHandle, 
					   WM_SETTEXT, 0, (LPARAM)(LPCTSTR)test1);
 
}

void Update_All(void)
{
	if (reg.pc > 0)
		PosListBox = reg.pc; // to see the middle
	UpdateRegs();
	UpdateListBox();
} 


LRESULT CALLBACK Main_Deb(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	char text[1024];
	switch(message)
	{
	case WM_INITDIALOG:

		MAIN_STOPA_EHandle=GetDlgItem(hDlg,  MAIN_STOPA_E);
		MAIN_NUMSTEPSHandle=GetDlgItem(hDlg, IDC_NUMSTEPS);
	 	MAIN_LIST_OPSHandle=GetDlgItem(hDlg, MAIN_LIST_OPS);
		MAIN_SKIPHandle=GetDlgItem(hDlg, MAIN_SKIP);
		MAIN_LOGCPUHandle=GetDlgItem(hDlg, MAIN_LOGCPU);
		Update_All();	
		return (TRUE);
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case (IDOK || IDCANCEL):
			EndDialog(hDlg,TRUE);
			DebuggerStatus = Debugger_OFF;
			return(TRUE);

		case IDC_DEBUGENTER:
			{
					int pepep = 0;
			}
			break;
		case IDC_CPU_STEPS:
			text[0] = 0;
			SendMessage(MAIN_NUMSTEPSHandle, WM_GETTEXT, 1024, (LPARAM)(LPCTSTR)text);
			DStep = true;
			DebuggerSteps = StrToInt(text);
			return (TRUE);
		case MAIN_STOPA_B:
			int address;
			text[0] = 0;
			SendMessage(MAIN_STOPA_EHandle, WM_GETTEXT, 1024, (LPARAM)(LPCTSTR)text);
			DStopOnAddress = true;
			StrToIntEx(text, STIF_SUPPORT_HEX, &address);
			DStopAddress = address;
			return (TRUE);
		case MAIN_SKIP:
			reg.pc += 4;
			return(TRUE);
		case MAIN_LOGCPU:
			if (LogCPU == false) LogCPU = true;
			else LogCPU = false;
			return(true);
		}
		break;
	}
	return(FALSE);
}

int CreateMainSheet()
{
	Main_psp[0].dwSize = sizeof(PROPSHEETPAGE);
	Main_psp[0].dwFlags = PSP_USETITLE;
	Main_psp[0].hInstance = hInstanceGlobal;
	Main_psp[0].pszTemplate = "DEB_MAIN";
	Main_psp[0].pszIcon = NULL;
	Main_psp[0].pfnDlgProc = (DLGPROC)Main_Deb;
	Main_psp[0].pszTitle = "Handling of CPU/RDP";
	Main_psp[0].lParam = 0;

	Main_psh.dwSize = sizeof(PROPSHEETHEADER);
	Main_psh.dwFlags = PSH_PROPSHEETPAGE | PSH_MODELESS;
	Main_psh.hwndParent = hDebugger;
	Main_psh.hInstance = hInstanceGlobal;
	Main_psh.pszIcon = NULL;
	Main_psh.pszCaption = (LPSTR) "Debugger";
	Main_psh.nPages = sizeof(Main_psp) / sizeof(PROPSHEETPAGE);
	Main_psh.ppsp = (LPCPROPSHEETPAGE) &Main_psp;
	return(PropertySheet(&Main_psh));
}

int StartDebugger()
{
	CreateRegSheet();
	CreateMainSheet();
	return 0;
}


void DebuggerAction()
{
//** stop the debugger if a specifeced address is reached
	if(DStopOnAddress)
	{
		long counter = 0;
		while (reg.pc != DStopAddress) 
		{
			Next_Cpu_Step();
			counter++;
			if (counter > 50000000) {
				MessageBox(hDebugger, "Over 50 millioen Operations\nare executed - we break\n (perhaps wrong Address)\n","ERROR" , MB_OK);
				DStopAddress = reg.pc;
			}
		}
		DStopOnAddress = false;
		DStopAddress = 0;
		Update_All();
	}
//** makes a specifecied of cpu steps (the loops are for redraw())
	if(DStep)
	{
		if (DebuggerSteps >= 100000)
		{
			for (int i; i < 100000; i++)
				Next_Cpu_Step();
			DebuggerSteps = DebuggerSteps - 100000;
			Update_All();
		}
		else
		{
			for (int i = 0; i < DebuggerSteps; i++)
				Next_Cpu_Step();
			DebuggerSteps = 0;
			DStep = false;
			Update_All();
		}
	}

/*	if(DRun)
	{
		for (int i=0; i <10000; i++)
			Next_Cpu_Step();
		Update_All();
	}*/

//** Laest den Emulator solange laufen, bis Bedingung erfuellt
	if(DRun)
	{
		int j,equal = 0;
		for (j = 0; j < 1000000; j++)
		{
			for (int i=0; i <32; i++)
			{
				if (rsp_reg.r[i] == 0xbf00) DRun = false;
			}
			Next_Cpu_Step();
		}
		Update_All();
	}


//** Protokolliert alle CPU-Stpes in der Datei "CPU.log"
}
