/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   FiRES,  Germany
**/

#include <stdio.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include "n64/registers.h"

extern rs4300i_reg reg;

typedef void (*cpu_instr)();

char text_row[1024];
char output[1024];

uWORD code, Address;

// ************
#define READ_CPU_INSTRUCTION(addr)                                                                  \
(                                                                                              \
        ( (addr < 0xa4000040) )                                                                \
        ?                                                                                      \
                ( *((uWORD *)(mem.rd_ram + (addr & 0x1fffffff))) )     /* instr in prgcode */   \
        :                                                                                      \
                ( (addr < 0xe0000000) )                                                        \
                ?                                                                              \
                        ( *(uWORD *)(mem.sp_dmem + (addr & 0xfff)) )   /* instr in bootcode */  \
                :                                                                              \
                        ( *(uWORD *)(rom.image + (addr & 0xfff)) )              /* for test */  \
)


#define __CODE          (code = READ_CPU_INSTRUCTION(Address))
#define __NEXT_CODE     (READ_CPU_INSTRUCTION((reg.pc + 4)))
#define __PREV_CODE     (READ_CPU_INSTRUCTION((reg.pc - 4)))

#define __OPCODE        ((uBYTE)(__CODE >> 26))

#define __RS            (((uBYTE)(code >> 21)) & 0x1f)

#define __RT            (((uBYTE)(code >> 16)) & 0x1f)

#define __RD            (((uBYTE)(code >> 11)) & 0x1f)

#define __SA            (((uBYTE)(code >>  6)) & 0x1f)

#define __F             ( (uBYTE)(code)       & 0x3f)

#define __I             ( (sWORD)(sHWORD)code )
#define __O             ( reg.pc + 4 + (__I << 2) )

#define ____T           (code & 0x3ffffff)
#define __T             ( (reg.pc & 0xf0000000) | (____T << 2) )


// ** FPU
#define FMT_S   16      /* 32bit binary floating-point */
#define FMT_D   17      /* 64bit binary floating-point */
#define FMT_W   20      /* 32bit binary fixed-point    */
#define FMT_L   21      /* 64bit binary fixed-point    */

#define __FMT   __RS

#define __FT    __RT

#define __FS    __RD

#define __FD    __SA



		#define PRINT_NOT_IMPLEMENTED

		#define PRINT_XINST
        #define PRINT_NOTHING

        #define PRINT_O         sprintf(text_row," $%08lx", __O);
        #define PRINT_RD        sprintf(text_row," r%02u", __RD);
        #define PRINT_RS        sprintf(text_row," r%02u", __RS);
        #define PRINT_T         sprintf(text_row," $%08lx", __T); 

        #define PRINT_RD_RS     sprintf(text_row," r%02u, r%02u", __RD, __RS);
        #define PRINT_RS_I      sprintf(text_row," r%02u, $%08lx (%ld)", __RS, (sWORD)__I, (sWORD)__I);
        #define PRINT_RS_O      sprintf(text_row," r%02u, $%08lx", __RS, __O);
        #define PRINT_RS_RD     sprintf(text_row," r%02u, r%02u", __RS, __RD);
        #define PRINT_RS_RT     sprintf(text_row," r%02u, r%02u", __RS, __RT);
        #define PRINT_RT_I      sprintf(text_row," r%02u, $%04hx ($%08lx)", __RT, (uHWORD)__I, (sWORD)(__I << 16));
        #define PRINT_RT_RD     sprintf(text_row," r%02u, r%02u", __RT, __RD);

        #define PRINT_RD_RS_RT  sprintf(text_row," r%02u, r%02u, r%02u", __RD, __RS, __RT);
        #define PRINT_RD_RT_SA  sprintf(text_row," r%02u, r%02u, %02u", __RD, __RT, __SA);
        #define PRINT_RD_RT_RS  sprintf(text_row," r%02u, r%02u, r%02u", __RD, __RT, __RS);
        #define PRINT_RS_RT_O   sprintf(text_row," r%02u, r%02u, $%08lx", __RS, __RT, __O);
        #define PRINT_RT_O_B    sprintf(text_row," r%02u, $%04lx(r%02u)", __RT, __I, __RS);
        #define PRINT_RT_RS_I   sprintf(text_row," r%02u, r%02u, $%04hx (%hd)", __RT, __RS, (sHWORD)__I, (sHWORD)__I);

        #define PRINT_FD_FS_FT  sprintf(text_row," f%02u, f%02u, f%02u", __FD, __FS, __FT);
        #define PRINT_FD_FS     sprintf(text_row," f%02u, f%02u", __FD, __FS);
        #define PRINT_FT_O_B    sprintf(text_row," f%02u, $%04lx(r%02u)", __FT, __I, __RS);
        #define PRINT_RT_FS     sprintf(text_row," r%02u, f%02u", __RT, __FS);

        #define PRINT_C0        sprintf(text_row," r%02u, cpr0[%02u]", __RT, __RD);
        #define PRINT_C0C       sprintf(text_row," r%02u, ccr0[%02u]", __RT, __RD);
        #define PRINT_RT_C1     sprintf(text_row," r%02u, ccr1[%02u]", __RT, __FS);
        #define PRINT_C2_O_B    sprintf(text_row," cpr2[%02u], $%04lx(r%02u)", __RT, __I, __RS);


        #define PRINT_DEBUGGING_INFO(name, format)      \
                        sprintf(output,"%-15s", name);  \
                        PRINT_##format                  \
						strcat(output, text_row);       \


#define PRINT_FMT               \
        switch(__FMT)           \
        {                       \
            case FMT_S:         \
                putchar('s');   \
                break;          \
            case FMT_D:         \
                putchar('d');   \
                break;          \
            case FMT_W:         \
                putchar('w');   \
                break;          \
            case FMT_L:         \
                putchar('l');   \
                break;          \
            default:            \
                putchar('?');   \
        }

        #define PRINT_DEBUGGING_INFO_FPU(name, format)  \
                        sprintf(output,"%-15s", name);  \
                        PRINT_##format                  \
						strcat(output, text_row);		\

// *************************************************************************





static void rs4300i_reserved();
static void rs4300i_invalid();
static void rs4300i_special();
static void rs4300i_regimm();
static void rs4300i_j();
static void rs4300i_jal();
static void rs4300i_beq();
static void rs4300i_bne();
static void rs4300i_blez();
static void rs4300i_bgtz();
static void rs4300i_addi();
static void rs4300i_addiu();
static void rs4300i_slti();
static void rs4300i_sltiu();
static void rs4300i_andi();
static void rs4300i_ori();
static void rs4300i_xori();
static void rs4300i_lui();
static void rs4300i_cop0();
static void rs4300i_cop1();
static void rs4300i_cop2();
static void rs4300i_beql();
static void rs4300i_bnel();
static void rs4300i_blezl();
static void rs4300i_bgtzl();
static void rs4300i_daddi();
static void rs4300i_daddiu();
static void rs4300i_ldl();
static void rs4300i_ldr();
static void rs4300i_lb();
static void rs4300i_lh();
static void rs4300i_lwl();
static void rs4300i_lw();
static void rs4300i_lbu();
static void rs4300i_lhu();
static void rs4300i_lwr();
static void rs4300i_lwu();
static void rs4300i_sb();
static void rs4300i_sh();
static void rs4300i_swl();
static void rs4300i_sw();
static void rs4300i_sdl();
static void rs4300i_sdr();
static void rs4300i_swr();
static void rs4300i_cache();
static void rs4300i_ll();
static void rs4300i_lwc1();
static void rs4300i_lwc2();
static void rs4300i_lld();
static void rs4300i_ldc1();
static void rs4300i_ldc2();
static void rs4300i_ld();
static void rs4300i_sc();
static void rs4300i_swc1();
static void rs4300i_swc2();
static void rs4300i_scd();
static void rs4300i_sdc1();
static void rs4300i_sdc2();
static void rs4300i_sd();
static void rs4300i_special_sll();
static void rs4300i_special_srl();
static void rs4300i_special_sra();
static void rs4300i_special_sllv();
static void rs4300i_special_srlv();
static void rs4300i_special_srav();
static void rs4300i_special_jr();
static void rs4300i_special_jalr();
static void rs4300i_special_syscall();
static void rs4300i_special_break();
static void rs4300i_special_sync();
static void rs4300i_special_mfhi();
static void rs4300i_special_mthi();
static void rs4300i_special_mflo();
static void rs4300i_special_mtlo();
static void rs4300i_special_dsllv();
static void rs4300i_special_dsrlv();
static void rs4300i_special_dsrav();
static void rs4300i_special_mult();
static void rs4300i_special_multu();
static void rs4300i_special_div();
static void rs4300i_special_divu();
static void rs4300i_special_dmult();
static void rs4300i_special_dmultu();
static void rs4300i_special_ddiv();
static void rs4300i_special_ddivu();
static void rs4300i_special_add();
static void rs4300i_special_addu();
static void rs4300i_special_sub();
static void rs4300i_special_subu();
static void rs4300i_special_and();
static void rs4300i_special_or();
static void rs4300i_special_xor();
static void rs4300i_special_nor();
static void rs4300i_special_slt();
static void rs4300i_special_sltu();
static void rs4300i_special_dadd();
static void rs4300i_special_daddu();
static void rs4300i_special_dsub();
static void rs4300i_special_dsubu();
static void rs4300i_special_tge();
static void rs4300i_special_tgeu();
static void rs4300i_special_tlt();
static void rs4300i_special_tltu();
static void rs4300i_special_teq();
static void rs4300i_special_tne();
static void rs4300i_special_dsll();
static void rs4300i_special_dsrl();
static void rs4300i_special_dsra();
static void rs4300i_special_dsll32();
static void rs4300i_special_dsrl32();
static void rs4300i_special_dsra32();
static void rs4300i_regimm_bltz();
static void rs4300i_regimm_bgez();
static void rs4300i_regimm_bltzl();
static void rs4300i_regimm_bgezl();
static void rs4300i_regimm_tgei();
static void rs4300i_regimm_tgeiu();
static void rs4300i_regimm_tlti();
static void rs4300i_regimm_tltiu();
static void rs4300i_regimm_teqi();
static void rs4300i_regimm_tnei();
static void rs4300i_regimm_bltzal();
static void rs4300i_regimm_bgezal();
static void rs4300i_regimm_bltzall();
static void rs4300i_regimm_bgezall();
static void rs4300i_cop0_rs_mf();
static void rs4300i_cop0_rs_dmf();
static void rs4300i_cop0_rs_cf();
static void rs4300i_cop0_rs_mt();
static void rs4300i_cop0_rs_dmt();
static void rs4300i_cop0_rs_ct();
static void rs4300i_cop0_rs_bc();
static void rs4300i_cop0_rs_co();
static void rs4300i_cop0_rt_bcf();
static void rs4300i_cop0_rt_bct();
static void rs4300i_cop0_rt_bcfl();
static void rs4300i_cop0_rt_bctl();
static void rs4300i_cop0_tlbr();
static void rs4300i_cop0_tlbwi();
static void rs4300i_cop0_tlbwr();
static void rs4300i_cop0_tlbp();
static void rs4300i_cop0_eret();
static void rs4300i_cop1_rs_mf();
static void rs4300i_cop1_rs_dmf();
static void rs4300i_cop1_rs_cf();
static void rs4300i_cop1_rs_mt();
static void rs4300i_cop1_rs_dmt();
static void rs4300i_cop1_rs_ct();
static void rs4300i_cop1_rs_bc();
static void rs4300i_cop1_rs_co();
static void rs4300i_cop1_rt_bcf();
static void rs4300i_cop1_rt_bct();
static void rs4300i_cop1_rt_bcfl();
static void rs4300i_cop1_rt_bctl();
static void rs4300i_cop1_add();
static void rs4300i_cop1_sub();
static void rs4300i_cop1_mul();
static void rs4300i_cop1_div();
static void rs4300i_cop1_sqrt();
static void rs4300i_cop1_abs();
static void rs4300i_cop1_mov();
static void rs4300i_cop1_neg();
static void rs4300i_cop1_roundl();
static void rs4300i_cop1_truncl();
static void rs4300i_cop1_ceill();
static void rs4300i_cop1_floorl();
static void rs4300i_cop1_roundw();
static void rs4300i_cop1_truncw();
static void rs4300i_cop1_ceilw();
static void rs4300i_cop1_floorw();
static void rs4300i_cop1_cvts();
static void rs4300i_cop1_cvtd();
static void rs4300i_cop1_cvtw();
static void rs4300i_cop1_cvtl();
static void rs4300i_cop1_c();
static void rs4300i_cop2_rs_not_implemented();

                   
static cpu_instr instruction[64] =
{
        rs4300i_special,
        rs4300i_regimm,
        rs4300i_j,
        rs4300i_jal,
        rs4300i_beq,
        rs4300i_bne,
        rs4300i_blez,
        rs4300i_bgtz,
        rs4300i_addi,
        rs4300i_addiu,
        rs4300i_slti,
        rs4300i_sltiu,
        rs4300i_andi,
        rs4300i_ori,
        rs4300i_xori,
        rs4300i_lui,
        rs4300i_cop0,
        rs4300i_cop1,
        rs4300i_cop2,
        rs4300i_reserved,
        rs4300i_beql,
        rs4300i_bnel,
        rs4300i_blezl,
        rs4300i_bgtzl,
        rs4300i_daddi,
        rs4300i_daddiu,
        rs4300i_ldl,
        rs4300i_ldr,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_lb,
        rs4300i_lh,
        rs4300i_lwl,
        rs4300i_lw,
        rs4300i_lbu,
        rs4300i_lhu,
        rs4300i_lwr,
        rs4300i_lwu,
        rs4300i_sb,
        rs4300i_sh,
        rs4300i_swl,
        rs4300i_sw,
        rs4300i_sdl,
        rs4300i_sdr,
        rs4300i_swr,
        rs4300i_cache,
        rs4300i_ll,
        rs4300i_lwc1,
        rs4300i_lwc2,
        rs4300i_reserved,
        rs4300i_lld,
        rs4300i_ldc1,
        rs4300i_ldc2,
        rs4300i_ld,
        rs4300i_sc,
        rs4300i_swc1,
        rs4300i_swc2,
        rs4300i_reserved,
        rs4300i_scd,
        rs4300i_sdc1,
        rs4300i_sdc2,
        rs4300i_sd
};





static cpu_instr special_instruction[64] =
{
        rs4300i_special_sll,
        rs4300i_reserved,
        rs4300i_special_srl,
        rs4300i_special_sra,
        rs4300i_special_sllv,
        rs4300i_reserved,
        rs4300i_special_srlv,
        rs4300i_special_srav,
        rs4300i_special_jr,
        rs4300i_special_jalr,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_special_syscall,
        rs4300i_special_break,
        rs4300i_reserved,
        rs4300i_special_sync,
        rs4300i_special_mfhi,
        rs4300i_special_mthi,
        rs4300i_special_mflo,
        rs4300i_special_mtlo,
        rs4300i_special_dsllv,
        rs4300i_reserved,
        rs4300i_special_dsrlv,
        rs4300i_special_dsrav,
        rs4300i_special_mult,
        rs4300i_special_multu,
        rs4300i_special_div,
        rs4300i_special_divu,
        rs4300i_special_dmult,
        rs4300i_special_dmultu,
        rs4300i_special_ddiv,
        rs4300i_special_ddivu,
        rs4300i_special_add,
        rs4300i_special_addu,
        rs4300i_special_sub,
        rs4300i_special_subu,
        rs4300i_special_and,
        rs4300i_special_or,
        rs4300i_special_xor,
        rs4300i_special_nor,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_special_slt,
        rs4300i_special_sltu,
        rs4300i_special_dadd,
        rs4300i_special_daddu,
        rs4300i_special_dsub,
        rs4300i_special_dsubu,
        rs4300i_special_tge,
        rs4300i_special_tgeu,
        rs4300i_special_tlt,
        rs4300i_special_tltu,
        rs4300i_special_teq,
        rs4300i_reserved,
        rs4300i_special_tne,
        rs4300i_reserved,
        rs4300i_special_dsll,
        rs4300i_reserved,
        rs4300i_special_dsrl,
        rs4300i_special_dsra,
        rs4300i_special_dsll32,
        rs4300i_reserved,
        rs4300i_special_dsrl32,
        rs4300i_special_dsra32
};





static cpu_instr regimm_instruction[32] =
{
        rs4300i_regimm_bltz,
        rs4300i_regimm_bgez,
        rs4300i_regimm_bltzl,
        rs4300i_regimm_bgezl,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_regimm_tgei,
        rs4300i_regimm_tgeiu,
        rs4300i_regimm_tlti,
        rs4300i_regimm_tltiu,
        rs4300i_regimm_teqi,
        rs4300i_reserved,
        rs4300i_regimm_tnei,
        rs4300i_reserved,
        rs4300i_regimm_bltzal,
        rs4300i_regimm_bgezal,
        rs4300i_regimm_bltzall,
        rs4300i_regimm_bgezall,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved
};





static cpu_instr cop0_rs_instruction[32] =
{
        rs4300i_cop0_rs_mf,
        rs4300i_cop0_rs_dmf,
        rs4300i_cop0_rs_cf,
        rs4300i_reserved,
        rs4300i_cop0_rs_mt,
        rs4300i_cop0_rs_dmt,
        rs4300i_cop0_rs_ct,
        rs4300i_reserved,
        rs4300i_cop0_rs_bc,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co,
        rs4300i_cop0_rs_co
};





static cpu_instr cop0_rt_instruction[32] =
{
        rs4300i_cop0_rt_bcf,
        rs4300i_cop0_rt_bct,
        rs4300i_cop0_rt_bcfl,
        rs4300i_cop0_rt_bctl,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved
};





static cpu_instr cop0_instruction[64] =
{
        rs4300i_invalid,
        rs4300i_cop0_tlbr,
        rs4300i_cop0_tlbwi,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_cop0_tlbwr,
        rs4300i_invalid,
        rs4300i_cop0_tlbp,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_reserved,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_cop0_eret,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid
};





static cpu_instr cop1_rs_instruction[32] =
{
        rs4300i_cop1_rs_mf,
        rs4300i_cop1_rs_dmf,
        rs4300i_cop1_rs_cf,
        rs4300i_reserved,
        rs4300i_cop1_rs_mt,
        rs4300i_cop1_rs_dmt,
        rs4300i_cop1_rs_ct,
        rs4300i_reserved,
        rs4300i_cop1_rs_bc,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co,
        rs4300i_cop1_rs_co
};





static cpu_instr cop1_rt_instruction[32] =
{
        rs4300i_cop1_rt_bcf,
        rs4300i_cop1_rt_bct,
        rs4300i_cop1_rt_bcfl,
        rs4300i_cop1_rt_bctl,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved,
        rs4300i_reserved
};





static cpu_instr cop1_instruction[64] =
{
        rs4300i_cop1_add,
        rs4300i_cop1_sub,
        rs4300i_cop1_mul,
        rs4300i_cop1_div,
        rs4300i_cop1_sqrt,
        rs4300i_cop1_abs,
        rs4300i_cop1_mov,
        rs4300i_cop1_neg,
        rs4300i_cop1_roundl,
        rs4300i_cop1_truncl,
        rs4300i_cop1_ceill,
        rs4300i_cop1_floorl,
        rs4300i_cop1_roundw,
        rs4300i_cop1_truncw,
        rs4300i_cop1_ceilw,
        rs4300i_cop1_floorw,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_cop1_cvts,
        rs4300i_cop1_cvtd,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_cop1_cvtw,
        rs4300i_cop1_cvtl,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_invalid,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c,
        rs4300i_cop1_c
};





static cpu_instr cop2_rs_instruction[32] =
{
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented,
        rs4300i_cop2_rs_not_implemented
};





void which_code(DWORD address)
{
    Address = address;
    instruction[__OPCODE]();
} /* void cpu_step() */

  /******************************************************************************\
*                                                                              *
*   Central Processing Unit (CPU)                                              *
*                                                                              *
\******************************************************************************/
static void rs4300i_reserved()
{
}
static void rs4300i_invalid()
{
}




static void rs4300i_special()
{
        PRINT_XINST("special instruction");
        special_instruction[__F]();
}





static void rs4300i_regimm()
{
        PRINT_XINST("regimm instructioni");
        regimm_instruction[__RT]();
}





static void rs4300i_j()
{
        PRINT_DEBUGGING_INFO("J", T);
}





static void rs4300i_jal()
{
        PRINT_DEBUGGING_INFO("JAL", T);
}





static void rs4300i_beq()
{
        PRINT_DEBUGGING_INFO("BEQ", RS_RT_O);
}





static void rs4300i_bne()
{
        PRINT_DEBUGGING_INFO("BNE", RS_RT_O);
}





static void rs4300i_blez()
{
        PRINT_DEBUGGING_INFO("BLEZ", RS_O);
}





static void rs4300i_bgtz()
{
        PRINT_DEBUGGING_INFO("BGTZ", RS_O);
}





static void rs4300i_addi()
{
        PRINT_DEBUGGING_INFO("ADDI", RT_RS_I);
}





static void rs4300i_addiu()
{
        PRINT_DEBUGGING_INFO("ADDIU", RT_RS_I);
}





static void rs4300i_slti()
{
        PRINT_DEBUGGING_INFO("SLTI", RT_RS_I);
}





static void rs4300i_sltiu()
{
        PRINT_DEBUGGING_INFO("SLTIU", RT_RS_I);
}





static void rs4300i_andi()
{
        PRINT_DEBUGGING_INFO("ANDI", RT_RS_I);
}





static void rs4300i_ori()
{
        PRINT_DEBUGGING_INFO("ORI", RT_RS_I);
}





static void rs4300i_xori()
{
        PRINT_DEBUGGING_INFO("XORI", RT_RS_I);
}





static void rs4300i_lui()
{
        PRINT_DEBUGGING_INFO("LUI", RT_I);
}





static void rs4300i_cop0()
{
        PRINT_XINST("cop0 instruction");
		cop0_rs_instruction[__RS]();
}





static void rs4300i_cop1()
{
        PRINT_XINST("cop1 instruction");
		cop1_rs_instruction[__RS]();
}





static void rs4300i_cop2()
{
        PRINT_XINST("cop2 instruction");
		cop2_rs_instruction[__RS]();
}





static void rs4300i_beql()
{
        PRINT_DEBUGGING_INFO("BEQL", RS_RT_O);

}





static void rs4300i_bnel()
{
        PRINT_DEBUGGING_INFO("BNEL", RS_RT_O);
}





static void rs4300i_blezl()
{
        PRINT_DEBUGGING_INFO("BLEZL", RS_O);
}





static void rs4300i_bgtzl()
{
        PRINT_DEBUGGING_INFO("BGTZL", RS_O);
}





static void rs4300i_daddi()
{
        PRINT_DEBUGGING_INFO("DADDI", RT_RS_I);
}





static void rs4300i_daddiu()
{
        PRINT_DEBUGGING_INFO("DADDIU", RT_RS_I);
}





static void rs4300i_ldl()
{
        PRINT_DEBUGGING_INFO("LDL", RT_O_B);
}





static void rs4300i_ldr()
{

        PRINT_DEBUGGING_INFO("LDR", RT_O_B);
}





static void rs4300i_lb()
{
        PRINT_DEBUGGING_INFO("LB", RT_O_B);
}



static void rs4300i_lh()
{
        PRINT_DEBUGGING_INFO("LH", RT_O_B);
}



static void rs4300i_lwl()
{
        PRINT_DEBUGGING_INFO("LWL", RT_O_B);
}





static void rs4300i_lw()
{
        PRINT_DEBUGGING_INFO("LW", RT_O_B);
}





static void rs4300i_lbu()
{
        PRINT_DEBUGGING_INFO("LBU", RT_O_B);
}





static void rs4300i_lhu()
{
        PRINT_DEBUGGING_INFO("LHU", RT_O_B);
}





static void rs4300i_lwr()
{
        PRINT_DEBUGGING_INFO("LWR", RT_O_B);
}





static void rs4300i_lwu()
{
        PRINT_DEBUGGING_INFO("LWU", RT_O_B);
}





static void rs4300i_sb()
{
        PRINT_DEBUGGING_INFO("SB", RT_O_B);
}





static void rs4300i_sh()
{
        PRINT_DEBUGGING_INFO("SH", RT_O_B);
}





static void rs4300i_swl()
{
        PRINT_DEBUGGING_INFO("SWL", RT_O_B);
}





static void rs4300i_sw()
{
        PRINT_DEBUGGING_INFO("SW", RT_O_B);
}





static void rs4300i_sdl()
{
        PRINT_DEBUGGING_INFO("SDL", RT_O_B);
}





static void rs4300i_sdr()
{
        PRINT_DEBUGGING_INFO("SDR", RT_O_B);
}





static void rs4300i_swr()
{
        PRINT_DEBUGGING_INFO("SWR", RT_O_B);
}





static void rs4300i_cache()
{
        PRINT_DEBUGGING_INFO("CACHE (IGNORED!)", NOTHING);

}





static void rs4300i_ll()
{
        PRINT_DEBUGGING_INFO("LL", RT_O_B);
}





static void rs4300i_lwc1()
{
        PRINT_DEBUGGING_INFO("LWC1", FT_O_B);
}





static void rs4300i_lwc2()
{
        PRINT_DEBUGGING_INFO("LWC2", C2_O_B);
}





static void rs4300i_lld()
{
        PRINT_DEBUGGING_INFO("LLD", RT_O_B);
}





static void rs4300i_ldc1()
{
        PRINT_DEBUGGING_INFO("LDC1", FT_O_B);
}





static void rs4300i_ldc2()
{
        PRINT_DEBUGGING_INFO("LDC2", C2_O_B);
}





static void rs4300i_ld()
{
        PRINT_DEBUGGING_INFO("LD", RT_O_B);
}





static void rs4300i_sc()
{
        PRINT_DEBUGGING_INFO("SC", RT_O_B);
}





static void rs4300i_swc1()
{
        PRINT_DEBUGGING_INFO("SWC1", FT_O_B);
}





static void rs4300i_swc2()
{
        PRINT_DEBUGGING_INFO("SWC2", C2_O_B);
}





static void rs4300i_scd()
{
        PRINT_DEBUGGING_INFO("SCD", RT_O_B);
}





static void rs4300i_sdc1()
{
        PRINT_DEBUGGING_INFO("SDC1", FT_O_B);
}





static void rs4300i_sdc2()
{
        PRINT_DEBUGGING_INFO("SDC2", C2_O_B);
}





static void rs4300i_sd()
{
        PRINT_DEBUGGING_INFO("SD", RT_O_B);
}





static void rs4300i_special_sll()
{
        PRINT_DEBUGGING_INFO("SLL", RD_RT_SA);
}





static void rs4300i_special_srl()
{
        PRINT_DEBUGGING_INFO("SRL", RD_RT_SA);
}





static void rs4300i_special_sra()
{
        PRINT_DEBUGGING_INFO("SRA", RD_RT_SA);
}





static void rs4300i_special_sllv()
{
        PRINT_DEBUGGING_INFO("SLLV", RD_RT_RS);
}





static void rs4300i_special_srlv()
{
        PRINT_DEBUGGING_INFO("SRLV", RD_RT_RS);
}





static void rs4300i_special_srav()
{
        PRINT_DEBUGGING_INFO("SRAV", RD_RT_RS);
}





static void rs4300i_special_jr()
{
        PRINT_DEBUGGING_INFO("JR", RS);
}





static void rs4300i_special_jalr()
{
        PRINT_DEBUGGING_INFO("JALR", RD_RS);
}





static void rs4300i_special_syscall()
{
        PRINT_DEBUGGING_INFO("SYSCALL", NOTHING);
}





static void rs4300i_special_break()
{
        PRINT_DEBUGGING_INFO("BREAK", NOTHING);
}





static void rs4300i_special_sync()
{
        PRINT_DEBUGGING_INFO("SYNC (IGNORED!)", NOTHING); 
}





static void rs4300i_special_mfhi()
{
        PRINT_DEBUGGING_INFO("MFHI", RD);
}





static void rs4300i_special_mthi()
{
        PRINT_DEBUGGING_INFO("MTHI", RS);
}





static void rs4300i_special_mflo()
{
        PRINT_DEBUGGING_INFO("MFLO", RD);
}





static void rs4300i_special_mtlo()
{
        PRINT_DEBUGGING_INFO("MTLO", RS);
}





static void rs4300i_special_dsllv()
{
        PRINT_DEBUGGING_INFO("DSLLV", RD_RT_RS);
}





static void rs4300i_special_dsrlv()
{
        PRINT_DEBUGGING_INFO("DSRLV", RD_RT_RS);
}



static void rs4300i_special_dsrav()
{
        PRINT_DEBUGGING_INFO("DSRAV", RD_RT_RS);
}





static void rs4300i_special_mult()
{
        PRINT_DEBUGGING_INFO("MULT", RS_RT);
} /* static void rs4300i_special_mult() */





static void rs4300i_special_multu()
{
        PRINT_DEBUGGING_INFO("MULTU", RS_RT);
} /* static void rs4300i_special_multu() */





static void rs4300i_special_div()
{
        PRINT_DEBUGGING_INFO("DIV", RS_RT);
} /* static void rs4300i_special_div() */





static void rs4300i_special_divu()
{
        PRINT_DEBUGGING_INFO("DIVU", RS_RT);
} /* static void rs4300i_special_divu() */





static void rs4300i_special_dmult()
{
        PRINT_DEBUGGING_INFO("DMULT", RS_RT);
} /* static void rs4300i_special_dmult() */





static void rs4300i_special_dmultu()
{
        PRINT_DEBUGGING_INFO("DMULTU", RS_RT);
} /* static void rs4300i_special_dmultu() */





static void rs4300i_special_ddiv()
{
        PRINT_DEBUGGING_INFO("DDIV", RS_RT);
} /* static void rs4300i_special_ddiv() */





static void rs4300i_special_ddivu()
{
        PRINT_DEBUGGING_INFO("DDIVU", RS_RT);
} /* static void rs4300i_special_ddivu() */





static void rs4300i_special_add()
{
        PRINT_DEBUGGING_INFO("ADD", RD_RS_RT);
}





static void rs4300i_special_addu()
{
        PRINT_DEBUGGING_INFO("ADDU", RD_RS_RT);
}





static void rs4300i_special_sub()
{
        PRINT_DEBUGGING_INFO("SUB", RD_RS_RT);
}





static void rs4300i_special_subu()
{
        PRINT_DEBUGGING_INFO("SUBU", RD_RS_RT);
}





static void rs4300i_special_and()
{
        PRINT_DEBUGGING_INFO("AND", RD_RS_RT);
}





static void rs4300i_special_or()
{
        PRINT_DEBUGGING_INFO("OR", RD_RS_RT);
}





static void rs4300i_special_xor()
{
        PRINT_DEBUGGING_INFO("XOR", RD_RS_RT);
}





static void rs4300i_special_nor()
{
        PRINT_DEBUGGING_INFO("NOR", RD_RS_RT);
}





static void rs4300i_special_slt()
{
        PRINT_DEBUGGING_INFO("SLT", RD_RS_RT);
}





static void rs4300i_special_sltu()
{
        PRINT_DEBUGGING_INFO("SLTU", RD_RS_RT);
}





static void rs4300i_special_dadd()
{
        PRINT_DEBUGGING_INFO("DADD", RD_RS_RT);
}





static void rs4300i_special_daddu()
{
        PRINT_DEBUGGING_INFO("DADDU", RD_RS_RT);
}



static void rs4300i_special_dsub()
{
        PRINT_DEBUGGING_INFO("DSUB", RD_RS_RT);
}





static void rs4300i_special_dsubu()
{
        PRINT_DEBUGGING_INFO("DSUBU", RD_RS_RT);
}





static void rs4300i_special_tge()
{
        PRINT_DEBUGGING_INFO("TGE", RS_RT);
}





static void rs4300i_special_tgeu()
{
        PRINT_DEBUGGING_INFO("TGEU", RS_RT);
}





static void rs4300i_special_tlt()
{
        PRINT_DEBUGGING_INFO("TLT", RS_RT);
}





static void rs4300i_special_tltu()
{
        PRINT_DEBUGGING_INFO("TLTU", RS_RT);
}





static void rs4300i_special_teq()
{
        PRINT_DEBUGGING_INFO("TEQ", RS_RT);
}





static void rs4300i_special_tne()
{
        PRINT_DEBUGGING_INFO("TNE", RS_RT);
}





static void rs4300i_special_dsll()
{
        PRINT_DEBUGGING_INFO("DSLL", RD_RT_SA);
}





static void rs4300i_special_dsrl()
{
        PRINT_DEBUGGING_INFO("DSRL", RD_RT_SA);
}





static void rs4300i_special_dsra()
{
        PRINT_DEBUGGING_INFO("DSRA", RD_RT_SA);
}





static void rs4300i_special_dsll32()
{
        PRINT_DEBUGGING_INFO("DSLL32", RD_RT_SA);
}





static void rs4300i_special_dsrl32()
{
        PRINT_DEBUGGING_INFO("DSRL32", RD_RT_SA);
}





static void rs4300i_special_dsra32()
{
        PRINT_DEBUGGING_INFO("DSRA32", RD_RT_SA);
}





static void rs4300i_regimm_bltz()
{
        PRINT_DEBUGGING_INFO("BLTZ", RS_O);
}





static void rs4300i_regimm_bgez()
{
        PRINT_DEBUGGING_INFO("BGEZ", RS_O);
}





static void rs4300i_regimm_bltzl()
{
        PRINT_DEBUGGING_INFO("BLTZL", RS_O);
}





static void rs4300i_regimm_bgezl()
{
        PRINT_DEBUGGING_INFO("BGEZL", RS_O);
}





static void rs4300i_regimm_tgei()
{
        PRINT_DEBUGGING_INFO("TGEI", RS_I);
}





static void rs4300i_regimm_tgeiu()
{
        PRINT_DEBUGGING_INFO("TGEIU", RS_I);
}





static void rs4300i_regimm_tlti()
{
        PRINT_DEBUGGING_INFO("TLTI", RS_I);
}





static void rs4300i_regimm_tltiu()
{
        PRINT_DEBUGGING_INFO("TLTIU", RS_I);
}





static void rs4300i_regimm_teqi()
{
        PRINT_DEBUGGING_INFO("TEQI", RS_I);
}





static void rs4300i_regimm_tnei()
{
        PRINT_DEBUGGING_INFO("TNEI", RS_I);
}





static void rs4300i_regimm_bltzal()
{
        PRINT_DEBUGGING_INFO("BLTZAL", RS_O);
}





static void rs4300i_regimm_bgezal()
{
        PRINT_DEBUGGING_INFO("BGEZAL", RS_O);
}





static void rs4300i_regimm_bltzall()
{
        PRINT_DEBUGGING_INFO("BLTZALL", RS_O);
}





static void rs4300i_regimm_bgezall()
{
        PRINT_DEBUGGING_INFO("BGEZALL", RS_O);
}





/******************************************************************************\
*                                                                              *
*   CoProcessor0 (COP0)                                                        *
*                                                                              *
\******************************************************************************/



static void rs4300i_cop0_rs_mf()
{
        PRINT_DEBUGGING_INFO("MFC0", C0);
}





static void rs4300i_cop0_rs_dmf()
{
        PRINT_DEBUGGING_INFO("DMFC0", C0);
}





static void rs4300i_cop0_rs_cf()
{
        PRINT_DEBUGGING_INFO("CFC0", C0C);
}





static void rs4300i_cop0_rs_mt()
{
        PRINT_DEBUGGING_INFO("MTC0", C0);
}





static void rs4300i_cop0_rs_dmt()
{
        PRINT_DEBUGGING_INFO("DMTC0", C0);
}





static void rs4300i_cop0_rs_ct()
{
        PRINT_DEBUGGING_INFO("CTC0", C0C);
}





static void rs4300i_cop0_rs_bc()
{
        PRINT_XINST("cop0 bc (rt) instruction");

        cop0_rt_instruction[__RT]();
  
}





static void rs4300i_cop0_rs_co()
{
        PRINT_XINST("cop0 co (funct) instruction");

        cop0_instruction[__F]();
   
}





static void rs4300i_cop0_rt_bcf()
{
        PRINT_NOT_IMPLEMENTED("BCF0");
}





static void rs4300i_cop0_rt_bct()
{
        PRINT_NOT_IMPLEMENTED("BCT0");
}





static void rs4300i_cop0_rt_bcfl()
{
        PRINT_NOT_IMPLEMENTED("BCFL0");
}





static void rs4300i_cop0_rt_bctl()
{
        PRINT_NOT_IMPLEMENTED("BCTL0");
}






static void rs4300i_cop0_tlbr()
{
        PRINT_DEBUGGING_INFO("TLBR", NOTHING);
} /* static void rs4300i_cop0_tlbr() */





static void rs4300i_cop0_tlbwi()
{
        PRINT_DEBUGGING_INFO("TLBWI", NOTHING);
} /* static void rs4300i_cop0_tlbwi() */





static void rs4300i_cop0_tlbwr()
{
        PRINT_DEBUGGING_INFO("TLBWR", NOTHING);
} /* static void rs4300i_cop0_tlbwr() */





static void rs4300i_cop0_tlbp()
{
        PRINT_DEBUGGING_INFO("TLBP", NOTHING);
} /* static void rs4300i_cop0_tlbp() */










static void rs4300i_cop0_eret()
{
        PRINT_DEBUGGING_INFO("ERET", NOTHING);
} /* static void rs4300i_cop0_eret() */





/******************************************************************************\
*                                                                              *
*   CoProcessor1 (COP1)                                                        *
*                                                                              *
\******************************************************************************/



static void rs4300i_cop1_rs_mf()
{
        PRINT_DEBUGGING_INFO("MFC1", RT_FS);
}





static void rs4300i_cop1_rs_dmf()
{
        PRINT_DEBUGGING_INFO("DMFC1", RT_FS);
}





static void rs4300i_cop1_rs_cf()
{
        PRINT_DEBUGGING_INFO("CFC1", RT_C1);
}





static void rs4300i_cop1_rs_mt()
{
        PRINT_DEBUGGING_INFO("MTC1", RT_FS);
}





static void rs4300i_cop1_rs_dmt()
{
        PRINT_DEBUGGING_INFO("DMTC1", RT_FS);
}





static void rs4300i_cop1_rs_ct()
{
        PRINT_DEBUGGING_INFO("CFC1", RT_C1);
}





static void rs4300i_cop1_rs_bc()
{
        PRINT_XINST("cop1 bc (rt) instruction");

        cop1_rt_instruction[__RT]();

}





static void rs4300i_cop1_rs_co()
{
        PRINT_XINST("cop1 co (funct) instruction");

        cop1_instruction[__F]();
   
}





static void rs4300i_cop1_rt_bcf()
{
        PRINT_DEBUGGING_INFO("BCF1", O);
}





static void rs4300i_cop1_rt_bct()
{
        PRINT_DEBUGGING_INFO("BCT1", O);
}





static void rs4300i_cop1_rt_bcfl()
{
        PRINT_DEBUGGING_INFO("BCFL1", O);
}





static void rs4300i_cop1_rt_bctl()
{
        PRINT_DEBUGGING_INFO("BCTL1", O);
}





static void rs4300i_cop1_add()
{
        PRINT_DEBUGGING_INFO_FPU("ADD.", FD_FS_FT);
}





static void rs4300i_cop1_sub()
{
        PRINT_DEBUGGING_INFO_FPU("SUB.", FD_FS_FT);
}





static void rs4300i_cop1_mul()
{
        PRINT_DEBUGGING_INFO_FPU("MUL.", FD_FS_FT);
}





static void rs4300i_cop1_div()
{
        PRINT_DEBUGGING_INFO_FPU("DIV.", FD_FS_FT);
}





static void rs4300i_cop1_sqrt()
{
        PRINT_DEBUGGING_INFO_FPU("SQRT.", FD_FS);
}





static void rs4300i_cop1_abs()
{
        PRINT_DEBUGGING_INFO_FPU("ABS.", FD_FS);
}





static void rs4300i_cop1_mov()
{
        PRINT_DEBUGGING_INFO_FPU("MOV.", FD_FS);

}





static void rs4300i_cop1_neg()
{
      PRINT_DEBUGGING_INFO_FPU("NEG.", FD_FS);
}





static void rs4300i_cop1_roundl()
{
        PRINT_DEBUGGING_INFO_FPU("ROUND.L.", FD_FS);
}





static void rs4300i_cop1_truncl()
{
        PRINT_DEBUGGING_INFO_FPU("TRUNC.L.", FD_FS);
}





static void rs4300i_cop1_ceill()
{
        PRINT_DEBUGGING_INFO_FPU("CEIL.L.", FD_FS);
}





static void rs4300i_cop1_floorl()
{
        PRINT_DEBUGGING_INFO_FPU("FLOOR.L.", FD_FS);
}





static void rs4300i_cop1_roundw()
{
        PRINT_DEBUGGING_INFO_FPU("FLOOR.W.", FD_FS);
}





static void rs4300i_cop1_truncw()
{
        PRINT_DEBUGGING_INFO_FPU("TRUNC.W.", FD_FS);
}





static void rs4300i_cop1_ceilw()
{
        PRINT_DEBUGGING_INFO_FPU("CEIL.W.", FD_FS);
}





static void rs4300i_cop1_floorw()
{
        PRINT_DEBUGGING_INFO_FPU("FLOOR.W.", FD_FS);
}





static void rs4300i_cop1_cvts()
{
        PRINT_DEBUGGING_INFO_FPU("CVT.S.", FD_FS);
}





static void rs4300i_cop1_cvtd()
{
        PRINT_DEBUGGING_INFO_FPU("CVT.S.", FD_FS);
}





static void rs4300i_cop1_cvtw()
{
        PRINT_DEBUGGING_INFO_FPU("CVT.W.", FD_FS);
}





static void rs4300i_cop1_cvtl()
{
        PRINT_DEBUGGING_INFO_FPU("CVT.L.", FD_FS);
}





static void rs4300i_cop1_c()
{
//        PRINT_DEBUGGING_INFO_FPU_COND;
} /* void rs4300i_cop1_c() */






/******************************************************************************\
*                                                                              *
*   CoProcessor2 (COP2)                                                        *
*                                                                              *
\******************************************************************************/

/* 32 cop2 rs instructions */

static void rs4300i_cop2_rs_not_implemented()
{
        PRINT_NOT_IMPLEMENTED("cop2");
}

