/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

#ifndef __MNEMONIC_H
#define __MNEMONIC_H


#include <type_sizes.h>
#include "../file_loader_extern.h"

/**********************\
*                      *
* instruction reader   *
*                      *
\**********************/


#define READ_CPU_INSTRUCTION(addr)                                                             \
(                                                                                              \
        ( (addr < 0xa4000040) )                                                                \
        ?                                                                                      \
                ( *((uWORD *)(mem.rd_ram + (addr & 0x1fffffff))) )     /* instr in prgcode */  \
        :                                                                                      \
                ( (addr < 0xe0000000) )                                                        \
                ?                                                                              \
                        ( *(uWORD *)(mem.sp_dmem + (addr & 0xfff)) )   /* instr in bootcode */ \
                :                                                                              \
                        ( *(uWORD *)(rom.image + (addr & 0xfff)) )              /* for test */ \
)





/**********************\
*                      *
* mnemonic definitions *
*                      *
\**********************/

#define __CODE          (reg.code = READ_CPU_INSTRUCTION(reg.pc))
#define __NEXT_CODE     (READ_CPU_INSTRUCTION((reg.pc + 4)))
#define __PREV_CODE     (READ_CPU_INSTRUCTION((reg.pc - 4)))

#define __OPCODE        ((uBYTE)(__CODE >> 26))

#define __RS            (((uBYTE)(reg.code >> 21)) & 0x1f)

#define __RT            (((uBYTE)(reg.code >> 16)) & 0x1f)

#define __RD            (((uBYTE)(reg.code >> 11)) & 0x1f)

#define __SA            (((uBYTE)(reg.code >>  6)) & 0x1f)

#define __F             ( (uBYTE)(reg.code)       & 0x3f)

#define __I             ( (sWORD)(sHWORD)reg.code )
#define __O             ( reg.pc + 4 + (__I << 2) )

#define ____T           (reg.code & 0x3ffffff)
#define __T             ( (reg.pc & 0xf0000000) | (____T << 2) )





/******************************\
*                              *
* mnemonic definitions for fpu *
*                              *
\******************************/

#define FMT_S   16      /* 32bit binary floating-point */
#define FMT_D   17      /* 64bit binary floating-point */
#define FMT_W   20      /* 32bit binary fixed-point    */
#define FMT_L   21      /* 64bit binary fixed-point    */


#define __FMT   __RS

#define __FT    __RT

#define __FS    __RD

#define __FD    __SA

#define FT                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FT]                                        \
        :                                                               \
                ((reg.cpr[1][__FT+1] << 32) | (reg.cpr[1][__FT] & 0xffffffff))       \
)

#define FD                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FD]                                        \
        :                                                               \
                ((reg.cpr[1][__FD+1] << 32) | (reg.cpr[1][__FD] & 0xffffffff))       \
)


#define FS                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FS]                                        \
        :                                                               \
                ((reg.cpr[1][__FS+1] << 32) | (reg.cpr[1][__FS] & 0xffffffff))       \
)


#define STORE_FPR(value)                                                \
{                                                                       \
        if(reg.cpr[0][12] & 0x04000000)                                 \
        {                                                               \
                (reg.cpr[1][__FD] = value);                             \
        }                                                               \
        else                                                            \
        {                                                               \
                (reg.cpr[1][__FD+0] = (sDWORD)(sWORD)(value & 0xffffffff));          \
                (reg.cpr[1][__FD+1] = (sDWORD)(sWORD)(value >> 32));    \
        }                                                               \
}

#define STORE_FPR_F(value)                                                \
{                                                                       \
	(reg.cpr[1][__FD] = value);                             \
}


#define STORE_FPR_D(value)                                                \
{                                                                       \
	(reg.cpr[1][__FD+0] = (sDWORD)(sWORD)(value & 0xffffffff));         \
    (reg.cpr[1][__FD+1] = (sDWORD)(sWORD)(value >> 32));				\
}


/*
#define FT                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FT]                                        \
        :                                                               \
                ((reg.cpr[1][__FT+1] << 32) | (reg.cpr[1][__FT] & 0xffffffff))       \
)

#define FD                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FD]                                        \
        :                                                               \
                ((reg.cpr[1][__FD+1] << 32) | (reg.cpr[1][__FD] & 0xffffffff))       \
)


#define FS                                                              \
(                                                                       \
        (reg.cpr[0][12] & 0x04000000) ?                                 \
                reg.cpr[1][__FS]                                        \
        :                                                               \
                ((reg.cpr[1][__FS+1] << 32) | (reg.cpr[1][__FS] & 0xffffffff))       \
)


#define STORE_FPR(value)                                                \
{                                                                       \
        if(reg.cpr[0][12] & 0x04000000)                                 \
        {                                                               \
                (reg.cpr[1][__FD] = value);                             \
        }                                                               \
        else                                                            \
        {                                                               \
                (reg.cpr[1][__FD+0] = (sDWORD)(sWORD)value);            \
                (reg.cpr[1][__FD+1] = (sDWORD)(sWORD)(value >> 32));    \
        }                                                               \
}
*/        




/**********************\
*                      *
* mnemonic debugging   *
*                      *
\**********************/

#define PRINT_NOT_IMPLEMENTED(x) ;

#define PRINT_SYMBOL(a)                                                 \
{                                                                       \
        int     i;                                                      \
                                                                        \
        for(i=0; prefs.symbols==USE_SYMBOLS && i<prefs.n_symbols; i++)  \
        {                                                               \
                if(((prefs.symtab)[i]).addr == a)                       \
                {                                                       \
                        printf(" (%s)", ((prefs.symtab)[i]).name);      \
                }                                                       \
        }                                                               \
}
#endif


// for debug mesages
	#define PRINT_DEBUGGING_INFO(name, format) ;
	#define PRINT_DEBUGGING_INFO_FPU(name, format) ;
	#define PRINT_DEBUGGING_INFO_FPU_COND ;
	#define PRINT_XINST(x) ;
	#define CPU_HALT ;


        #define PRINT_NOTHING

        #define PRINT_O
        #define PRINT_RD
        #define PRINT_RS
        #define PRINT_T

        #define PRINT_RD_RS
        #define PRINT_RS_I
        #define PRINT_RS_O
        #define PRINT_RS_RD
        #define PRINT_RS_RT
        #define PRINT_RT_I
        #define PRINT_RT_RD

        #define PRINT_RD_RS_RT
        #define PRINT_RD_RT_SA
        #define PRINT_RD_RT_RS
        #define PRINT_RS_RT_O
        #define PRINT_RT_O_B
        #define PRINT_RT_RS_I

        #define PRINT_FD_FS_FT
        #define PRINT_FD_FS
        #define PRINT_FT_O_B
        #define PRINT_RT_FS

        #define PRINT_C0
        #define PRINT_C0C
        #define PRINT_RT_C1
        #define PRINT_C2_O_B



		#define FD_FS_FT
		#define FD_FS