/**
 * TrueReality - N64/rdp.c
 * Copyright (C) 1998 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

/**
 * Version information:
 * 1999042700
 *      Moved most variables into the global reg_rdp struct.
 *
 * 1999040600
 *      Implemented MOVEMEM VIEWPORT.
 *
 * 1999040500:
 *      So much new HLE stuff.
 *
 * 1999032300:
 *      More HLE stuff.
 *
 * 1999032200:
 *      Fixed instruction table.
 *      Added HLE stuff.
 *
 * 1999031900:
 *      Added the missing commands to process display lists directly.
 *
**/



#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>
#include <memory_extern.h>
#include <type_sizes.h>

#include "../output/display/display.h"

#include "rsp.h"
#include "rdp.h"
#include "rsp_registers.h"
#include "rdp_registers.h"
#include "mnemonic.h"
#include "registers.h"
#include "memory.h"


#include "rdp_gl.h"


struct  rdp_reg rdp_reg;
//** Debug
//	#define WRITE_LOG		  //** Write RDP to LogFile
//** Libary
//	#define WAVERUNNER_SE   //** from Jeff 
//  #define WAVERUNNER_US


void    rdp_dl_check();                 /* executes a displaylist @ address a */
static  uWORD segoffset2addr(uWORD so);   /* calcs the real addr out of seg and offset */
static  void matrix_mult(float r[4][4], float m1[4][4], float m2[4][4]);
static  void vertex_mult(float r[4], float v[4], float m[4][4]);





/* rdp instructions (audio & graphics) */

static void rdp_fixme();
static void rdp_spnoop();





/* rdp audio instructions */

static void rdp_adpcm();
static void rdp_clearbuff();
static void rdp_envmixer();
static void rdp_loadbuff();
static void rdp_resample();
static void rdp_savebuff();
static void rdp_segment();
static void rdp_setbuff();
static void rdp_setvol();
static void rdp_dmemmove();
static void rdp_loadadpcm();
static void rdp_mixer();
static void rdp_interleave();
static void rdp_polef();
static void rdp_setloop();





/* rdp graphics instructions */

static void rdp_matrix();
static void rdp_reserved0();
static void rdp_movemem();
static void rdp_vertex();
static void rdp_reserved1();
static void rdp_displaylist();
static void rdp_reserved2();
static void rdp_reserved3();
static void rdp_sprite2d();
static void rdp_tri1();
static void rdp_culldl();
static void rdp_popmatrix();
static void rdp_moveword();
static void rdp_texturemode();
static void rdp_setothermode_h();
static void rdp_setothermode_l();
static void rdp_enddl();
static void rdp_setgeometrymode();
static void rdp_cleargeometrymode();
static void rdp_line3d();
static void rdp_rdphalf_1();
static void rdp_rdphalf_2();
static void rdp_rdphalf_cont();
static void rdp_tri2();
static void rdp_noop();
static void rdp_texrect();
static void rdp_texrectflip();
static void rdp_loadsync();
static void rdp_pipesync();
static void rdp_tilesync();
static void rdp_fullsync();
static void rdp_setkeygb();
static void rdp_setkeyr();
static void rdp_setconvert();
static void rdp_setscissor();
static void rdp_setprimdepth();
static void rdp_setothermode();
static void rdp_loadtlut();
static void rdp_settilesize();
static void rdp_loadblock();
static void rdp_loadtile();
static void rdp_settile();
static void rdp_fillrect();
static void rdp_setfillcolor();
static void rdp_setfogcolor();
static void rdp_setblendcolor();
static void rdp_setprimcolor();
static void rdp_setenvcolor();
static void rdp_setcombine();
static void rdp_settextureimage();
static void rdp_setdepthimage();
static void rdp_setcolorimage();
static void rdp_trifill();
static void rdp_trishade();
static void rdp_tritxtr();
static void rdp_trishadetxtr();
static void rdp_trifillz();
static void rdp_trishadez();
static void rdp_tritxtrz();
static void rdp_trishadetxtrz();





/* rdp audio instruction function pointer */

static rdp_instr aud_instruction[256] =
{
    /**
    * Just 16 commands are used (as far as i know)
    **/
        rdp_spnoop,             rdp_adpcm,              rdp_clearbuff,          rdp_envmixer,
        rdp_loadbuff,           rdp_resample,           rdp_savebuff,           rdp_segment,
        rdp_setbuff,            rdp_setvol,             rdp_dmemmove,           rdp_loadadpcm,
        rdp_mixer,              rdp_interleave,         rdp_polef,              rdp_setloop,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme
};





/* rdp graphics instruction function pointer */

static rdp_instr gfx_instruction[256] =
{
    /**
    * DMA commands:
    * The next 64 commands (just 10 are used) are not executed by the RDP.
    * They are "emulated" by the RSP!
    * But in the display lists they appear with these bit combinations.
    * We will need this if we execute the display lists directly in future.
    * These commands include a pointer to mem in rdmem which is loaded via DMA.
    **/
        rdp_spnoop,             rdp_matrix,             rdp_reserved0,          rdp_movemem,
        rdp_vertex,             rdp_reserved1,          rdp_displaylist,        rdp_reserved2,
        rdp_reserved3,          rdp_sprite2d,           rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
    /**
    * The next 64 commands are not used (as far as i know).
    **/
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
    /**
    * IMMEDIATE commands:
    * The next 64 commands (just 14 are used) are not executed by the RDP.
    * They are like the commands above except that they do not carry a pointer.
    * All data they need is in the command itself.
    * So no DMA transfer is needed.
    **/
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_tri2,               rdp_rdphalf_cont,       rdp_rdphalf_2,
        rdp_rdphalf_1,          rdp_line3d,             rdp_cleargeometrymode,  rdp_setgeometrymode,
        rdp_enddl,              rdp_setothermode_l,     rdp_setothermode_h,     rdp_texturemode,
        rdp_moveword,           rdp_popmatrix,          rdp_culldl,             rdp_tri1,
    /**
    * RDP commands:
    * These 64 commands are the real RDP commands.
    * I've never seen the TRI commands. They might also be generated by the RSP.
    **/
        rdp_noop,               rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_trifill,            rdp_trifillz,           rdp_tritxtr,            rdp_tritxtrz,
        rdp_trishade,           rdp_trishadez,          rdp_trishadetxtr,       rdp_trishadetxtrz,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_texrect,            rdp_texrectflip,        rdp_loadsync,           rdp_pipesync,
        rdp_tilesync,           rdp_fullsync,           rdp_setkeygb,           rdp_setkeyr,
        rdp_setconvert,         rdp_setscissor,         rdp_setprimdepth,       rdp_setothermode,
        rdp_loadtlut,           rdp_fixme,              rdp_settilesize,        rdp_loadblock,
        rdp_loadtile,           rdp_settile,            rdp_fillrect,           rdp_setfillcolor,
        rdp_setfogcolor,        rdp_setblendcolor,      rdp_setprimcolor,       rdp_setenvcolor,
        rdp_setcombine,         rdp_settextureimage,    rdp_setdepthimage,      rdp_setcolorimage
};





/* define macros - to speed up things and to make things looking nice */


	#define PRINT_RDP_AUDIO_NOT_IMPLEMENTED(x) ;
	#define PRINT_RDP_NOT_IMPLEMENTED(name);
	#define PRINT_RDP_WARNING(text);
	#define PRINT_RDP_INFO(name) writeit(name);
	void    PRINT_RDP_MNEMONIC(char *fmt, ...)
	{
#ifdef WRITE_LOG
		FILE *stream;
		va_list ap;

		va_start(ap, fmt);
		stream = fopen("Display.lst","at");
		vfprintf(stream, fmt, ap);
		va_end(ap);

		fclose(stream);
#endif
	}



void writeit(char *text)
{
#ifdef WRITE_LOG
	FILE *stream;
	stream = fopen("Display.lst","at");
	fprintf(stream, "%s\n", text);
	fclose(stream);
#endif 
}






/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - global routines                          *
*                                                                              *
\******************************************************************************/




void rdp_go()
{
        return;
} /* void rdp_go() */





void rdp_dis()
{

} /* void rdp_dis() */





void rdp_reset()
{
        rdp_reg.pc_i = 0;
        rdp_reg.halt = 1;
		rdp_reg.m_CurTile = &rdp_reg.td[0];
} /* void rdp_reset() */






void rdp_step()
{
        uWORD    a;


        a = rdp_reg.pc[rdp_reg.pc_i];

        rdp_reg.cmd0 = ((uWORD *)mem.rd_ram)[(a>>2)+0];
        rdp_reg.cmd1 = ((uWORD *)mem.rd_ram)[(a>>2)+1];


        rdp_reg.curr = ((uWORD *)mem.dpc_reg)[2] = \
        rdp_reg.pc[rdp_reg.pc_i] = a + 8;   /* point to next instruction */

        switch(rdp_reg.task)
        {
            case AUDIO_TASK:
    /* if audio is not compiled in we simulate the BREAK instr immediately */
//#ifdef AUDIO
                aud_instruction[rdp_reg.cmd0>>24]();
//#endif
                rdp_reg.size -= 8;
                if(rdp_reg.size == 0)
                {

                        PRINT_RDP_MNEMONIC("RDP stops here - RSP BREAK will be simulated\n");
						rdp_fullsync();
                        rsp_reg.code = 0x0000000d;
                        rsp_dis(0x0000000d);   /* RSP BREAK instruction */
                        rdp_reg.halt = 1;
//                 seems not to work with this
//                        reg.do_or_check_sthg = RS4300I_DO_AI_INTERRUPT;
//                
                }
                break;

            case GFX_TASK:
                gfx_instruction[rdp_reg.cmd0>>24]();
                break;
        }

} /* void rdp_step() */




extern int DrawBackground;
void rdp_dl_check()
{
        uWORD us, uds, dss;



        us  = ((uWORD *)mem.sp_dmem)[0xfd4>>2];  /* ucode size */
        uds = ((uWORD *)mem.sp_dmem)[0xfdc>>2];  /* ucode data size */
        dss = ((uWORD *)mem.sp_dmem)[0xfe4>>2];  /* dram stack size */

        switch(((uWORD *)mem.sp_dmem)[0xfc0>>2])
        {
            case 1:                             /* Graphic Task */
                PRINT_RDP_MNEMONIC("GRAPHICS TASK -");

                rdp_reg.task = GFX_TASK;

                if(us != 0x1000)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(uds != 0x800)
                {
                        PRINT_RDP_MNEMONIC("\n        UCODE DATA SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(dss < 640)
                {
                        PRINT_RDP_MNEMONIC("\n        DRAM STACK SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }

                ClearVisual();
				DrawBackground = 0;
				PRINT_RDP_INFO("\n************************************ \n");
				PRINT_RDP_INFO("*********** IT BEGINS **************\n");
				PRINT_RDP_INFO("************************************\n");
                break;

            case 2:                             /* Audio Task */
                PRINT_RDP_MNEMONIC("AUDIO TASK -");

                rdp_reg.task = AUDIO_TASK;

                if(us > 0x1000)
                {
                        PRINT_RDP_MNEMONIC("	UCODE SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(uds > 0x800)
                {
                        PRINT_RDP_MNEMONIC("	UCODE DATA SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                if(dss != 0)
                {
                        PRINT_RDP_MNEMONIC("	DRAM STACK SIZE TEST FAILED");
                        rdp_reg.task = INVALID_TASK;
                }
                break;

            default:
                PRINT_RDP_MNEMONIC("	UNKNOWN TASK - WILL NOT RUN HLE");
                return;

        } /* switch(mem.sp_dmem[0xfc0>>2]) */

    
    
    /* for audio and graphic task */

        if(rdp_reg.task == INVALID_TASK)
        {
                puts("\n        WILL NOT RUN IN HLE !!!");
        }
        else
        {
                rdp_reg.pc[0] = rdp_reg.curr = ((uWORD *)mem.dpc_reg)[2] =
                        ((uWORD *)mem.sp_dmem)[0xff0>>2];
                rdp_reg.size = ((uWORD *)mem.sp_dmem)[0xff4>>2];  /* we need this in audio task */
                rdp_reg.pc_i = 0;
                rdp_reg.halt = 0;
                PRINT_RDP_MNEMONIC(" WILL RUN HLE\n");
        }

} /* void rdp_dl_check() */







/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - helper routines                          *
*                                                                              *
\******************************************************************************/

uWORD *adr_test;
static uWORD segoffset2addr(uWORD so)
{
	return( (rdp_reg.segment[ (so>>24) & 0x0f] + (so & 0x00ffffff) & 0x003fffff) );
} /* static uWORD segoffset2addr(uWORD so) */





/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - static routines (graphics and audio)     *
*                                                                              *
\******************************************************************************/



static void rdp_fixme()
{
	PRINT_RDP_NOT_IMPLEMENTED("FIXME")
} /* static void rdp_fixme() */


static void rdp_spnoop()
{
	PRINT_RDP_INFO("SPNOOP")
} /* static void rdp_spnoop() */


/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - static graphic routines                  *
*                                                                              *
\******************************************************************************/



/**
*
* Most matrix stuff was first done by Jeff.
* Great work, Jeff! (Niki)
*
**/

static void rdp_matrix()
{

	uWORD   a = segoffset2addr(rdp_reg.cmd1);
	uBYTE   command = (uBYTE)((rdp_reg.cmd0 >> 16) & 0xff);
	float  m[4][4], *mtx;
	int     i=0,j=0;

	PRINT_RDP_INFO("MATRIX ")

	a = a >> 1;   // we have half uWORDs 

	mtx = &m[0][0];
	for(i = 0; i < 16; i+=4)
	{
		for(j = 0; j < 4; j++)
			{
				m[i>>2][j] = (float)(
							 (
                             (((sWORD)((uHWORD *)mem.rd_ram)[(a + i + j)^1]) << 16)
                             |
                             ((uHWORD *)mem.rd_ram)[(a + i + j + 16)^1]
                             ) / 65536.0);
			}
	}

        switch(command)
        {
            case 0: /* modelview  mul  nopush */
				glMatrixMode(GL_MODELVIEW);
                glMultMatrixf(mtx);
				PRINT_RDP_INFO("rdp_matrix() - modelview  mul  nopush");
                break;

            case 1: /* projection mul  nopush */
            case 5: /* projection mul  push   */ /* the same, because there is just 1 proj matrix */
				glMatrixMode(GL_PROJECTION);
				glMultMatrixf(mtx);
				glMatrixMode(GL_MODELVIEW);
				PRINT_RDP_INFO("rdp_matrix() - projection mul  nopush");
                break;

            case 2: /* modelview  load nopush */
                /*
                * we can use memcopy here because we have native float values here
                * -> endian does not matter
                */
				glMatrixMode(GL_MODELVIEW);
				glLoadMatrixf(mtx);
				PRINT_RDP_INFO("rdp_matrix() - modelview  load nopush");
                break;

            case 3: /* projection load nopush */
            case 7: /* projection load push   */ /* the same, because there is just 1 proj matrix */ /* thanks Gil */
				glMatrixMode(GL_PROJECTION);
				glLoadMatrixf(mtx);
				glMatrixMode(GL_MODELVIEW);
				PRINT_RDP_INFO("rdp_matrix() - projection load nopush");
                break;

            case 4: /* modelview  mul  push   */								
				glMatrixMode(GL_MODELVIEW);
				glPushMatrix();
                glMultMatrixf(mtx);
				PRINT_RDP_INFO("rdp_matrix() - modelview  mul  push"); 
                break;

            case 6: /* modelview  load push   */
				glMatrixMode(GL_MODELVIEW);
				glPushMatrix();				
                glLoadMatrixf(mtx);
				PRINT_RDP_INFO("rdp_matrix() - modelview  load push");

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MATRIX: UNKNOWN COMMAND")
                CPU_HALT

        } /* switch(command) */

        //CPU_HALT


} /* static void rdp_matrix() */





static void rdp_reserved0()
{
	PRINT_RDP_NOT_IMPLEMENTED("RESERVED0")
} /* static void rdp_reserved0() */





static void rdp_movemem()
{

        uWORD a;



    /* we do '>> 1' here because we just need sHWORD's */
        a = segoffset2addr(rdp_reg.cmd1) >> 1;

        switch((rdp_reg.cmd0 >> 16) & 0xff)
        {
            case 0x80:
                rdp_reg.vp[0] = ((float)((sHWORD *)mem.rd_ram)[(a+0)^1]) / 4.0f;
                rdp_reg.vp[1] = ((float)((sHWORD *)mem.rd_ram)[(a+1)^1]) / 4.0f;
                rdp_reg.vp[2] = ((float)((sHWORD *)mem.rd_ram)[(a+2)^1]) / 4.0f;
                rdp_reg.vp[3] = ((float)((sHWORD *)mem.rd_ram)[(a+3)^1]) / 4.0f;
                rdp_reg.vp[4] = ((float)((sHWORD *)mem.rd_ram)[(a+4)^1]) / 4.0f;
                rdp_reg.vp[5] = ((float)((sHWORD *)mem.rd_ram)[(a+5)^1]) / 4.0f;
                rdp_reg.vp[6] = ((float)((sHWORD *)mem.rd_ram)[(a+6)^1]) / 4.0f;
                rdp_reg.vp[7] = ((float)((sHWORD *)mem.rd_ram)[(a+7)^1]) / 4.0f;
                PRINT_RDP_INFO("MOVEMEM VIEWPORT: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%ld, offset=$%lx)\n",
                        a<<1,
                        (rdp_reg.cmd1 >> 24) & 0x0f,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC(
                        "        scale: % 7.2f % 7.2f % 7.2f % 7.2f\n" 
                        "        trans: % 7.2f % 7.2f % 7.2f % 7.2f\n",
                        rdp_reg.vp[0],
                        rdp_reg.vp[1],
                        rdp_reg.vp[2],
                        rdp_reg.vp[3],
                        rdp_reg.vp[4],
                        rdp_reg.vp[5],
                        rdp_reg.vp[6],
                        rdp_reg.vp[7] );
                break;

            case 0x82:
                rdp_reg.lookat_x.rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.lookat_x.rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.lookat_x.dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.lookat_x.dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.lookat_x.dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.lookat_x.dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LOOKATY")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.lookat_x.rgba,
                        rdp_reg.lookat_x.rgba_copy,
                        rdp_reg.lookat_x.dir[0],
                        rdp_reg.lookat_x.dir[1],
                        rdp_reg.lookat_x.dir[2],
                        (uBYTE)rdp_reg.lookat_x.dir[0],
                        (uBYTE)rdp_reg.lookat_x.dir[1],
                        (uBYTE)rdp_reg.lookat_x.dir[2] );
                break;

            case 0x84:
                rdp_reg.lookat_y.rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.lookat_y.rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.lookat_y.dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.lookat_y.dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.lookat_y.dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.lookat_y.dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LOOKATY")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.lookat_y.rgba,
                        rdp_reg.lookat_y.rgba_copy,
                        rdp_reg.lookat_y.dir[0],
                        rdp_reg.lookat_y.dir[1],
                        rdp_reg.lookat_y.dir[2],
                        (uBYTE)rdp_reg.lookat_y.dir[0],
                        (uBYTE)rdp_reg.lookat_y.dir[1],
                        (uBYTE)rdp_reg.lookat_y.dir[2] );
                break;

            case 0x86:
                rdp_reg.light[0].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[0].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[0].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[0].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[0].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[0].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 0: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[0].rgba,
                        rdp_reg.light[0].rgba_copy,
                        rdp_reg.light[0].dir[0],
                        rdp_reg.light[0].dir[1],
                        rdp_reg.light[0].dir[2],
                        (uBYTE)rdp_reg.light[0].dir[0],
                        (uBYTE)rdp_reg.light[0].dir[1],
                        (uBYTE)rdp_reg.light[0].dir[2] );
                break;

            case 0x88:
                rdp_reg.light[1].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[1].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[1].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[1].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[1].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[1].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 1: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[1].rgba,
                        rdp_reg.light[1].rgba_copy,
                        rdp_reg.light[1].dir[0],
                        rdp_reg.light[1].dir[1],
                        rdp_reg.light[1].dir[2],
                        (uBYTE)rdp_reg.light[1].dir[0],
                        (uBYTE)rdp_reg.light[1].dir[1],
                        (uBYTE)rdp_reg.light[1].dir[2] );
                break;

            case 0x8a:
                rdp_reg.light[2].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[2].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[2].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[2].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[2].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[2].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 2: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[2].rgba,
                        rdp_reg.light[2].rgba_copy,
                        rdp_reg.light[2].dir[0],
                        rdp_reg.light[2].dir[1],
                        rdp_reg.light[2].dir[2],
                        (uBYTE)rdp_reg.light[2].dir[0],
                        (uBYTE)rdp_reg.light[2].dir[1],
                        (uBYTE)rdp_reg.light[2].dir[2] );
                break;

            case 0x8c:
                rdp_reg.light[3].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[3].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[3].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[3].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[3].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[3].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 3: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[3].rgba,
                        rdp_reg.light[3].rgba_copy,
                        rdp_reg.light[3].dir[0],
                        rdp_reg.light[3].dir[1],
                        rdp_reg.light[3].dir[2],
                        (uBYTE)rdp_reg.light[3].dir[0],
                        (uBYTE)rdp_reg.light[3].dir[1],
                        (uBYTE)rdp_reg.light[3].dir[2] );
                break;

            case 0x8e:
                rdp_reg.light[4].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[4].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[4].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[4].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[4].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[4].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 4: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[4].rgba,
                        rdp_reg.light[4].rgba_copy,
                        rdp_reg.light[4].dir[0],
                        rdp_reg.light[4].dir[1],
                        rdp_reg.light[4].dir[2],
                        (uBYTE)rdp_reg.light[4].dir[0],
                        (uBYTE)rdp_reg.light[4].dir[1],
                        (uBYTE)rdp_reg.light[4].dir[2] );
                break;

            case 0x90:
                rdp_reg.light[5].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[5].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[5].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[5].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[5].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[5].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 5: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[5].rgba,
                        rdp_reg.light[5].rgba_copy,
                        rdp_reg.light[5].dir[0],
                        rdp_reg.light[5].dir[1],
                        rdp_reg.light[5].dir[2],
                        (uBYTE)rdp_reg.light[5].dir[0],
                        (uBYTE)rdp_reg.light[5].dir[1],
                        (uBYTE)rdp_reg.light[5].dir[2] );
                break;

            case 0x92:
                rdp_reg.light[6].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[6].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[6].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[6].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[6].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[6].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 6: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[6].rgba,
                        rdp_reg.light[6].rgba_copy,
                        rdp_reg.light[6].dir[0],
                        rdp_reg.light[6].dir[1],
                        rdp_reg.light[6].dir[2],
                        (uBYTE)rdp_reg.light[6].dir[0],
                        (uBYTE)rdp_reg.light[6].dir[1],
                        (uBYTE)rdp_reg.light[6].dir[2] );
                break;

            case 0x94:
                rdp_reg.light[7].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                rdp_reg.light[7].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                rdp_reg.light[7].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                rdp_reg.light[7].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                rdp_reg.light[7].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                rdp_reg.light[7].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];

                PRINT_RDP_INFO("MOVEMEM LIGHT 7: ")
                PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx)\n",
                        a,
                        rdp_reg.cmd1 >> 24,
                        rdp_reg.cmd1 & 0x00ffffff );
                PRINT_RDP_MNEMONIC("	rgba=$%08lx rgba_cp=$%08lx xyz=%d,%d,%d ($%02x,$%02x,$%02x)\n",
                        rdp_reg.light[7].rgba,
                        rdp_reg.light[7].rgba_copy,
                        rdp_reg.light[7].dir[0],
                        rdp_reg.light[7].dir[1],
                        rdp_reg.light[7].dir[2],
                        (uBYTE)rdp_reg.light[7].dir[0],
                        (uBYTE)rdp_reg.light[7].dir[1],
                        (uBYTE)rdp_reg.light[7].dir[2] );
                break;

            case 0x98:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 0")
                CPU_HALT
                break;

            case 0x9a:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 1")
                CPU_HALT
                break;

            case 0x9c:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 2")
                CPU_HALT
                break;

            case 0x9e:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 3")
                CPU_HALT
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM ?")
                CPU_HALT

        } /* switch((rdp_reg.cmd0 >> 8) & 0xffff) */


} /* static void rdp_movemem() */


static void rdp_vertex()
{
        uWORD   a = segoffset2addr(rdp_reg.cmd1) & 0x003fffff;
        uHWORD  len = (uHWORD)(rdp_reg.cmd0 & 0xffff);
        int     v0, i, n;

		float TScaleT, TScaleS, width, height;

		PRINT_RDP_INFO("VERTEX")

		v0 = (rdp_reg.cmd0 >> 16) & 0xf;
//** orginal        n = ((rdp_reg.cmd0 >> 20) & 0xf) + 1;
		n = len / 0x10;

#ifdef WAVERUNNER_SE   
		v0 = ((rdp_reg.cmd0 >> 16) & 0x2f) >> 0x1;
		n  = (len+1) / 0x410;
#endif

#ifdef WAVERUNNER_US
		v0 = ((rdp_reg.cmd0 >> 16) & 0x2f) >> 0x1;
		v0 = ((rdp_reg.cmd0 >> 17) & 0x2f);
		n  = (len+1) / 0x210;
#endif


		rdp_reg.m_CurTile=&rdp_reg.td[0];

		width=(float)((short)rdp_reg.m_CurTile->lrs >> rdp_reg.m_CurTile->shifts);
		if(width==0) width=1.f;

		height=(float)((short)rdp_reg.m_CurTile->lrt >> rdp_reg.m_CurTile->shiftt);
		if(height==0) height=1.f;

		TScaleS = rdp_reg.m_CurTile->SScale / width;
		TScaleT = rdp_reg.m_CurTile->TScale / height;


        for(i = 0; i < (n<<4); i+=16)
        {
			rdp_reg.vtx[v0].x     = (float) ((sHWORD *)mem.rd_ram)[(((a+i) >> 1) + 0)^1];
			rdp_reg.vtx[v0].y     = (float) ((sHWORD *)mem.rd_ram)[(((a+i) >> 1) + 1)^1];
			rdp_reg.vtx[v0].z     = (float) ((sHWORD *)mem.rd_ram)[(((a+i) >> 1) + 2)^1];
			rdp_reg.vtx[v0].flags =			((uHWORD *)mem.rd_ram)[(((a+i) >> 1) + 3)^1];
			rdp_reg.vtx[v0].s     = ((float)((sHWORD *)mem.rd_ram)[(((a+i) >> 1) + 4)^1]) * TScaleS;
			rdp_reg.vtx[v0].t     = ((float)((sHWORD *)mem.rd_ram)[(((a+i) >> 1) + 5)^1]) * TScaleT;
			rdp_reg.vtx[v0].r     = ((uBYTE  *)mem.rd_ram)[(a+i+12)^3];
			rdp_reg.vtx[v0].g     = ((uBYTE  *)mem.rd_ram)[(a+i+13)^3];
			rdp_reg.vtx[v0].b     = ((uBYTE  *)mem.rd_ram)[(a+i+14)^3];
			rdp_reg.vtx[v0].a     = ((uBYTE  *)mem.rd_ram)[(a+i+15)^3];
			v0++;
		}
} /* static void rdp_vertex() */ 




static void rdp_reserved1()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED1")

        CPU_HALT

} /* static void rdp_reserved1() */





static void rdp_displaylist()
{

        uWORD addr = segoffset2addr(rdp_reg.cmd1);

        PRINT_RDP_INFO("DISPLAYLIST: ")

        rdp_reg.pc_i++;
        if(rdp_reg.pc_i > 9)
        {
                PRINT_RDP_WARNING("DL STACK OVERFLOW (too much (>9) DISPLAYLIST calls\n"
                                  "will not call dl!!! - ignored")
                return;
        }

        PRINT_RDP_MNEMONIC("	call dl#%d; addr=$%08lx (seg#%d, offset=$%lx)\n",
                rdp_reg.pc_i,
                addr,
                (rdp_reg.cmd1 >> 24) & 0x0f,
                rdp_reg.cmd1 & 0x00ffffff );

        rdp_reg.pc[rdp_reg.pc_i] = addr;

} /* static void rdp_displaylist() */





static void rdp_reserved2()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED2")

        CPU_HALT

} /* static void rdp_reserved2() */





static void rdp_reserved3()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED3")

        CPU_HALT

} /* static void rdp_reserved3() */





static void rdp_sprite2d()
{

/*        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };

        uWORD    a               = segoffset2addr(rdp_reg.cmd1);
        uWORD    dst_p;  //** dest pointer 
        uWORD    dst_w;  //** dest with (normally 320 or 640) 
        uWORD    dst_b;  //** dest depth (16bit: 2; 32bit: 3) 

        uWORD    src             = ((uWORD *) mem.rd_ram)[((a>>2)+ 0)^0];
        uWORD    src_ci          = ((uWORD *) mem.rd_ram)[((a>>2)+ 1)^0];
        uHWORD   src_w           = ((uHWORD *)mem.rd_ram)[((a>>1)+ 4)^1];
        uHWORD   sprite_w        = ((uHWORD *)mem.rd_ram)[((a>>1)+ 5)^1];
        uHWORD   sprite_h        = ((uHWORD *)mem.rd_ram)[((a>>1)+ 6)^1];
        uBYTE    src_type        = ((uBYTE *) mem.rd_ram)[( a    +14)^3];   //** rgba, ci, ai, a, i 
        uBYTE    src_bits        = ((uBYTE *) mem.rd_ram)[( a    +15)^3];   //** 4, 8, 16, 32 
        float   scale_x         = ((float)((uHWORD *)mem.rd_ram)[((a>>1)+ 8)^1]) / 1024.f;   //** s5.10 
        float   scale_y         = ((float)((uHWORD *)mem.rd_ram)[((a>>1)+ 9)^1]) / 1024.f;   //** s5.10 
        uBYTE    flip_x          = ((uBYTE *) mem.rd_ram)[( a    +20)^3];
        uBYTE    flip_y          = ((uBYTE *) mem.rd_ram)[( a    +21)^3];
        uHWORD   src_ox          = ((uHWORD *)mem.rd_ram)[((a>>1)+11)^1];
        uHWORD   src_oy          = ((uHWORD *)mem.rd_ram)[((a>>1)+12)^1];
        uHWORD   dst_ox          = ((uHWORD *)mem.rd_ram)[((a>>1)+13)^1];
        uHWORD   dst_oy          = ((uHWORD *)mem.rd_ram)[((a>>1)+14)^1];
        dst_b = ((uWORD *)mem.vi_reg)[0] & 0x03;
        if(dst_b != 2 && dst_b != 3)
        {
                PRINT_RDP_NOT_IMPLEMENTED("SPRITE2D: dest screen has unsupported depth\n")
                CPU_HALT
                return;
        }
        dst_w = ((uWORD *)mem.vi_reg)[2];
        dst_p = ((uWORD *)mem.vi_reg)[1];
        dst_p &= 0x003fffff;

        PRINT_RDP_INFO("SPRITE2D: ")
        PRINT_RDP_MNEMONIC("	addr=$%08lx (seg#%d, offset=$%lx):\n"
                           "        src     : $%08lx\n"
                           "        src_ci  : $%08lx\n"
                           "        src_w   : $%04lx (%d)\n"
                           "        sprite_w: $%04lx (%d)\n"
                           "        sprite_h: $%04lx (%d)\n"
                           "        src_type: $%02lx   (%s)\n"
                           "        src_bits: $%02lx   (%s)\n"
                           "        scale_x : %f\n"
                           "        scale_y : %f\n"
                           "        flip_x  : $%02lx\n"
                           "        flip_y  : $%02lx\n"
                           "        src_ox  : $%04lx (%d)\n"
                           "        src_oy  : $%04lx (%d)\n"
                           "        dst_ox  : $%04lx (%d)\n"
                           "        dst_oy  : $%04lx (%d)\n",
                           a, (rdp_reg.cmd1 >> 24) & 0x0f, rdp_reg.cmd1 & 0x00ffffff,
                           src, src_ci, src_w, src_w,
                           sprite_w, sprite_w, sprite_h, sprite_h, src_type, format[src_type], src_bits, size[src_bits],
                           scale_x, scale_y,
                           flip_x, flip_y,
                           src_ox, src_ox, src_oy, src_oy,
                           dst_ox, dst_ox, dst_oy, dst_oy);
        PRINT_RDP_WARNING("scale and flip are ignored")



        //** BUG? the following might be just used with the TEXRECT and FILLRECT commands 
        //** - i don't know 
        if(!(rdp_reg.mode_h & 0x00200000))
        {
            //** if we are not in copy or fill mode (we are in 1 or 2 cycle mode) 
            /** we must not draw the bottom and right edges 
                //xl--;
                //yl--;
        }

//**
        draw2d(src, src_ci, src_type, src_bits, src_w, src_ox, src_oy,
                sprite_w, sprite_h, flip_x, flip_y,
                dst_p, 0, dst_b, dst_w, dst_ox, dst_oy );

*/
} //** static void rdp_sprite2d() 


	


static void rdp_tri1()
{
        int   vn[3], f;
		uWORD cmd0 = rdp_reg.cmd0;
		uWORD cmd1 = rdp_reg.cmd1;

        PRINT_RDP_INFO("TRI1")

        vn[0] = ((rdp_reg.cmd1 >> 16) & 0xff) / 10;
        vn[1] = ((rdp_reg.cmd1 >>  8) & 0xff) / 10;
        vn[2] = ( rdp_reg.cmd1        & 0xff) / 10;

#ifdef WAVERUNNER_SE 
        vn[0] = ((rdp_reg.cmd1 >> 16) & 0xff) / 2;
        vn[1] = ((rdp_reg.cmd1 >>  8) & 0xff) / 2;
        vn[2] = ( rdp_reg.cmd1        & 0xff) / 2;
#endif
#ifdef WAVERUNNER_US
        vn[0] = ((rdp_reg.cmd1 >> 16) & 0xff) / 5;
        vn[1] = ((rdp_reg.cmd1 >>  8) & 0xff) / 5;
        vn[2] = ( rdp_reg.cmd1        & 0xff) / 5;
#endif
        f     =  (rdp_reg.cmd1 >> 24) & 0xff;

        DrawVisualTriangle(vn);
         
} /* static void rdp_tri1() */

static void rdp_tri2()
{
        int vn[6];

		PRINT_RDP_INFO("TRI2")
        vn[0] = ((rdp_reg.cmd0 >> 16) & 0xff) / 10;
        vn[1] = ((rdp_reg.cmd0 >>  8) & 0xff) / 10;
        vn[2] = ( rdp_reg.cmd0        & 0xff) / 10;
        vn[3] = ((rdp_reg.cmd1 >> 16) & 0xff) / 10;
        vn[4] = ((rdp_reg.cmd1 >>  8) & 0xff) / 10;
        vn[5] = ( rdp_reg.cmd1        & 0xff) / 10;

#ifdef WAVERUNNER_SE  
        vn[0] = ((rdp_reg.cmd0 >> 16) & 0xff) / 2;
        vn[1] = ((rdp_reg.cmd0 >>  8) & 0xff) / 2;
        vn[2] = ( rdp_reg.cmd0        & 0xff) / 2;
        vn[3] = ((rdp_reg.cmd1 >> 16) & 0xff) / 2;
        vn[4] = ((rdp_reg.cmd1 >>  8) & 0xff) / 2;
        vn[5] = ( rdp_reg.cmd1        & 0xff) / 2;
#endif

        DrawVisualTriangle(vn);
		DrawVisualTriangle(vn+3);

} /* static void rdp_tri2() */




static void rdp_culldl()
{
        PRINT_RDP_NOT_IMPLEMENTED("CULLDL")

        CPU_HALT

} /* static void rdp_culldl() */





static void rdp_popmatrix()
{
    /* Compleately from Gil - thanks */
        uWORD            param;

	

        PRINT_RDP_INFO("POPMATRIX ")

        param = rdp_reg.cmd1;

        switch(param)
        {
            case 0: //** modelview pop 
					glMatrixMode(GL_MODELVIEW);
					glPopMatrix();
                break;

            case 1: //** projection pop 
                    //** only 1 projection matrix . Not possible! 
                break;
        }

} /* static void rdp_popmatrix() */





static void rdp_moveword()
{

        switch(rdp_reg.cmd0 & 0xff)
        {
            case 0x00:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD MATRIX")
                CPU_HALT
                break;

            case 0x02:
                rdp_reg.lights = (rdp_reg.cmd1 - 0x80000000) / 32 - 1;
                PRINT_RDP_INFO("MOVEWORD NUMLIGHT: ")
                PRINT_RDP_MNEMONIC("	lights: %d\n", rdp_reg.lights);
                break;

            case 0x04:
                PRINT_RDP_INFO("MOVEWORD CLIP: ")
                switch((rdp_reg.cmd0 >> 8) & 0xffff)
                {
                    case 0x0004:
                        PRINT_RDP_MNEMONIC("	-x = ");
                        rdp_reg.clip.nx = rdp_reg.cmd1;
                        break;

                    case 0x000c:
                        PRINT_RDP_MNEMONIC("	-y = ");
                        rdp_reg.clip.ny = rdp_reg.cmd1;
                        break;

                    case 0x0014:
                        PRINT_RDP_MNEMONIC("	+x = ");
                        rdp_reg.clip.px = rdp_reg.cmd1;
                        break;

                    case 0x001c:
                        PRINT_RDP_MNEMONIC("	+y = ");
                        rdp_reg.clip.py = rdp_reg.cmd1;
                        break;

                    default:
                        PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD CLIP - wrong offset")
                        CPU_HALT

                } /* switch((rdp_reg.cmd0 >> 8) & 0xffff) */
                
                PRINT_RDP_MNEMONIC("	%d\n", (sWORD)(sHWORD)rdp_reg.cmd1);
                PRINT_RDP_WARNING("ignored");
                break;

            case 0x06:
                PRINT_RDP_INFO("MOVEWORD SEGMENT: ")
                PRINT_RDP_MNEMONIC("	$%08lx -> seg#%d\n", rdp_reg.cmd1, (rdp_reg.cmd0 >> 10) & 0xf);
                rdp_reg.segment[(rdp_reg.cmd0 >> 10) & 0xf] = rdp_reg.cmd1;
                break;

            case 0x08:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD FOG")
                CPU_HALT
                break;

            case 0x0a:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD LIGHTCOL")
                CPU_HALT
                break;

            case 0x0c:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD POINTS")
                CPU_HALT
                break;

            case 0x0e:
                PRINT_RDP_INFO("MOVEWORD PERSPNORM: ")
                PRINT_RDP_MNEMONIC("	factor: %d ($%lx)\n", rdp_reg.cmd1, rdp_reg.cmd1);
            /**
            *
            * We don't need this. This is to make multipltiplication for s15.16 values 
            * more precise. We use float values. They are precise enough!
            *
            **/
/*
                perspective_normal = ((float)rdp_reg.cmd1) / 65536.0;
*/
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD ?")
                CPU_HALT

        } /* switch(rdp_reg.cmd0 & 0xff) */

} /* static void rdp_moveword() */





static void rdp_texturemode()
{
        int tile = (rdp_reg.cmd0 >> 8)  & 0x07;				//** tile
		uWORD mipmap_level = (rdp_reg.cmd0 >> 11) & 0x07;	//** mipmap_level   - not used yet
		uWORD on = (rdp_reg.cmd0 & 0xff);			//** 1: on - 0:off

		sHWORD s = ((rdp_reg.cmd1 >> 16) & 0xffff);
		sHWORD t =  (rdp_reg.cmd1 & 0xffff);

		t_tile *tmp_tile = &rdp_reg.td[tile];
		tmp_tile->Texture_on = on;

//** Skale S and T 
		if(s<=1) tmp_tile->SScale=1.f;
		else tmp_tile->SScale=(float )s/65535.f;

		if(t<=1) tmp_tile->TScale=1.f;
		else  tmp_tile->TScale=(float )t/65535.f;

		tmp_tile->TScale/=32.f;
		tmp_tile->SScale/=32.f;


        PRINT_RDP_INFO("TEXTUREMODE");
/*      PRINT_RDP_MNEMONIC("	tile: %ld; scale_s: %9.6f; scale_t: %9.6f; mipmap_level: %ld; on: %ld\n",
                tile,
                rdp_reg.texture[tile].scale_s,
                rdp_reg.texture[tile].scale_t,
                rdp_reg.texture[tile].mipmap_level,
                rdp_reg.texture[tile].on );
*/        
} /* static void rdp_texture() */





static void rdp_setothermode_h()
{
        static char *ad[] = { "pattern", "notpattern", "noise", "disable" };
        static char *rd[] = { "magicsq", "bayer", "noise", "?" };
        static char *ck[] = { "none", "key" };
        static char *tc[] = { "conv", "?", "?", "?", "?", "filtconv", "filt", "?" };
        static char *tf[] = { "point", "?", "bilerp", "average" };
        static char *tt[] = { "none", "?", "rgba16", "ia16" };
        static char *tl[] = { "tile", "lod" };
        static char *td[] = { "clamp", "sharpen", "detail", "?" };
        static char *tp[] = { "none", "persp" };
        static char *ct[] = { "1cycle", "2cycle", "copy", "fill" };
        static char *cd[] = { "disable(hw>1)", "enable(hw>1)", "disable(hw1)", "enable(hw1)" };
        static char *pm[] = { "nprimitive", "1primitive" };

        switch((rdp_reg.cmd0 >> 8) & 0xff)
        {
            case 0x00:
                PRINT_RDP_WARNING("SETOTHERMODE_H BLENDMASK - ignored")
                CPU_HALT
                break;

            case 0x04:
                PRINT_RDP_INFO("SETOTHERMODE_H ALPHADITHER: ")
                PRINT_RDP_MNEMONIC("	%s\n", ad[(rdp_reg.cmd1>>0x04) & 0x3]);
                rdp_reg.mode_h &= ~0x00000030;
                rdp_reg.cmd1   &=  0x00000030;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x06:
                PRINT_RDP_INFO("SETOTHERMODE_H RGBDITHER: ")
                PRINT_RDP_MNEMONIC("	%s\n", rd[(rdp_reg.cmd1>>0x06) & 0x3]);
                rdp_reg.mode_h &= ~0x000000c0;
                rdp_reg.cmd1   &=  0x000000c0;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x08:
                PRINT_RDP_INFO("SETOTHERMODE_H COMBINEKEY: ")
                PRINT_RDP_MNEMONIC("	%s\n", ck[(rdp_reg.cmd1>>0x08) & 0x1]);
                rdp_reg.mode_h &= ~0x00000100;
                rdp_reg.cmd1   &=  0x00000100;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x09:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTURECONVERT: ")
                PRINT_RDP_MNEMONIC("	%s\n", tc[(rdp_reg.cmd1>>0x09) & 0x7]);
                rdp_reg.mode_h &= ~0x00000e00;
                rdp_reg.cmd1   &=  0x00000e00;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x0c:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTUREFILTER: ")
                PRINT_RDP_MNEMONIC("	%s\n", tf[(rdp_reg.cmd1>>0x0c) & 0x3]);
                rdp_reg.mode_h &= ~0x00003000;
                rdp_reg.cmd1   &=  0x00003000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x0e:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTURELUT: ")
                PRINT_RDP_MNEMONIC("	%s\n", tt[(rdp_reg.cmd1>>0x0e) & 0x3]);
                rdp_reg.mode_h &= ~0x0000c000;
                rdp_reg.cmd1   &=  0x0000c000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x10:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTURELOD: ")
                PRINT_RDP_MNEMONIC("	%s\n", tl[(rdp_reg.cmd1>>0x10) & 0x1]);
                rdp_reg.mode_h &= ~0x00010000;
                rdp_reg.cmd1   &=  0x00010000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x11:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTUREDETAIL: ")
                PRINT_RDP_MNEMONIC("	%s\n", td[(rdp_reg.cmd1>>0x11) & 0x3]);
                rdp_reg.mode_h &= ~0x00060000;
                rdp_reg.cmd1   &=  0x00060000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x13:
                PRINT_RDP_INFO("SETOTHERMODE_H TEXTUREPERSP: ")
                PRINT_RDP_MNEMONIC("	%s\n", tp[(rdp_reg.cmd1>>0x13) & 0x1]);
                rdp_reg.mode_h &= ~0x00080000;
                rdp_reg.cmd1   &=  0x00080000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x14:
                PRINT_RDP_INFO("SETOTHERMODE_H CYCLETYPE: ")
                PRINT_RDP_MNEMONIC("	%s\n", ct[(rdp_reg.cmd1>>0x14) & 0x3]);
                rdp_reg.mode_h &= ~0x00300000;
                rdp_reg.cmd1   &=  0x00300000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x16:
                PRINT_RDP_INFO("SETOTHERMODE_H COLORDITHER: ")
                PRINT_RDP_MNEMONIC("	%s\n", cd[(rdp_reg.cmd1>>0x16) & 0x1]);
                rdp_reg.mode_h &= ~0x00400000;
                rdp_reg.cmd1   &=  0x00400000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            case 0x17:
                PRINT_RDP_INFO("SETOTHERMODE_H PIPELINEMODE: ")
                PRINT_RDP_MNEMONIC("	%s\n", pm[(rdp_reg.cmd1>>0x17) & 0x1]);
                rdp_reg.mode_h &= ~0x00800000;
                rdp_reg.cmd1   &=  0x00800000;
                rdp_reg.mode_h |=  rdp_reg.cmd1;
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_H ?")
                CPU_HALT

        } /* switch((rdp_reg.cmd0 >> 8) & 0xff) */


} /* static void rdp_setothermode_h() */





static void rdp_setothermode_l()
{
        static char *ac[] = { "none", "threshold", "?", "diter" };
        static char *zs[] = { "pixel", "prim" };
        static char *a1[] =
                {
                        "        bl_1ma (1)",
                        "        bl_a_mem (1)",
                        "        bl_1 (1)",
                        "        bl_0 (1)"
                };
        static char *b1[] =
                {
                        "        bl_clr_in (1)",
                        "        bl_clr_mem (1)",
                        "        bl_clr_bl (1)",
                        "        bl_clr_fog (1)"
                };
        static char *c1[] =
                {
                        "        bl_a_in (1)",
                        "        bl_a_fog (1)",
                        "        bl_a_shade (1)",
                        "        bl_0 (1)"
                };
        static char *d1[] =
                {
                        "        bl_1ma (1)",
                        "        bl_a_mem (1)",
                        "        bl_1 (1)",
                        "        bl_0 (1)" 
                };
        static char *a2[] =
                {
                        "        bl_1ma (2)",
                        "        bl_a_mem (2)",
                        "        bl_1 (2)",
                        "        bl_0 (2)"
                };
        static char *b2[] =
                {
                        "        bl_clr_in (2)",
                        "        bl_clr_mem (2)",
                        "        bl_clr_bl (2)",
                        "        bl_clr_fog (2)"
                };
        static char *c2[] =
                {
                        "        bl_a_in (2)",
                        "        bl_a_fog (2)",
                        "        bl_a_shade (2)",
                        "        bl_0 (2)"
                };
        static char *d2[] =
                {
                        "        bl_1ma (2)",
                        "        bl_a_mem (2)",
                        "        bl_1 (2)",
                        "        bl_0 (2)" 
                };

        switch((rdp_reg.cmd0 >> 8) & 0xff)
        {
            case 0x00:
                PRINT_RDP_INFO("SETOTHERMODE_L ALPHACOMPARE: ")
                PRINT_RDP_MNEMONIC("	%s\n", ac[(rdp_reg.cmd1>>0x00) & 0x3]);
                rdp_reg.mode_l &= ~0x00000003;
                rdp_reg.cmd1   &=  0x00000003;
                rdp_reg.mode_l |=  rdp_reg.cmd1;
                break;

            case 0x02:
                PRINT_RDP_INFO("SETOTHERMODE_L ZSRCSEL: ")
                PRINT_RDP_MNEMONIC("	%s\n", zs[(rdp_reg.cmd1>>0x02) & 0x1]);
                rdp_reg.mode_l &= ~0x00000004;
                rdp_reg.cmd1   &=  0x00000004;
                rdp_reg.mode_l |=  rdp_reg.cmd1;
                break;

            case 0x03:
                PRINT_RDP_INFO("SETOTHERMODE_L RENDERMODE: ")
                PRINT_RDP_MNEMONIC("	$%08lx:\n", rdp_reg.cmd1 & 0xfffffff8,
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n",
                        rdp_reg.cmd1 & 0xfffffff8,
                        (rdp_reg.cmd1 & 0x00000008) ? "        anti alias\n" : "",
                        (rdp_reg.cmd1 & 0x00000010) ? "        z_cmp\n" : "",
                        (rdp_reg.cmd1 & 0x00000020) ? "        z_upd\n" : "",
                        (rdp_reg.cmd1 & 0x00000040) ? "        im_rd\n" : "",
                        (rdp_reg.cmd1 & 0x00000080) ? "        clr_on_cvg\n" : "",
                        (rdp_reg.cmd1 & 0x00000100) ? "        cvg_dst_warp\n" : "",
                        (rdp_reg.cmd1 & 0x00000200) ? "        cvg_dst_full\n" : "",
                        (rdp_reg.cmd1 & 0x00000400) ? "        z_inter\n" : "",
                        (rdp_reg.cmd1 & 0x00000800) ? "        z_xlu\n" : "",
                        (rdp_reg.cmd1 & 0x00001000) ? "        cvg_x_alpha\n" : "",
                        (rdp_reg.cmd1 & 0x00002000) ? "        alpha_cvg_sel\n" : "",
                        (rdp_reg.cmd1 & 0x00004000) ? "        force_bl\n" : "",
                        (rdp_reg.cmd1 & 0x00008000) ? "        tex_edge?\n" : "",
                        a2[(rdp_reg.cmd1>>16) & 0x3],
                        a1[(rdp_reg.cmd1>>18) & 0x3],
                        b2[(rdp_reg.cmd1>>20) & 0x3],
                        b1[(rdp_reg.cmd1>>22) & 0x3],
                        c2[(rdp_reg.cmd1>>24) & 0x3],
                        c1[(rdp_reg.cmd1>>26) & 0x3],
                        d2[(rdp_reg.cmd1>>28) & 0x3],
                        d1[(rdp_reg.cmd1>>30) & 0x3]
                        );

                rdp_reg.mode_l &= ~0xfffffff8;
                rdp_reg.cmd1   &=  0xfffffff8;
                rdp_reg.mode_l |=  rdp_reg.cmd1;
                break;

            case 0x16:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_L BLENDER")
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_L ?")
                CPU_HALT

        } /* switch((rdp_reg.cmd0 >> 8) & 0xff) */


} /* static void rdp_setothermode_l() */





static void rdp_enddl()
{

        PRINT_RDP_INFO("ENDDL: ")
        PRINT_RDP_MNEMONIC("	return from dl#%d\n", rdp_reg.pc_i);

        if(rdp_reg.pc_i < 0)
        {
                PRINT_RDP_WARNING("DL STACK UNDERRUN (too much ENDDL calls)\n"
                                  "will not return from dl!!! - ignored")
                return;
        }

        if(rdp_reg.pc_i == 0)
        {
                PRINT_RDP_MNEMONIC("	RDP stops here - RSP BREAK will be simulated\n");

                FlushVisualRenderBuffer();

                rsp_reg.code = 0x0000000d;
                rsp_dis(0x0000000d);   /* RSP BREAK instruction */
                rdp_reg.halt = 1;
        }

        rdp_reg.pc_i--;

} /* static void rdp_enddl() */




int Textures;
static void rdp_setgeometrymode()
{
        PRINT_RDP_INFO("SETGEOMETRYMODE ")
        PRINT_RDP_MNEMONIC(
                "+$%08lx:\n"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s",
                rdp_reg.cmd1,
                (rdp_reg.cmd1 & 0x00000001) ? "        zbuffer\n" : "",
                (rdp_reg.cmd1 & 0x00000002) ? "        texture\n" : "",
                (rdp_reg.cmd1 & 0x00000004) ? "        shade\n" : "",
                (rdp_reg.cmd1 & 0x00000200) ? "        shade smooth\n" : "",
                (rdp_reg.cmd1 & 0x00001000) ? "        cull front\n" : "",
                (rdp_reg.cmd1 & 0x00002000) ? "        cull back\n" : "",
                (rdp_reg.cmd1 & 0x00010000) ? "        fog\n" : "",
                (rdp_reg.cmd1 & 0x00020000) ? "        lightning\n" : "",
                (rdp_reg.cmd1 & 0x00040000) ? "        texture gen\n" : "",
                (rdp_reg.cmd1 & 0x00080000) ? "        texture gen lin\n" : "",
                (rdp_reg.cmd1 & 0x00100000) ? "        lod\n" : ""
                );

//** stuff by GiL
        rdp_reg.geometrymode |= rdp_reg.cmd1;
		if (rdp_reg.cmd1 & 0x00000001) glEnable(GL_DEPTH_TEST);
		if (rdp_reg.cmd1 & 0x00000001) Textures = 1;

		switch(rdp_reg.geometrymode & 0x00003000) //** Must be geometrymode, andnot cmd1 as far as I can see 
		{
			case 0x1000 : glCullFace(GL_FRONT); glEnable(GL_CULL_FACE); break;
			case 0x2000 : glCullFace(GL_BACK);  glEnable(GL_CULL_FACE); break;
			case 0x3000 : glCullFace(GL_FRONT_AND_BACK); glEnable(GL_CULL_FACE); break;
			default:      glDisable(GL_CULL_FACE);
		}
} /* static void rdp_setgeometrymode() */





static void rdp_cleargeometrymode()
{
        PRINT_RDP_INFO("CLEARGEOMETRYMODE ")
        PRINT_RDP_MNEMONIC(
                "-$%08lx\n"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s",
                rdp_reg.cmd1,
                (rdp_reg.cmd1 & 0x00000001) ? "        zbuffer\n" : "",
                (rdp_reg.cmd1 & 0x00000002) ? "        texture\n" : "",
                (rdp_reg.cmd1 & 0x00000004) ? "        shade\n" : "",
                (rdp_reg.cmd1 & 0x00000200) ? "        shade smooth\n" : "",
                (rdp_reg.cmd1 & 0x00001000) ? "        cull front\n" : "",
                (rdp_reg.cmd1 & 0x00002000) ? "        cull back\n" : "",
                (rdp_reg.cmd1 & 0x00010000) ? "        fog\n" : "",
                (rdp_reg.cmd1 & 0x00020000) ? "        lightning\n" : "",
                (rdp_reg.cmd1 & 0x00040000) ? "        texture gen\n" : "",
                (rdp_reg.cmd1 & 0x00080000) ? "        texture gen lin\n" : "",
                (rdp_reg.cmd1 & 0x00100000) ? "        lod\n" : ""
                );

        rdp_reg.geometrymode &= ~rdp_reg.cmd1;

		if (rdp_reg.cmd1 & 0x00000001) glDisable(GL_DEPTH_TEST);
		if (rdp_reg.cmd1 & 0x00000001) Textures = 0;

		switch(rdp_reg.geometrymode & 0x00003000) //** Must be geometrymode, andnot cmd1 as far as I can see 
		{
			case 0x1000 : glCullFace(GL_FRONT); glEnable(GL_CULL_FACE); break;
			case 0x2000 : glCullFace(GL_BACK);  glEnable(GL_CULL_FACE); break;
			case 0x3000 : glCullFace(GL_FRONT_AND_BACK); glEnable(GL_CULL_FACE); break;
			default:      glDisable(GL_CULL_FACE);
		}
} /* static void rdp_cleargeometrymode() */





static void rdp_line3d()
{
        PRINT_RDP_NOT_IMPLEMENTED("LINE3D")

        CPU_HALT

} /* static void rdp_line3d() */





static void rdp_rdphalf_1()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_1")

        CPU_HALT

} /* static void rdp_rdphalf_1() */





static void rdp_rdphalf_2()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_2")

        CPU_HALT

} /* static void rdp_rdphalf_2() */





static void rdp_rdphalf_cont()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_CONT")

        CPU_HALT

} /* static void rdp_rdphalf_cont() */





static void rdp_noop()
{
        PRINT_RDP_INFO("NOOP\n")

} /* static void rdp_noop() */


static void rdp_texrect()
{
        uWORD    a = rdp_reg.pc[rdp_reg.pc_i];
        uWORD    cmd2 = ((uWORD *)mem.rd_ram)[(a>>2)+1];
        uWORD    cmd3 = ((uWORD *)mem.rd_ram)[(a>>2)+3];

        uHWORD   xh     = (rdp_reg.cmd0 & 0x00fff000) >> 14; //1.
        uHWORD   yh     = (rdp_reg.cmd0 & 0x00000fff) >> 2;  //2.
        uHWORD   xl     = (rdp_reg.cmd1 & 0x00fff000) >> 14; //4
        uHWORD   yl     = (rdp_reg.cmd1 & 0x00000fff) >> 2;  //5
        int     tile   = (rdp_reg.cmd1 & 0x07000000) >> 24;  //3

        uHWORD   s      = (cmd2 & 0xffff0000) >> 16;  //6
        uHWORD   t      = cmd2 & 0x0000ffff; // 7

        float   f_s    = ((float)((cmd2 & 0xffff0000) >> 16) / 31.f);
        float   f_t    = ((float)(cmd2 & 0x0000ffff) / 31.f);                  /* is this correct jeff? */

        sHWORD  dsdx   = (cmd3 & 0xffff0000) >> 16; // 8
        sHWORD  dsdy   = cmd3 & 0x0000ffff; // 9
        float   f_dsdx = (float)((dsdx & 0xfc00) + (dsdx & 0x02ff) / 1023.f);
        float   f_dsdy = (float)((dsdy & 0xfc00) + (dsdy & 0x02ff) / 1023.f);

//        float   f_dsdx = (((float)((cmd3 & 0xffff0000) >> 16)) / 1023.f);
//        float   f_dsdy = (((float)(cmd3 & 0x0000ffff)) / 1023.f);


	    DrawTexRectangle((float)xl, 
						 (float)yl, 
						 (float)xh, 
						 (float)yh, f_s, f_t);


        PRINT_RDP_INFO("rdp_texrect()")

        PRINT_RDP_MNEMONIC("xh=%4hu; yh=%4hu; tile=%u; xl=%4hu; yl=%4hu; s=%+9.4f; t=%+9.4f; dsdx=%+7.4f; dtdy=%+7.4f\n",
               xh, yh, tile, xl, yl, f_s, f_t, f_dsdx, f_dsdy);
        PRINT_RDP_WARNING("some parameter are ignored")



    //** so rdp_rdphalf_1 and rdp_rdphalf_2 are never reached 
        ((uWORD *)mem.dpc_reg)[2] = a + 16;
        rdp_reg.pc[rdp_reg.pc_i] += 16;

} /* static void rdp_texrect() */


static void rdp_texrectflip()
{
        PRINT_RDP_NOT_IMPLEMENTED("TEXRECTFLIP")

        CPU_HALT

} /* static void rdp_texrectflip() */





static void rdp_loadsync()
{
        PRINT_RDP_INFO("LOADSYNC - ignored")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_loadsync() */





static void rdp_pipesync()
{
        PRINT_RDP_INFO("PIPESYNC - ignored")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_pipesync() */





static void rdp_tilesync()
{
        PRINT_RDP_INFO("TILESYNC - ignored")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_tilesync() */





static void rdp_fullsync()
{
        PRINT_RDP_INFO("FULLSYNC")

        reg.do_or_check_sthg = RS4300I_DO_DP_INTERRUPT;
        //((uWORD *)mem.dpc_reg)[2] = 0;

        //CPU_HALT

} /* static void rdp_fullsync() */





static void rdp_setkeygb()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYGB")

        CPU_HALT

} /* static void rdp_setkeygb() */





static void rdp_setkeyr()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYR")

        CPU_HALT

} /* static void rdp_setkeyr() */





static void rdp_setconvert()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETCONVERT")

        CPU_HALT

} /* static void rdp_setconvert() */





static void rdp_setscissor()
{
        rdp_reg.scissor.ulx             = (rdp_reg.cmd0 & 0x00fff000) >> 14;
        rdp_reg.scissor.uly             = (rdp_reg.cmd0 & 0x00000fff) >> 2;
        rdp_reg.scissor.interlace_mode  = (rdp_reg.cmd1 & 0x03000000) >> 24;
        rdp_reg.scissor.lrx             = (rdp_reg.cmd1 & 0x00fff000) >> 14;
        rdp_reg.scissor.lry             = (rdp_reg.cmd1 & 0x00000fff) >> 2;

    /* It seems that most demo coders thought that SETSCISSOR uses:
          x, y, width, height
       arguments. That is NOT correct. SETSCISSOR uses:
          ulx, uly, lrx, lry
       args!!! */

        PRINT_RDP_WARNING("interlace_mode is ignored")

        SetVisualClipRectangle(rdp_reg.scissor.ulx, rdp_reg.scissor.uly, rdp_reg.scissor.lrx, rdp_reg.scissor.lry);

} /* static void rdp_setscissor() */





static void rdp_setprimdepth()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETPRIMDEPTH")

        CPU_HALT

} /* static void rdp_setprimdepth() */





static void rdp_setothermode()
{
        PRINT_RDP_INFO("SETOTHERMODE")
        PRINT_RDP_WARNING("this should clear/init SETOTHERMODE_[HL] - is not done")

} /* static void rdp_setothermode() */





static void rdp_loadtlut()
{
        uWORD  tile   = (rdp_reg.cmd1 >> 24) & 0x07;
        uHWORD count  = ((uHWORD)(rdp_reg.cmd1 >> 14) & 0x03ff);


        PRINT_RDP_INFO("LOADTLUT ")
        PRINT_RDP_MNEMONIC("	tile=%ld (->tmem=%lx), count=%hx\n", 
               tile,
               rdp_reg.td[tile].tmem,
               count
               );

        rdp_reg.tlut_8_fmt  = rdp_reg.TextureImage.fmt;
        rdp_reg.tlut_8_size = rdp_reg.TextureImage.size;
        rdp_reg.tlut_8_addr = rdp_reg.TextureImage.addr;

        rdp_reg.tlut_4_fmt[rdp_reg.td[tile].palette]  = rdp_reg.TextureImage.fmt; 
        rdp_reg.tlut_4_size[rdp_reg.td[tile].palette] = rdp_reg.TextureImage.size;
        rdp_reg.tlut_4_addr[rdp_reg.td[tile].palette] = rdp_reg.TextureImage.addr;


} /* static void rdp_loadtlut() */





static void rdp_settilesize()
{
        uWORD tile = (rdp_reg.cmd1 >> 24) & 0x07;
		t_tile *tmpTile = &rdp_reg.td[tile];

        tmpTile->uls = (uHWORD)((rdp_reg.cmd0 >> 12) & 0x0fff);
        tmpTile->ult = (uHWORD)( rdp_reg.cmd0        & 0x0fff);
        tmpTile->lrs = (uHWORD)((rdp_reg.cmd1 >> 12) & 0x0fff);
        tmpTile->lrt = (uHWORD)( rdp_reg.cmd1        & 0x0fff);

        tmpTile->uls = (tmpTile->uls >> 2);
        tmpTile->ult = (tmpTile->ult >> 2);
        tmpTile->lrs = (tmpTile->lrs >> 2) + 1;
        tmpTile->lrt = (tmpTile->lrt >> 2) + 1;


		if(tmpTile->lrs == tmpTile->uls) tmpTile->lrs++;
		if(tmpTile->lrt == tmpTile->ult) tmpTile->lrt++;
		rdp_reg.m_CurTile=tmpTile;


		tmpTile->Width = tmpTile->lrs - tmpTile->uls;
		tmpTile->Height= tmpTile->lrt - tmpTile->ult;

//	WORD size=Width*Height*tmpTile->size;
//	if(size==0) size=(Width*Height)/2;

		SetCurTexture(rdp_reg.TextureImage.addr, rdp_reg.TextureImage.ID);


        PRINT_RDP_INFO("SETTILESIZE ")
        PRINT_RDP_MNEMONIC("	uls=%hu; ult=%hu; lrs=%hu; lrt=%hu; tile=%lu\n",
                rdp_reg.td[tile].uls,
                rdp_reg.td[tile].ult,
                rdp_reg.td[tile].lrs,
                rdp_reg.td[tile].lrt,
                tile
                );

} /* static void rdp_settilesize() */





static void rdp_loadblock()
{
		t_tile *tmpTile;
        uWORD TILE = (uWORD)((rdp_reg.cmd1 >> 24) & 0x07);

        uHWORD uls = (uHWORD)((rdp_reg.cmd0 >> 12) & 0x0fff);
        uHWORD ult = (uHWORD)( rdp_reg.cmd0        & 0x0fff);
        uHWORD lrs = (uHWORD)((rdp_reg.cmd1 >> 12) & 0x0fff);
        uHWORD dxt = (uHWORD)( rdp_reg.cmd1        & 0x0fff);
		uHWORD lrt = dxt;

		uHWORD width = (dxt>>11);
		uHWORD bytes = lrs;

		uDWORD Offset;

		tmpTile = &rdp_reg.td[TILE];
		if(dxt&0x7ff) width++;
			lrs++;

//** MAKE IT BETTER
		switch(tmpTile->size)
		{
			case 0:
				width>>=1;
				bytes>>=1;
				break;
			case 1:
				break;
			case 2:
				width<<=1;
				bytes<<=1;
				break;
			case 3:
				width<<=2;
				bytes<<=2;
				break;
		}

		width<<=1;
		if(tmpTile->size==0) lrs<<=2;
		if(tmpTile->size==1) lrs<<=1;
//** MAKE IT BETTER - end

// **  I doesnt know if we need this Offset (FiRES)
	uls/=4;
	lrs/=4;
	ult/=4;
	lrt/=4;

//	Offset = (width*tmpTile->size)+lrs;
//	rdp_reg.TextureImage.addr = rdp_reg.TextureImage.addr + Offset;


//	rdp_reg.TextureImage.LoadAddr = &mem.rd_ram[rdp_reg.TextureImage.addr];
//** Generate an ID to identify Block/Texture
	rdp_reg.TextureImage.ID = ( (uDWORD)width << 48) |
							  ( (uDWORD)bytes << 32) |
							  ( (uDWORD)ult   << 16) |
							  ( (uDWORD)dxt   <<  0);

	SetCurTexture((uDWORD)rdp_reg.TextureImage.addr, rdp_reg.TextureImage.ID);

	PRINT_RDP_INFO("LOADBLOCK ")
/*	PRINT_RDP_MNEMONIC("	uls=%hu; ult=%hu; lrs=%hu; (real line width)dxt=%hu(*64bit); tile=%lu (->tmem=%hx)\n",
			rdp_reg.td[TILE].uls,
			rdp_reg.td[TILE].ult,
			rdp_reg.td[TILE].lrs,
			rdp_reg.td[TILE].dxt,
			TILE,
			rdp_reg.td[TILE].tmem
			);
*/
} /* static void rdp_loadblock() */





static void rdp_loadtile()
{
    int CURR_TILE = (rdp_reg.cmd1 & 0x07000000) >> 24;
    int uls = (rdp_reg.cmd0 & 0x00fff000) >> 12;
	int	ult =  rdp_reg.cmd0 & 0x00000fff;
	int	lrs = (rdp_reg.cmd1 & 0x00fff000) >> 12; 
	int	lrt =  rdp_reg.cmd1 & 0x00000fff;
	uWORD width, height, size;
	uDWORD Offset;


	t_tile *tmpTile=&rdp_reg.td[CURR_TILE];

	uls = uls >> 2;
	lrs = lrs >> 2;
	ult = ult >> 2;
	lrt = lrt >> 2;

	if(lrs & 1) lrs++;
	if(lrt & 1) lrt++;

	width=(lrs-uls);
	height=(lrt-ult);
	size=rdp_reg.TextureImage.size;

	switch(rdp_reg.TextureImage.size)
	{
	case 0:
		size=1;
		width>>=1;
		rdp_reg.TextureImage.width>>=1;
		break;
	case 1:
		size=1;
		break;
	case 2:
		size=2;
		break;
	case 3:
		size=4;
		break;
	}

	tmpTile->Width = width;
	tmpTile->Height = height;


	Offset = (ult*(width*size)+uls);

	rdp_reg.TextureImage.addr += Offset;
								

	rdp_reg.TextureImage.ID =( (uDWORD)uls << 24) |
							 ( (uDWORD)lrs << 16) |
							 ( (uDWORD)ult <<  8) |
							   (uDWORD)lrt;

	SetCurTexture((uDWORD)rdp_reg.TextureImage.addr, rdp_reg.TextureImage.ID);
	
	PRINT_RDP_INFO("LOADTILE")
	PRINT_RDP_MNEMONIC("	uls=%lu, ult=%lu, tile=%lu; lrs=%lu, lrt=%lu\n", 
		(rdp_reg.cmd0 & 0x00fff000) >> 12, 
		(rdp_reg.cmd0 & 0x00000fff),
		(rdp_reg.cmd1 & 0x07000000) >> 24,
		(rdp_reg.cmd1 & 0x00fff000) >> 12, 
		(rdp_reg.cmd1 & 0x00000fff) );

} /* static void rdp_loadtile() */





static void rdp_settile()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };
        static char *cm[]       = { "NOMIRROR/WARP(NOCLAMP)", "MIRROR", "CLAMP", "MIRROR&CLAMP" };

        uWORD tile = (uWORD)((rdp_reg.cmd1 >> 24) & 0x07);
		t_tile *tmpTile = &rdp_reg.td[tile];

        tmpTile->format  = (uBYTE) ((rdp_reg.cmd0 >> 21) & 0x07);
        tmpTile->size    = (uBYTE) ((rdp_reg.cmd0 >> 19) & 0x02);
        tmpTile->line    = (uHWORD)((rdp_reg.cmd0 >>  9) & 0x01ff);
        tmpTile->tmem    = (uHWORD)( rdp_reg.cmd0        & 0x01ff);
        tmpTile->palette = (uBYTE) ((rdp_reg.cmd1 >> 20) & 0x0f);
        tmpTile->clampt  = (uBYTE) ((rdp_reg.cmd1 >> 19) & 0x01);
        tmpTile->mirrort = (uBYTE) ((rdp_reg.cmd1 >> 18) & 0x01);
        tmpTile->maskt   = (uBYTE) ((rdp_reg.cmd1 >> 14) & 0x0f);
        tmpTile->shiftt  = (uBYTE) ((rdp_reg.cmd1 >> 10) & 0x0f);
        tmpTile->clamps  = (uBYTE) ((rdp_reg.cmd1 >>  9) & 0x01);
        tmpTile->mirrors = (uBYTE) ((rdp_reg.cmd1 >>  8) & 0x01);
        tmpTile->masks   = (uBYTE) ((rdp_reg.cmd1 >>  4) & 0x0f);
        tmpTile->shifts  = (uBYTE) ( rdp_reg.cmd1        & 0x0f);

		if(tile!=7)
		{
			rdp_reg.m_CurTile=tmpTile;
		}



        //** setup line from "64bit pixel" to correct 32, 16, 8 or 4bit pixel 
/*        switch(rdp_reg.td[tile].size)
        {
                case 0:   //** 4bit 
                        rdp_reg.td[tile].line <<= 4;
                        break;
                case 1:   //** 8bit 
                        rdp_reg.td[tile].line <<= 3;
                        break;
                case 2:   //** 16bit 
                        rdp_reg.td[tile].line <<= 2;
                        break;
                case 3:   //** 32bit 
                        rdp_reg.td[tile].line <<= 1;
                        break;
        }

        PRINT_RDP_INFO("SETTILE ")
        PRINT_RDP_MNEMONIC(
                "	format=%s; size=%s; line=%lu (pixel); tmem=%lu; tile=%lu; palette=$%lx;\n"
                "	cmt=%s; maskt=%lu; shiftt=%lu; cms=%s; masks=%lu; shifts=%lu\n",
                format[rdp_reg.td[tile].format],
                size[rdp_reg.td[tile].size],
                rdp_reg.td[tile].line,
                rdp_reg.td[tile].tmem,
                tile,
                rdp_reg.td[tile].palette,
                cm[(rdp_reg.td[tile].clampt << 1) | rdp_reg.td[tile].mirrort],
                rdp_reg.td[tile].maskt,
                rdp_reg.td[tile].shiftt,
                cm[(rdp_reg.td[tile].clamps << 1) | rdp_reg.td[tile].mirrors],
                rdp_reg.td[tile].masks,
                rdp_reg.td[tile].shifts);
*/
} /* static void rdp_settile() */





static void rdp_fillrect()
{
        uWORD    ulx, uly, lrx, lry;
        uWORD    color;



        lrx = (rdp_reg.cmd0 & 0x00ffc000) >> 14;
        lry = (rdp_reg.cmd0 & 0x00000ffc) >> 2;
        ulx = (rdp_reg.cmd1 & 0x00ffc000) >> 14;
        uly = (rdp_reg.cmd1 & 0x00000ffc) >> 2;


        
        PRINT_RDP_INFO("FILLRECT ")
        PRINT_RDP_MNEMONIC("	ulx=%lu; uly=%lu; lrx=%lu; lry=%lu\n",
                           ulx,
                           uly,
                           lrx,
                           lry);
        PRINT_RDP_WARNING("modes are ignored")


    /* modes: copy mode is not used (not allowed?) */
        if(!(rdp_reg.mode_h & 0x00200000))
        {
            /* if we are not in 1 or 2 cycle mode (not in copy or fill mode) */
            /* we must not draw the bottom and right edges                   */
                lrx--;
                lry--;
                color = rdp_reg.blendcolor;
        }
        else
        {
                color = rdp_reg.fillcolor;
        }

        SetVisualColor(color);

        DrawVisualRectangle(ulx, uly, lrx, lry);


} /* static void rdp_fillrect() */





static void rdp_setfillcolor()
{
        PRINT_RDP_INFO("SETFILLCOLOR ")
        PRINT_RDP_MNEMONIC("	color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.fillcolor = rdp_reg.cmd1;

} /* static void rdp_setfillcolor() */





static void rdp_setfogcolor()
{
        PRINT_RDP_INFO("SETFOGCOLOR ")
        PRINT_RDP_MNEMONIC("	color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.fogcolor = rdp_reg.cmd1;

} /* static void rdp_setfogcolor() */





static void rdp_setblendcolor()
{
        PRINT_RDP_INFO("SETBLENDCOLOR ")
        PRINT_RDP_MNEMONIC("	color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.blendcolor = rdp_reg.cmd1;

} /* static void rdp_setblendcolor() */





static void rdp_setprimcolor()
{
        PRINT_RDP_INFO("SETPRIMCOLOR ")
        PRINT_RDP_MNEMONIC("	color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.primcolor = rdp_reg.cmd1;

} /* static void rdp_setprimcolor() */





static void rdp_setenvcolor()
{
        PRINT_RDP_INFO("SETENVCOLOR ")
        PRINT_RDP_MNEMONIC("	color=$%08lx\n", rdp_reg.cmd1);

        rdp_reg.envcolor = rdp_reg.cmd1;

} /* static void rdp_setenvcolor() */





static void rdp_setcombine()
{
        PRINT_RDP_INFO("SETCOMBINE - ignored")
        PRINT_RDP_MNEMONIC("	mode0=$%06lx; mode1=$%08lx\n",
               (rdp_reg.cmd0 & 0x00ffffff),
               (rdp_reg.cmd1 & 0xffffffff));
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_setcombine() */





static void rdp_settextureimage()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };


        rdp_reg.TextureImage.fmt	  = (uBYTE)((rdp_reg.cmd0 >> 21) & 0x07);
        rdp_reg.TextureImage.size	  = (uBYTE)((rdp_reg.cmd0 >> 19) & 0x03);
        rdp_reg.TextureImage.width	  = (uHWORD)(1 + (rdp_reg.cmd0 & 0x00000fff));
        rdp_reg.TextureImage.addr	  = segoffset2addr(rdp_reg.cmd1) & 0x003fffff;
		rdp_reg.TextureImage.vaddr	  = rdp_reg.cmd0;
		rdp_reg.TextureImage.ID		  = 0;


//		rdp_reg.TextureImage.LoadAddr = &mem.rd_ram[rdp_reg.TextureImage.addr];


        PRINT_RDP_INFO("SETTEXTUREIMAGE ")
/*        PRINT_RDP_MNEMONIC("	format=%s; size=%s; width=%hu; addr=$%08lx\n",
               format[rdp_reg.textureimg_fmt],
               size[rdp_reg.textureimg_size],
               rdp_reg.textureimg_width,
               rdp_reg.textureimg_addr);*/

} /* static void rdp_settextureimage() */





static void rdp_setdepthimage()
{
        PRINT_RDP_INFO("SETDEPTHIMAGE - ignored")
        PRINT_RDP_MNEMONIC("	addr=$%lx\n", rdp_reg.cmd1);
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_setdepthimage() */





static void rdp_setcolorimage()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };



        rdp_reg.colorimg_fmt   = (uBYTE)((rdp_reg.cmd0 & 0x00e00000) >> 21);
        rdp_reg.colorimg_size  = (uBYTE)((rdp_reg.cmd0 & 0x00180000) >> 19);
        rdp_reg.colorimg_width = (uHWORD)(1 + (rdp_reg.cmd0 & 0x00000fff));
        rdp_reg.colorimg_addr  = rdp_reg.cmd1;

        PRINT_RDP_INFO("SETCOLORIMAGE ")
        PRINT_RDP_MNEMONIC("	format=%s; size=%s; width=%hu; addr=$%lx\n",
                           format[rdp_reg.colorimg_fmt],
                           size[rdp_reg.colorimg_size],
                           rdp_reg.colorimg_width,
                           rdp_reg.colorimg_addr);

//  	   FlushVisualRenderBuffer();
//         SetVisualRenderBuffer(
//                        rdp_reg.colorimg_fmt, rdp_reg.colorimg_size,
//                        rdp_reg.colorimg_width, rdp_reg.colorimg_addr + (uBYTE *)mem.rd_ram);


} /* static void rdp_setcolorimage() */





static void rdp_trifill()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILL")

        CPU_HALT

} /* static void rdp_trifill() */





static void rdp_trishade()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADE")

        CPU_HALT

} /* static void rdp_trishade() */





static void rdp_tritxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTR")

        CPU_HALT

} /* static void rdp_tritxtr() */





static void rdp_trishadetxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTR")

        CPU_HALT

} /* static void rdp_trishadetxtr() */





static void rdp_trifillz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILLZ")

        CPU_HALT

} /* static void rdp_trifillz() */





static void rdp_trishadez()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADEZ")

        CPU_HALT

} /* static void rdp_trishadez() */





static void rdp_tritxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTRZ")

        CPU_HALT

} /* static void rdp_tritxtrz() */





static void rdp_trishadetxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTRZ")

        CPU_HALT

} /* static void rdp_trishadetxtrz() */










/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - static audio routines                    *
*                                                                              *
\******************************************************************************/



static void rdp_adpcm()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_ADPCM")

} /* static void rdp_adpcm() */





static void rdp_clearbuff()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_CLEARBUFF")

} /* static void rdp_clearbuff() */





static void rdp_envmixer()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_ENVMIXER")

} /* static void rdp_envmixer() */





static void rdp_loadbuff()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_LOADBUFF")

} /* static void rdp_loadbuff() */





static void rdp_resample()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_RESAMPLE")

} /* static void rdp_resample() */





static void rdp_savebuff()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_SAVEBUFF")

} /* static void rdp_savebuff() */





static void rdp_segment()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_SEGMENT")

} /* static void rdp_segment() */





static void rdp_setbuff()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_SETBUFF")

} /* static void rdp_setbuff() */





static void rdp_setvol()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_SETVOL")

} /* static void rdp_setvol() */





static void rdp_dmemmove()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_DMEMMOVE")

} /* static void rdp_dmemmove() */





static void rdp_loadadpcm()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_LOADADPCM")

} /* static void rdp_loadadpcm() */





static void rdp_mixer()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_MIXER")

} /* static void rdp_mixer() */





static void rdp_interleave()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_INTERLEAVE")

} /* static void rdp_interleave() */





static void rdp_polef()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_POLEF")

} /* static void rdp_polef() */





static void rdp_setloop()
{
        PRINT_RDP_AUDIO_NOT_IMPLEMENTED("RDP_SETLOOP")

} /* static void rdp_setloop() */






