#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <memory_extern.h>
#include "rdp_registers.h"

#include <stdio.h>
#include <stdlib.h>


#define CACHE_TEXTURES
#define MAX_CACHED_TEXTURES 4096

/******************************************************************************\
*                                                                              *
*   helper routines									  						   *
*                                                                              *
\******************************************************************************/

typedef struct
{
	uDWORD ID1;
	uDWORD ID2;
} t_cached_texture;

static t_cached_texture cached_texture[MAX_CACHED_TEXTURES];
static uWORD num_cached_texture = 0;
static uDWORD cur_ID1, cur_ID2;

//** I think the tbuffers are to large... we only
//** need 4096 (TMEM) * 4 (RGBA) * 2 (MAX 2Pixel pro Byte example given IA4) (FiRES)
static unsigned char tbuffer1[256*256*4];
static unsigned char tbuffer2[256*256*4];

void SetCurTexture(uDWORD ID1, uDWORD ID2)
{
	cur_ID1 = ID1;
	cur_ID2 = ID2;
}

int TestPower2(int width, int height, int *nw, int *nh)
{
	int w;
	int temp, NewWidth = *nw, NewHeight = *nh;
	
//** test if width has the power of 2
	temp = width;
	for (w =11; w>= 0; w--)
	{
		if(temp&0x800)
		{
			NewWidth = (1<<w);
			break;
		}
		temp<<=1;
	}
//** test if height has the power of 2
	temp = height;
	for (w =11; w>= 0; w--)
	{
		if(temp&0x800)
		{
			NewHeight = (1<<w);
			break;
		}
		temp<<=1;
	}
//** return
	if ((NewWidth == width) && (NewHeight == height)) return 1;
	else
	{
		*nw = NewWidth;
		*nh = NewHeight;
		return 0;
	}
	return 1;
}


void CopyTextureRGBA32(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureRGBA16(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureIA4   (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureIA8   (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureIA16  (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureI     (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureCI4   (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureCI8   (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);
void CopyTextureYUV   (uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height);

void writetex()
{
/*	int i, j = 0;
	FILE *stream = fopen("texture.raw","wb");
	for (i = 0; i < rdp_reg.td[0].Width*rdp_reg.td[0].Height; i++)
	{
		fprintf(stream,"%c",tbuffer1[j]); j++;
		fprintf(stream,"%c",tbuffer1[j]); j++;
		fprintf(stream,"%c",tbuffer1[j]); j++;
		j++;
	}
	fclose(stream);	
*/
}


int DecompressTexture(t_tile *tmpTile)
{
	unsigned char *src  = &mem.rd_ram[rdp_reg.TextureImage.addr];
	unsigned char *dest = (unsigned char*)tbuffer1;

	uBYTE size   = rdp_reg.TextureImage.size;
	uBYTE format = rdp_reg.TextureImage.fmt;

	switch(format)
	{
	case 0:	//rgba
		{
			if(size==3)
			{
				CopyTextureRGBA32(dest, src, tmpTile->Width, tmpTile->Height);
			}
			else
			{
				CopyTextureRGBA16(dest, src, tmpTile->Width, tmpTile->Height);
			}
			break;
		}
	case 1:	
		{
			//**not be implemented
			CopyTextureYUV(dest, src, tmpTile->Width, tmpTile->Height);
			break;
		}
	case 2:	//CI
		{
			if(size==0)
			{
				//**not be implemented
				CopyTextureCI4(dest, src, tmpTile->Width, tmpTile->Height);
			}
			else
			{
				CopyTextureCI8(dest, src, tmpTile->Width, tmpTile->Height);
			}
			break;
		}
	case 3:	//IA
		{
			switch(size)
			{
			case 0:
				{
					CopyTextureIA4(dest, src, tmpTile->Width, tmpTile->Height);
					break;
				}
			case 1:
				{
					CopyTextureIA8(dest, src, tmpTile->Width, tmpTile->Height);
					break;
				}
			case 2:
				{
					CopyTextureIA16(dest, src, tmpTile->Width, tmpTile->Height);
					break;
				}
			}
			break;
		}
	case 4:	//I
		{
			//**not be implemented
			CopyTextureI(dest, src, tmpTile->Width, tmpTile->Height);
			break;
		}
	default:
		{
			CopyTextureRGBA32(dest, src, tmpTile->Width, tmpTile->Height);
		}
	}

//	writetex();
	return 0;
}


int SetActiveTexture()
{
	uWORD i, new_width, new_height;

//** Use m_CurTile now (FiRES)  //**	t_tile *t_p = &rdp_reg.td[0];
	t_tile *t_p = rdp_reg.m_CurTile;

	uDWORD addr = rdp_reg.TextureImage.addr;
	uBYTE  size = rdp_reg.TextureImage.size;
	
#ifdef CACHE_TEXTURES
	for (i=0; i < num_cached_texture; i++)
		if ((cached_texture[i].ID1 == cur_ID1) && 
            (cached_texture[i].ID2 == cur_ID2))
			{				
				glBindTexture(GL_TEXTURE_2D, i);
				return 1;
			}
#endif

//** texture is not cached yet... so we cache it
	glBindTexture(GL_TEXTURE_2D, num_cached_texture);
	cached_texture[num_cached_texture].ID1 = cur_ID1;
    cached_texture[num_cached_texture].ID2 = cur_ID2;

	DecompressTexture(t_p);

	if (t_p->clampt) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	if (t_p->clamps) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

//	if (rdp_reg.geometrymode & 0x00080000)
//	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
//	}
//	else
//	{
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
//** Build the texture...	
	if (TestPower2(t_p->Width, t_p->Height, &new_width, &new_height))
	{
		glTexImage2D(GL_TEXTURE_2D, 0, 4, t_p->Width, t_p->Height, 0, GL_RGBA,
				 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer1);
	}
	else
	{
		gluScaleImage(GL_RGBA, t_p->Width, t_p->Height, GL_UNSIGNED_BYTE, (unsigned char*)tbuffer1,
					   new_width, new_height, GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2);

		glTexImage2D(GL_TEXTURE_2D, 0, 4, new_width, new_height, 0, GL_RGBA,
					 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2);
	}

	glDisable(GL_BLEND); // is needed for a lot of demos (FiRES)

#ifdef CACHE_TEXTURES
	num_cached_texture++;
	if (num_cached_texture >= MAX_CACHED_TEXTURES) num_cached_texture=0;
#endif

	return 1;
}
/******************************************************************************\
*                                                                              *
*   3d RenderStuff for RDP (OPENGL - helper routines)  						   *
*                                                                              *
\******************************************************************************/
void ClearVisual(void)
{
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

}

extern int Textures;
void DrawVisualTriangle(int vn[3])
{
	int i;
	if (1==1) //if (Textures == 0) //** doesnt work yet
	{
		SetActiveTexture();
		glColor3f(1,1,1);
		glEnable(GL_TEXTURE_2D);
		glBegin(GL_TRIANGLES);
		for(i=0; i<3; i++)
		{
			t_vtx *vertex = &rdp_reg.vtx[vn[i]];
			glTexCoord2f(vertex->s, vertex->t);

			glVertex3f((float)vertex->x,
					   (float)vertex->y,
					   (float)vertex->z);
		}
		glEnd();
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_TRIANGLES);
		for(i=0; i<3; i++)
		{
			t_vtx *vertex = &rdp_reg.vtx[vn[i]];
			glColor4ub(vertex->r, vertex->g, vertex->b, vertex->a);
			glVertex3f((float)vertex->x, 
					   (float)vertex->y,
					   (float)vertex->z);
		}
		glEnd();		
	}
}

void DrawTexRectangle(float xh, float yh, 
					  float xl, float yl, float f_s, float f_t)
{
	uWORD s_width  = rdp_reg.colorimg_width, 
		  s_height = (s_width * 3) >> 2;

	int depth_test;

//*** draw it
	glGetIntegerv(GL_DEPTH_FUNC, &depth_test);
	glDepthFunc(GL_ALWAYS);
	

    glMatrixMode(GL_PROJECTION);
	glPushMatrix();

    glLoadIdentity();
	glOrtho(0, s_width, s_height, 0, -1, 1);

    glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

    glLoadIdentity();

//** Draw the Rectangle
	SetActiveTexture();
	glColor3f(1,1,1);		// Textures are in modulate-mode...
	glEnable(GL_TEXTURE_2D);

	glBegin(GL_QUADS);
		glTexCoord2f(1, 1); glVertex2f(xl, yl);
		glTexCoord2f(0, 1); glVertex2f(xh, yl);
		glTexCoord2f(0, 0); glVertex2f(xh, yh);
		glTexCoord2f(1, 0); glVertex2f(xl, yh);
	glEnd();
//**
	glPopMatrix();

    glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);

	glDepthFunc(depth_test);


}

void SetVisualClipRectangle(uWORD ulx, uWORD uly, uWORD lrx, uWORD lry)
{

}

void SetVisualRenderBuffer(void)
{

}

void SetVisualColor(uWORD Color)
{

}


void DrawVisualRectangle(uWORD xh, uWORD yh, uWORD xl, uWORD yl)
{
}

//********************************************************************************
void CopyTextureRGBA32(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uWORD *src = (uWORD*)srcSurf;
               
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
		{
            *(dest) = (unsigned char)(*src >> 24) & 0xff; dest++;
            *(dest) = (unsigned char)(*src >> 16) & 0xff; dest++;
            *(dest) = (unsigned char)(*src >>  8) & 0xff; dest++;
            *(dest) = (unsigned char)(*src >>  0) & 0xff; dest++;
            src++;
         }


}
// correct !!
void CopyTextureRGBA16(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uHWORD *src = (uHWORD*)srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < (width << 2); j++)
        {
			dest+=4;
			*(dest) = (((*src >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =  ((*src & 0x01) * 0xff); dest++;
			src++;
			dest-=8;
			*(dest) = (((*src >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =  ((*src & 0x01) * 0xff); dest++;
			src++;
			dest+=4;

         }
}
void CopyTextureIA4(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uHWORD *src = (uHWORD*)srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < (width << 1); j++)
        {
        	uBYTE intensity = (*src & 0xf0) | ((*src & 0xf0) >> 4);
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
 	  	    *(dest) = intensity; dest++;
 	  	    
 	  	    intensity = (*src & 0x0f) | ((*src & 0x0f) << 4);
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
 	  	    *(dest) = intensity; dest++;
			src++;
         }

}

void CopyTextureIA8(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uBYTE *src = srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {
			*(dest) = *src; dest++;
			*(dest) = *src; dest++;
			*(dest) = *src; dest++;
 	  	    *(dest) = *src; dest++;
			src++;

         }

}
void CopyTextureIA16(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uHWORD *src = (uHWORD*)srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {
        	uBYTE intensity = ((*src >> 8) & 0xff);
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
			*(dest) = intensity; dest++;
 	  	    *(dest) = (*src & 0xff); dest++;
			 src++;
         }

}

void CopyTextureI(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uBYTE *src = srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {
			*(dest) = *src; dest++;
			*(dest) = *src; dest++;
			*(dest) = *src; dest++;
 	  	    *(dest) = 0xA0; dest++; //Is ist okay - mhhh - perhaps (FiRES)
			src++;

         }
}

void CopyTextureCI8(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst, *src = srcSurf;
    uHWORD color;
	uHWORD *palette = (uHWORD*)(mem.rd_ram + rdp_reg.tlut_8_addr);
	uBYTE index;

    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {	
			index = *src ^0x3;
			color = palette[index^0x2];
			*(dest) = (((color >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((color >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((color >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =  ((color & 0x01) * 0xff); dest++;
			src++;
		}
}

//*****************************************************************************
//*** the follow functions are not implemented yet
void CopyTextureCI4(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uHWORD *src = (uHWORD*)srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {
			*(dest) = 100; dest++;
			*(dest) = 150; dest++;
			*(dest) = 200; dest++;
 	  	    *(dest) = 255; dest++;
			 src++;
         }

}

void CopyTextureYUV(uBYTE *dst, uBYTE *srcSurf, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = dst;
    uHWORD *src = (uHWORD*)(uWORD*)srcSurf;
                
    for(i=0; i < height; i++)
	    for(j=0; j < width; j++)
        {
			*(dest) = 100; dest++;
			*(dest) = 150; dest++;
			*(dest) = 200; dest++;
 	  	    *(dest) = 255; dest++;
			 src++;
         }

}
