/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <type_sizes.h>
#include <config.h>
#include "file_loader_extern.h"


#include "N64/memory.h"
#include "N64/cpu.h"
#include "N64/registers.h"
#include "N64/controller.h"

#include "file_loader.h"
#include "_dispatch.h"

#include "output/display/display.h"

extern DWORD GetDInputData();		// come from DInput.cpp


void display_refresh(void);
uWORD controller_get_status(uWORD controller);

//**************************************++++*****
void display_refresh(void)
{
	RefreshVisual();
} 

//**
uWORD controller_get_status(uWORD controller)
{
	return (GetDInputData());
} 


int dispatch(int command, void *value)
{
        switch(command)
        {
            case EMU_RESET:
                clear_n64_mem();
                cpu_reset();
                copy_code();
                display_refresh();
                break;


            case SHUTDOWN:
                free_n64_mem();
                free(rom.image);
                exit(0);
                
            case MEMORY_ALLOC:
                return(alloc_n64_mem());
                
            case MEMORY_CLEAR:
                clear_n64_mem();
                return(0);
                
            case MEMORY_FREE:
                free_n64_mem();
                return(0);

            case DISPLAY_REFRESH:
                display_refresh();
                return(0);
                
            case CONTROLLER_GET_STATUS:
				*((uWORD *)value) = controller_get_status(*((uWORD *)value));				
                return(0);
                
            case CONTROLLER_INIT:
				// Dont use beacause the prefs.plugged_controller is set direct (Emu_Init)
                return(0);       
       
            case CPU_RESET:
                cpu_reset();
                return(0);
        
            case CPU_STOP:
                reg.halt = 1;
                return(0);
        
            case AUDIO_OPEN:
            case AUDIO_CLOSE:
            case AUDIO_UPDATE:
                return(0);
            case AUDIO_READY:
                return 1; //return(AUDIO_DMA_READY);
        
        } /* switch(command) */



        return(0);   /* to make compiler happy */

} /* int dispatch(int command, void *value) */
