/*
		draw2d(..) hab ich "erstmal" entfernt
*/
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <type_sizes.h>
#include <config.h>
#include <memory_extern.h>


// #include "draw2d_sw.h"
#include "rsp.h"
#include "rdp.h"
#include "rsp_registers.h"
#include "rdp_registers.h"
#include "mnemonic.h"
#include "registers.h"
#include "memory.h"




struct rdp_tile_descriptor
{
    /* SETTILESIZE / LOADBLOCK*/
        uHWORD   uls; /* Upper Left Texture X-Coordinate (12 bits) */
        uHWORD   ult; /* Upper Left Texture Y-Coordinate (12 bits) */
        uHWORD   lrs; /* Lower Right Texture X-Coordinate (12 bits) */
        uHWORD   lrt; /* Lower Right Texture Y-Coordinate (12 bits) */
        uHWORD   dxt; /* Increment dxt per scanline */

    /* SETTILE */
       uBYTE    format;
       uBYTE    size;
        int     line;
        int     tmem;
       uBYTE    palette;
       uBIT     clampt;
       uBIT     mirrort;
       uBYTE    maskt;
       uBYTE    shiftt;
       uBIT     clamps;
       uBIT     mirrors;
       uBYTE    masks;
       uBYTE    shifts;
};


/* general */

static uWORD  cmd0, cmd1;
static uWORD  start = 0, stop = 0, curr = 0;

/* color register */
static uWORD  rdp_fillcolor;     /* memory interface: 32bit */
                                /* i dont know if that is 1 32bit or 2 16bit pixel(s) */
static uWORD  rdp_primcolor;     /* color combiner: rgba */
static uWORD  rdp_envcolor;      /* color combiner: rgba */
static uWORD  rdp_fogcolor;      /* blender: rgba */
static uWORD  rdp_blendcolor;    /* blender: rgba */

/* This is the frame buffer location to draw into */
/* Used by rdp_setcolorimage() */
static  uBYTE  rdp_colorimg_fmt;          /* Frame Format (3 bits) */
static  uBYTE  rdp_colorimg_size;         /* Bits per pixel (2 bits) */
static  uHWORD rdp_colorimg_width;        /* width of frame in pixels (12 bits) */
static  uWORD  rdp_colorimg_addr;         /* RDRAM pointer to Image (32 bits) */

/* This is the frame buffer location to draw into */
/* Used by rdp_settextureimage() */
static uBYTE  rdp_textureimg_fmt;        /* see above */
static uBYTE  rdp_textureimg_size;
static uHWORD rdp_textureimg_width;
static uWORD  rdp_textureimg_addr;

/* This is the location of an 8 bit TLUT */
/* Used by rdp_loadtlut() */
static uBYTE  rdp_tlut_8_fmt;   /* TLUT Format (3 bits) */
static uBYTE  rdp_tlut_8_size;  /* Bits per pixel (2 bits) */
static uWORD  rdp_tlut_8_addr;  /* RDRAM pointer to TLUT (32 bits) */

/* This is the location of an 4 bit TLUT */
/* Used by rdp_loadtlut() */
static uBYTE  rdp_tlut_4_fmt[16];   /* TLUT Format (3 bits) */
static uBYTE  rdp_tlut_4_size[16];  /* Bits per pixel (2 bits) */
static uWORD  rdp_tlut_4_addr[16];  /* RDRAM pointer to TLUT (32 bits) */

/* These are the tile descriptors */
static struct   rdp_tile_descriptor rdp_td[8];

/* 4k of internal texture mem */
static uBYTE     *tmem = NULL;





/**
*
* This might be part of some sort of High Level Emulation (HLE).
* HLE was first hacked out by Epsilon and RealityMan - the authors of UltraHLE.
*
**/

#define INVALID_TASK    -1
#define GFX_TASK         1
#define AUDIO_TASK       2

static  int rdp_task = INVALID_TASK;

struct  rdp_reg rdp_reg;

static struct{
        uWORD tc;
        uWORD sc;
        uWORD level;
        uWORD tile;
        uWORD on;
}
texture;

static struct
{
        uWORD  rgba;
        uWORD  rgba_copy;
        sBYTE dir[3];
        uBYTE  dummy;
}
light[8];

static  uWORD geometrymode = 0;
static  uWORD mode_h = 0;
static  uWORD mode_l = 0;
static  uWORD lights = 0;

static  uWORD moveword_seg[16] =
        { 0,0,0,0,
          0,0,0,0,
          0,0,0,0,
          0,0,0,0 };                    /* segment pointer */

static  sHWORD vp[8] =
        { 0,0,0,0,
          0,0,0,0 };                    /* viewport (scale, translate) */

static uWORD perspective_normal = 1;

static  float mtx_proj[4][4] =          /* projection matrix */
        { { 0, 0, 0, 0 },
          { 0, 0, 0, 0 },
          { 0, 0, 0, 0 },
          { 0, 0, 0, 0 } };

static  float mtx_model[10][4][4];      /* 10 modelview matrices */

typedef  struct
{
        sHWORD  x, y, z;
        uHWORD  flags;
        sHWORD  s, t;
        uBYTE   r, g, b, a;
}
t_vtx;
static t_vtx vtx[16];                                /* 16 vertices */

static  struct
{
        sWORD   nx;
        sWORD   ny;
        sWORD   px;
        sWORD   py;
}
clip;                                   /* clipping stuff */


void    rdp_dl_check();                 /* executes a displaylist @ address a */
static  uWORD segoffset2addr(uWORD so);   /* calcs the real addr out of seg and offset */
static  void matrix_mult(float r[4][4], float m1[4][4], float m2[4][4]);
static  void vertex_mult(float r[4], float v[4], float m[4][4]);


/**
*
* End of HLE stuff.
*
**/




/* rdp instructions */

static void rdp_fixme();
static void rdp_spnoop();
static void rdp_matrix();
static void rdp_reserved0();
static void rdp_movemem();
static void rdp_vertex();
static void rdp_reserved1();
static void rdp_displaylist();
static void rdp_reserved2();
static void rdp_reserved3();
static void rdp_sprite2d();
static void rdp_tri1();
static void rdp_culldl();
static void rdp_popmatrix();
static void rdp_moveword();
static void rdp_texture();
static void rdp_setothermode_h();
static void rdp_setothermode_l();
static void rdp_enddl();
static void rdp_setgeometrymode();
static void rdp_cleargeometrymode();
static void rdp_line3d();
static void rdp_rdphalf_1();
static void rdp_rdphalf_2();
static void rdp_rdphalf_cont();
static void rdp_noop();
static void rdp_texrect();
static void rdp_texrectflip();
static void rdp_loadsync();
static void rdp_pipesync();
static void rdp_tilesync();
static void rdp_fullsync();
static void rdp_setkeygb();
static void rdp_setkeyr();
static void rdp_setconvert();
static void rdp_setscissor();
static void rdp_setprimdepth();
static void rdp_setothermode();
static void rdp_loadtlut();
static void rdp_settilesize();
static void rdp_loadblock();
static void rdp_loadtile();
static void rdp_settile();
static void rdp_fillrect();
static void rdp_setfillcolor();
static void rdp_setfogcolor();
static void rdp_setblendcolor();
static void rdp_setprimcolor();
static void rdp_setenvcolor();
static void rdp_setcombine();
static void rdp_settextureimage();
static void rdp_setdepthimage();
static void rdp_setcolorimage();
static void rdp_trifill();
static void rdp_trishade();
static void rdp_tritxtr();
static void rdp_trishadetxtr();
static void rdp_trifillz();
static void rdp_trishadez();
static void rdp_tritxtrz();
static void rdp_trishadetxtrz();





/* rdp instruction function pointer */

static rdp_instr instruction[256] =
{
    /**
    * DMA commands:
    * The next 64 commands (just 10 are used) are not executed by the RDP.
    * They are "emulated" by the RSP!
    * But in the display lists they appear with these bit combinations.
    * We will need this if we execute the display lists directly in future.
    * These commands include a pointer to mem in rdmem which is loaded via DMA.
    **/
        rdp_spnoop,             rdp_matrix,             rdp_reserved0,          rdp_movemem,
        rdp_vertex,             rdp_reserved1,          rdp_displaylist,        rdp_reserved2,
        rdp_reserved3,          rdp_sprite2d,           rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
    /**
    * The next 64 commands are not used (as far as i know).
    **/
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
    /**
    * IMMEDIATE commands:
    * The next 64 commands (just 14 are used) are not executed by the RDP.
    * They are like the commands above except that they do not carry a pointer.
    * All data they need is in the command itself.
    * So no DMA transfer is needed.
    **/
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_rdphalf_cont,       rdp_rdphalf_2,
        rdp_rdphalf_1,          rdp_line3d,             rdp_cleargeometrymode,  rdp_setgeometrymode,
        rdp_enddl,              rdp_setothermode_l,     rdp_setothermode_h,     rdp_texture,
        rdp_moveword,           rdp_popmatrix,          rdp_culldl,             rdp_tri1,
    /**
    * RDP commands:
    * These 64 commands are the real RDP commands.
    * I've never seen the TRI commands. They might also be generated by the RSP.
    **/
        rdp_noop,               rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_trifill,            rdp_trifillz,           rdp_tritxtr,            rdp_tritxtrz,
        rdp_trishade,           rdp_trishadez,          rdp_trishadetxtr,       rdp_trishadetxtrz,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_fixme,              rdp_fixme,              rdp_fixme,              rdp_fixme,
        rdp_texrect,            rdp_texrectflip,        rdp_loadsync,           rdp_pipesync,
        rdp_tilesync,           rdp_fullsync,           rdp_setkeygb,           rdp_setkeyr,
        rdp_setconvert,         rdp_setscissor,         rdp_setprimdepth,       rdp_setothermode,
        rdp_loadtlut,           rdp_fixme,              rdp_settilesize,        rdp_loadblock,
        rdp_loadtile,           rdp_settile,            rdp_fillrect,           rdp_setfillcolor,
        rdp_setfogcolor,        rdp_setblendcolor,      rdp_setprimcolor,       rdp_setenvcolor,
        rdp_setcombine,         rdp_settextureimage,    rdp_setdepthimage,      rdp_setcolorimage
};





/* define macros - to speed up things and to make things looking nice */


extern void msg_box(char *textl);

#define PRINT_RDP_NOT_IMPLEMENTED(name) // msg_box(name);
#define PRINT_RDP_WARNING(text)
#define PRINT_RDP_INFO(name)   // msg_box(name);
void    PRINT_RDP_MNEMONIC(char *fmt, ...)
{
}



/******************************************************************************\
*                                                                              *
*   GL Help-Functions - global routines										   *
*                                                                              *
\******************************************************************************/



// ** Begin a new Frame (clear COLOR and DEPTH Buffer)
void GL_BeginFrame(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

// ** Sets the DarwColor of OpenGl
void GL_SetColor(uWORD icolor)
{
		glColor4ub((icolor >> 24) & 0xff,
				   (icolor >> 16) & 0xff,
				   (icolor >>  8) & 0xff,
				   (icolor        & 0xff));
}

// ** Says the WindowsHandler, that he have to Flush the Screen 
int RePaint = 0;    // says the main that we have to flush OpenGl
void GL_FlushVisual(char *dst)
{
	FILE *tes; int i;
	tes = fopen ("bild", "wb");
	for (i=0; i< 4 * 1024; i++)
		fprintf(tes, "%c", tmem[i]);
	fclose(tes);
	RePaint = 1;
}


/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - global routines                          *
*                                                                              *
\******************************************************************************/




void rdp_go()		// ONLY FOR COMPATIBLE
{
	return; // WE USE HLE
} /* void rdp_go() */


void rdp_dis()
{

} /* void rdp_dis() */



void rdp_reset()
{
        if(tmem == NULL)
                tmem = (uBYTE *)malloc(4*1024);
        
        memset(tmem, 0, 4*1024);

        rdp_reg.pc_i = 0;
        rdp_reg.halt = 1;

} /* void rdp_reset() */



void rdp_step()
{
        int     i;
        uWORD    a;
	
        a = rdp_reg.pc[rdp_reg.pc_i];

        cmd0 = ((uWORD *)mem.rd_ram)[(a>>2)+0];
        cmd1 = ((uWORD *)mem.rd_ram)[(a>>2)+1];

        curr = ((uWORD *)mem.dpc_reg)[2] = \
        rdp_reg.pc[rdp_reg.pc_i] = a + 8;   /* point to next instruction */

        instruction[cmd0>>24]();

} /* void rdp_step() */





void rdp_dl_check()
{
        uWORD val;


// Which kind of Task
        switch(((uWORD *)mem.sp_dmem)[0xfc0>>2])
        {
            case 1:                             /* Graphic Task */
                rdp_task = GFX_TASK;
                break;

            case 2:                             /* Audio Task */
                rdp_task = AUDIO_TASK;
                PRINT_RDP_INFO("AUDIO TASK\n")
                break;

            default:
                rdp_task = INVALID_TASK;
                PRINT_RDP_INFO("NO KNOWN TASK\n")

        } /* switch(mem.sp_dmem[0xfc0>>2]) */

// Valid Task ??
        val = ((uWORD *)mem.sp_dmem)[0xfd4>>2];  /* ucode size */
        if(val != 0x1000)
        {
                rdp_task = INVALID_TASK;
                PRINT_RDP_INFO("UCODE TEST FAILED\n")
        }

        val = ((uWORD *)mem.sp_dmem)[0xfdc>>2];  /* ucode data size */
        if(val != 0x800)
        {
                rdp_task = INVALID_TASK;
                PRINT_RDP_INFO("UCODE DATA TEST FAILED\n")
        }

        val = ((uWORD *)mem.sp_dmem)[0xfe4>>2];  /* dram stack size */
        switch(rdp_task)
        {
            case GFX_TASK:
                if(val < 640)   // 0x280 = 640
                {
                        rdp_task = INVALID_TASK;
                        PRINT_RDP_INFO("UCODE DRAM STACK TEST FAILED\n")
                }
                break;

            case AUDIO_TASK:
                if(val != 0)
                {
                        rdp_task = INVALID_TASK;
                        PRINT_RDP_INFO("UCODE DRAM STACK TEST FAILED\n")
                }
                break;

        } /* switch(rdp_task) */



        switch(rdp_task)
        {
            case GFX_TASK:
                curr = ((uWORD *)mem.dpc_reg)[2] = val;
                rdp_reg.pc[0] = ((uWORD *)mem.sp_dmem)[0xff0>>2];
                rdp_reg.pc_i = 0;
                rdp_reg.halt = 0;
			// NEW FRAME BEGINNS
				GL_BeginFrame();
                break;

            case AUDIO_TASK:
                PRINT_RDP_INFO("AUDIO TASK - WILL NOT RUN HLE - AUDIO IS NOT DONE (yet;-)\n")
                break;

            case INVALID_TASK:
            default:
                PRINT_RDP_INFO("WILL NOT RUN HLE\n");
        
        } /* switch(rdp_task) */

} /* void rdp_dl_check() */





/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - helper routines                          *
*                                                                              *
\******************************************************************************/


static uWORD segoffset2addr(uWORD so)
{
        return( moveword_seg[(so>>24)&0x0f] + (so & 0x00ffffff) );

} /* static uWORD segoffset2addr(uWORD so) */



/******************************************************************************\
*                                                                              *
*   RDP (Reality Display Processor) - static routines                          *
*                                                                              *
\******************************************************************************/



static void rdp_fixme()
{
        PRINT_RDP_NOT_IMPLEMENTED("FIXME")
} /* static void rdp_fixme() */





static void rdp_spnoop()
{
} /* static void rdp_spnoop() */




/**
*
* The following matrix stuff was first done by Jeff.
* Great work, Jeff! (Niki)
*
**/

static void rdp_matrix()
{

        uWORD    a = segoffset2addr(cmd1);
        uBYTE    command = (uBYTE)((cmd0 >> 16) & 0xff);
        uHWORD   len = (uHWORD)(cmd0 & 0xffff);
        uWORD    t2_i;
        uHWORD   t2_f;
        sWORD    tmp_i;
/* BUG: ?*/
/*
        sHWORD  tmp_f;
*/
/* it might be that we have to use the following (s15.s15 instead if s15.16): */
        uHWORD   tmp_f = 0;
/* Jeff? You should know! */
        float   m[4][4];
/*        uHWORD   n[8][4]; */
        int     i, j;



        for(i = 0; i < 4; i++)
                for(j = 0; j < 4; j++)
                {					
					uWORD address = (a>>1) + (i<<2) + j;

//					t2_i   =  ((uHWORD *)mem.rd_ram)[(address)^1];
//                  t2_f   =  ((uHWORD *)mem.rd_ram)[(address + 16)^1];

					tmp_i   = ((sWORD)((sHWORD *)mem.rd_ram)[(address )^1]) << 16;
					tmp_f   = ((sWORD)((sHWORD *)mem.rd_ram)[(address + 16)^1]);

					m[i][j] = (float)(tmp_i + tmp_f) / 0xffff;
/*
					n[i][j] = t2_i;
					n[i+4][j] = t2_f;
*/
                }
        switch(command)
        {
            case 0: /* modelview  mul  nopush */
				glMatrixMode(GL_MODELVIEW);
                glMultMatrixf(m);
                break;

            case 1: /* projection mul  nopush */
				glMatrixMode(GL_PROJECTION);
				glMultMatrixf(m);
				glMatrixMode(GL_MODELVIEW);
                break;

            case 2: /* modelview  load nopush */
				glMatrixMode(GL_MODELVIEW);
				glLoadMatrixf(m);
                break;

            case 3: /* projection load nopush */
				glMatrixMode(GL_PROJECTION);
				glLoadMatrixf(m);
				glMatrixMode(GL_MODELVIEW);
                break;

            case 4: /* modelview  mul  push   */
				glPushMatrix();				
				glMatrixMode(GL_MODELVIEW);
                glMultMatrixf(m);
                break;

            case 5: /* projection mul  push   */
				glPushMatrix();				
				glMatrixMode(GL_PROJECTION);
				glMultMatrixf(m);
				glMatrixMode(GL_MODELVIEW);
                break;

            case 6: /* modelview  load push   */
				glPushMatrix();				
				glMatrixMode(GL_MODELVIEW);
                glLoadMatrixf(m);
                break;

            case 7: /* projection load push   */
				glPushMatrix();				
				glMatrixMode(GL_PROJECTION);
				glLoadMatrixf(m);
				glMatrixMode(GL_MODELVIEW);
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MATRIX: UNKNOWN COMMAND")
				break;

        } /* switch(command) */


} /* static void rdp_matrix() */





static void rdp_reserved0()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED0")
} /* static void rdp_reserved0() */





static void rdp_movemem()
{

        uWORD a = segoffset2addr(cmd1);

        switch((cmd0 >> 16) & 0xff)
        {
            case 0x80:
                vp[0] = ((uHWORD *)mem.rd_ram)[((a>>1)+0)^1];
                vp[1] = ((uHWORD *)mem.rd_ram)[((a>>1)+1)^1];
                vp[2] = ((uHWORD *)mem.rd_ram)[((a>>1)+2)^1];
                vp[3] = ((uHWORD *)mem.rd_ram)[((a>>1)+3)^1];
                vp[4] = ((uHWORD *)mem.rd_ram)[((a>>1)+4)^1];
                vp[5] = ((uHWORD *)mem.rd_ram)[((a>>1)+5)^1];
                vp[6] = ((uHWORD *)mem.rd_ram)[((a>>1)+6)^1];
                vp[7] = ((uHWORD *)mem.rd_ram)[((a>>1)+7)^1];
                break;

            case 0x82:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM LOOKATY")
                break;

            case 0x84:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM LOOKATX")
                break;

            case 0x86:
                light[0].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[0].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[0].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[0].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[0].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[0].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x88:
                light[1].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[1].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[1].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[1].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[1].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[1].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x8a:
                light[2].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[2].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[2].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[2].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[2].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[2].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x8c:
                light[3].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[3].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[3].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[3].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[3].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[3].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x8e:
                light[4].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[4].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[4].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[4].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[4].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[4].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x90:
                light[5].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[5].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[5].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[5].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[5].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[5].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x92:
                light[6].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[6].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[6].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[6].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[6].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[6].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x94:
                light[7].rgba      = ((uWORD *)mem.rd_ram)[ (a>>2)+ 0   ];
                light[7].rgba_copy = ((uWORD *)mem.rd_ram)[ (a>>2)+ 1   ];
                light[7].dir[0]    = ((uBYTE *)mem.rd_ram)[( a    + 8)^3];
                light[7].dir[1]    = ((uBYTE *)mem.rd_ram)[( a    + 9)^3];
                light[7].dir[2]    = ((uBYTE *)mem.rd_ram)[( a    +10)^3];
                light[7].dummy     = ((uBYTE *)mem.rd_ram)[( a    +11)^3];
                break;

            case 0x98:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 0")
                break;

            case 0x9a:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 1")
                break;

            case 0x9c:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 2")
                break;

            case 0x9e:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM MATRIX 3")
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEMEM ?")
				break;

        } /* switch((cmd0 >> 8) & 0xffff) */


} /* static void rdp_movemem() */





static void rdp_vertex()
{

        uWORD    a = segoffset2addr(cmd1);
        int     v0 =  (cmd0 >> 16) & 0xf;
        int     n  = ((cmd0 >> 20) & 0xf) + 1;
        uHWORD len = cmd0 & 0xffff;
        int     i;

       for(i = v0; i < (v0 + n); i++)
        {
                vtx[i].x     = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 0)^1];
                vtx[i].y     = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 1)^1];
                vtx[i].z     = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 2)^1];
                vtx[i].flags = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 3)^1];
                vtx[i].s     = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 4)^1];
                vtx[i].t     = ((uHWORD *)mem.rd_ram)[(((a+i*16) >> 1) + 5)^1];
                vtx[i].r     = ((uBYTE  *)mem.rd_ram)[(  a+i*16        +12)^3];
                vtx[i].g     = ((uBYTE  *)mem.rd_ram)[(  a+i*16        +13)^3];
                vtx[i].b     = ((uBYTE  *)mem.rd_ram)[(  a+i*16        +14)^3];
                vtx[i].a     = ((uBYTE  *)mem.rd_ram)[(  a+i*16        +15)^3];
        }
} /* static void rdp_vertex() */





static void rdp_reserved1()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED1")
} /* static void rdp_reserved1() */





static void rdp_displaylist()
{

        uWORD addr = segoffset2addr(cmd1);

        rdp_reg.pc_i++;
        if(rdp_reg.pc_i > 9)
        {
                PRINT_RDP_WARNING("DL STACK OVERFLOW (too much (>9) DISPLAYLIST calls\n"
                                  "will not call dl!!! - ignored")
                return;
        }
        rdp_reg.pc[rdp_reg.pc_i] = addr;
} /* static void rdp_displaylist() */





static void rdp_reserved2()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED2")
} /* static void rdp_reserved2() */





static void rdp_reserved3()
{
        PRINT_RDP_NOT_IMPLEMENTED("RESERVED3")
} /* static void rdp_reserved3() */





static void rdp_sprite2d()
{

        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };

        uWORD    a               = segoffset2addr(cmd1);
        float   fscale_x; 
        float   fscale_y; 
        uWORD    dst_p;  /* dest pointer */
        uWORD    dst_w;  /* dest with (normally 320 or 640) */
        uWORD    dst_b;  /* dest depth (16bit: 2; 32bit: 3) */

        uWORD    src             = ((uWORD *) mem.rd_ram)[((a>>2)+ 0)^0];
        uWORD    src_ci          = ((uWORD *) mem.rd_ram)[((a>>2)+ 1)^0];
        uHWORD   src_w           = ((uHWORD *)mem.rd_ram)[((a>>1)+ 4)^1];
        uHWORD   sprite_w        = ((uHWORD *)mem.rd_ram)[((a>>1)+ 5)^1];
        uHWORD   sprite_h        = ((uHWORD *)mem.rd_ram)[((a>>1)+ 6)^1];
        uBYTE    src_type        = ((uBYTE *) mem.rd_ram)[( a    +14)^3];   /* rgba, ci, ai, a, i */
        uBYTE    src_bits        = ((uBYTE *) mem.rd_ram)[( a    +15)^3];   /* 4, 8, 16, 32 */
        uHWORD   scale_x         = ((uHWORD *)mem.rd_ram)[((a>>1)+ 8)^1];   /* s5.10 */
        uHWORD   scale_y         = ((uHWORD *)mem.rd_ram)[((a>>1)+ 9)^1];   /* s5.10 */
        uBYTE    flip_x          = ((uBYTE *) mem.rd_ram)[( a    +20)^3];
        uBYTE    flip_y          = ((uBYTE *) mem.rd_ram)[( a    +21)^3];
        uHWORD   src_ox          = ((uHWORD *)mem.rd_ram)[((a>>1)+11)^1];
        uHWORD   src_oy          = ((uHWORD *)mem.rd_ram)[((a>>1)+12)^1];
        uHWORD   dst_ox          = ((uHWORD *)mem.rd_ram)[((a>>1)+13)^1];
        uHWORD   dst_oy          = ((uHWORD *)mem.rd_ram)[((a>>1)+14)^1];

        fscale_x = ((signed)scale_x >> 10);   // BUG: fraction is missing!
        fscale_y = ((signed)scale_y >> 10);   // BUG: fraction is missing!

        dst_b = ((uWORD *)mem.vi_reg)[0] & 0x03;
        if(dst_b != 2 && dst_b != 3)
        {
                PRINT_RDP_NOT_IMPLEMENTED("SPRITE2D: dest screen has unsupported depth\n")
                return;
        }
        dst_w = ((uWORD *)mem.vi_reg)[2];
        dst_p = ((uWORD *)mem.vi_reg)[1];
        dst_p &= 0x003fffff;

/*
        draw2d(src, src_ci, src_type, src_bits, src_w, src_ox, src_oy,
                sprite_w, sprite_h, flip_x, flip_y,
                dst_p, 0, dst_b, dst_w, dst_ox, dst_oy );

*/
} /* static void rdp_sprite2d() */


static void rdp_tri1()
{
	int   i, vn[3], f;
    vn[0] = ((cmd1 >> 16) & 0xff) / 10;
    vn[1] = ((cmd1 >>  8) & 0xff) / 10;
    vn[2] =  (cmd1        & 0xff) / 10;
    f     =  (cmd1 >> 24) & 0xff;


// Draw without Texture
	if (texture.on == 0)
	{
		glBegin(GL_TRIANGLES);
			for(i=0; i<3; i++)
			{
				t_vtx *vertex = &vtx[vn[i]];
				glColor4ub(vertex->r, vertex->g, vertex->b, vertex->a );
				glVertex3f((float)vertex->x, 
						   (float)vertex->y,
						   (float)vertex->z);
			}
		glEnd();
	}
// Draw textured
	else
	{
		glBindTexture(GL_TEXTURE_2D,0);
//		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, 

		glBegin(GL_TRIANGLES);
			for(i=0; i<3; i++)
			{
				t_vtx *vertex = &vtx[vn[i]];
				glColor4ub(0xff, 0xff, 0xff, 0xff);
				glVertex3f((float)vertex->x, 
						   (float)vertex->y,
						   (float)vertex->z);
			}
		glEnd();
	}


} /* static void rdp_tri1() */





static void rdp_culldl()
{
        PRINT_RDP_NOT_IMPLEMENTED("CULLDL")
} /* static void rdp_culldl() */





static void rdp_popmatrix()
{
	glPopMatrix();
} /* static void rdp_popmatrix() */





static void rdp_moveword()
{
        switch(cmd0 & 0xff)
        {
            case 0x00:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD MATRIX")
                break;

            case 0x02: //MOVEWORD NUMLIGHT: 
                lights = (cmd1 - 0x80000000) / 32 - 1;
                break;

            case 0x04:
                PRINT_RDP_INFO("MOVEWORD CLIP: ")
                switch((cmd0 >> 8) & 0xffff)
                {
                    case 0x0004:
                        PRINT_RDP_NOT_IMPLEMENTED("-x = ");
                        clip.nx = cmd1;
                        break;

                    case 0x000c:
                        PRINT_RDP_NOT_IMPLEMENTED("-y = ");
                        clip.ny = cmd1;
                        break;

                    case 0x0014:
                        PRINT_RDP_NOT_IMPLEMENTED("+x = ");
                        clip.px = cmd1;
                        break;

                    case 0x001c:
                        PRINT_RDP_NOT_IMPLEMENTED("+y = ");
                        clip.py = cmd1;
                        break;

                    default:
                        PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD CLIP - wrong offset")
						break;
                } /* switch((cmd0 >> 8) & 0xffff) */
                
                PRINT_RDP_MNEMONIC("%d\n", (sWORD)(sHWORD)cmd1);
                break;

            case 0x06://MOVEWORD SEGMENT
                moveword_seg[(cmd0 >> 10) & 0xf] = cmd1;
                break;

            case 0x08:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD FOG")
                break;

            case 0x0a:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD LIGHTCOL")
                break;

            case 0x0c:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD POINTS")
                break;

            case 0x0e:
                PRINT_RDP_INFO("MOVEWORD PERSPNORM: ")
                perspective_normal = cmd1;
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("MOVEWORD ?")
				break;
        } /* switch(cmd0 & 0xff) */


} /* static void rdp_moveword() */





static void rdp_texture()
{

        texture.sc    = cmd1 >> 16;
        texture.tc    = cmd1 & 0xffff;
        texture.level = (cmd0 >> 11) & 0x07;
        texture.tile  = (cmd0 >> 8)  & 0x07;
        texture.on    = cmd0 & 0xff;

		if (texture.on != 0) 
			glEnable(GL_TEXTURE_2D);
		else 
			glDisable(GL_TEXTURE_2D);
} /* static void rdp_texture() */





static void rdp_setothermode_h()
{
        static char *ad[] = { "pattern", "notpattern", "noise", "disable" };
        static char *rd[] = { "magicsq", "bayer", "noise", "?" };
        static char *ck[] = { "none", "key" };
        static char *tc[] = { "conv", "?", "?", "?", "?", "filtconv", "filt", "?" };
        static char *tf[] = { "point", "?", "bilerp", "average" };
        static char *tt[] = { "none", "?", "rgba16", "ia16" };
        static char *tl[] = { "tile", "lod" };
        static char *td[] = { "clamp", "sharpen", "detail", "?" };
        static char *tp[] = { "none", "persp" };
        static char *ct[] = { "1cycle", "2cycle", "copy", "fill" };
        static char *cd[] = { "disable(hw>1)", "enable(hw>1)", "disable(hw1)", "enable(hw1)" };
        static char *pm[] = { "nprimitive", "1primitive" };

        switch((cmd0 >> 8) & 0xff)
        {
            case 0x00:
                PRINT_RDP_WARNING("SETOTHERMODE_H BLENDMASK - ignored")
                break;

            case 0x04: // SETOTHERMODE_H ALPHADITHER
                mode_h &= ~0x00000030;
                cmd1   &=  0x00000030;
                mode_h |=  cmd1;
                break;

            case 0x06: // SETOTHERMODE_H RGBDITHER
                mode_h &= ~0x000000c0;
                cmd1   &=  0x000000c0;
                mode_h |=  cmd1;
                break;

            case 0x08: // SETOTHERMODE_H COMBINEKEY
                mode_h &= ~0x00000100;
                cmd1   &=  0x00000100;
                mode_h |=  cmd1;
                break;

            case 0x09: // SETOTHERMODE_H TEXTURECONVERT
                mode_h &= ~0x00000e00;
                cmd1   &=  0x00000e00;
                mode_h |=  cmd1;
                break;

            case 0x0c: // SETOTHERMODE_H TEXTUREFILTER
                mode_h &= ~0x00003000;
                cmd1   &=  0x00003000;
                mode_h |=  cmd1;
                break;

            case 0x0e: // SETOTHERMODE_H TEXTURELUT
                mode_h &= ~0x0000c000;
                cmd1   &=  0x0000c000;
                mode_h |=  cmd1;
                break;

            case 0x10: // SETOTHERMODE_H TEXTURELOD
                mode_h &= ~0x00010000;
                cmd1   &=  0x00010000;
                mode_h |=  cmd1;
                break;

            case 0x11: // SETOTHERMODE_H TEXTUREDETAIL
                mode_h &= ~0x00060000;
                cmd1   &=  0x00060000;
                mode_h |=  cmd1;
                break;

            case 0x13: // SETOTHERMODE_H TEXTUREPERSP
                mode_h &= ~0x00080000;
                cmd1   &=  0x00080000;
                mode_h |=  cmd1;
                break;

            case 0x14: // SETOTHERMODE_H CYCLETYPE
                mode_h &= ~0x00300000;
                cmd1   &=  0x00300000;
                mode_h |=  cmd1;
                break;

            case 0x16: // SETOTHERMODE_H COLORDITHER
                mode_h &= ~0x00400000;
                cmd1   &=  0x00400000;
                mode_h |=  cmd1;
                break;

            case 0x17: // SETOTHERMODE_H PIPELINEMODE
                mode_h &= ~0x00800000;
                cmd1   &=  0x00800000;
                mode_h |=  cmd1;
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_H ?")
				break;
        } /* switch((cmd0 >> 8) & 0xff) */

} /* static void rdp_setothermode_h() */





static void rdp_setothermode_l()
{
        static char *ac[] = { "none", "threshold", "?", "diter" };
        static char *zs[] = { "pixel", "prim" };
        static char *a1[] =
                {
                        "        bl_1ma (1)",
                        "        bl_a_mem (1)",
                        "        bl_1 (1)",
                        "        bl_0 (1)"
                };
        static char *b1[] =
                {
                        "        bl_clr_in (1)",
                        "        bl_clr_mem (1)",
                        "        bl_clr_bl (1)",
                        "        bl_clr_fog (1)"
                };
        static char *c1[] =
                {
                        "        bl_a_in (1)",
                        "        bl_a_fog (1)",
                        "        bl_a_shade (1)",
                        "        bl_0 (1)"
                };
        static char *d1[] =
                {
                        "        bl_1ma (1)",
                        "        bl_a_mem (1)",
                        "        bl_1 (1)",
                        "        bl_0 (1)" 
                };
        static char *a2[] =
                {
                        "        bl_1ma (2)",
                        "        bl_a_mem (2)",
                        "        bl_1 (2)",
                        "        bl_0 (2)"
                };
        static char *b2[] =
                {
                        "        bl_clr_in (2)",
                        "        bl_clr_mem (2)",
                        "        bl_clr_bl (2)",
                        "        bl_clr_fog (2)"
                };
        static char *c2[] =
                {
                        "        bl_a_in (2)",
                        "        bl_a_fog (2)",
                        "        bl_a_shade (2)",
                        "        bl_0 (2)"
                };
        static char *d2[] =
                {
                        "        bl_1ma (2)",
                        "        bl_a_mem (2)",
                        "        bl_1 (2)",
                        "        bl_0 (2)" 
                };

        switch((cmd0 >> 8) & 0xff)
        {
            case 0x00: // SETOTHERMODE_L ALPHACOMPARE
                mode_l &= ~0x00000003;
                cmd1   &=  0x00000003;
                mode_l |=  cmd1;
                break;

            case 0x02: // SETOTHERMODE_L ZSRCSEL
                mode_l &= ~0x00000004;
                cmd1   &=  0x00000004;
                mode_l |=  cmd1;
                break;

            case 0x03:
                PRINT_RDP_MNEMONIC("$%08lx:\n", cmd1 & 0xfffffff8,
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n"
                        "%s\n",
                        cmd1 & 0xfffffff8,
                        (cmd1 & 0x00000008) ? "        anti alias\n" : "",
                        (cmd1 & 0x00000010) ? "        z_cmp\n" : "",
                        (cmd1 & 0x00000020) ? "        z_upd\n" : "",
                        (cmd1 & 0x00000040) ? "        im_rd\n" : "",
                        (cmd1 & 0x00000080) ? "        clr_on_cvg\n" : "",
                        (cmd1 & 0x00000100) ? "        cvg_dst_warp\n" : "",
                        (cmd1 & 0x00000200) ? "        cvg_dst_full\n" : "",
                        (cmd1 & 0x00000400) ? "        z_inter\n" : "",
                        (cmd1 & 0x00000800) ? "        z_xlu\n" : "",
                        (cmd1 & 0x00001000) ? "        cvg_x_alpha\n" : "",
                        (cmd1 & 0x00002000) ? "        alpha_cvg_sel\n" : "",
                        (cmd1 & 0x00004000) ? "        force_bl\n" : "",
                        (cmd1 & 0x00008000) ? "        tex_edge?\n" : "",
                        a2[(cmd1>>16) & 0x3],
                        a1[(cmd1>>18) & 0x3],
                        b2[(cmd1>>20) & 0x3],
                        b1[(cmd1>>22) & 0x3],
                        c2[(cmd1>>24) & 0x3],
                        c1[(cmd1>>26) & 0x3],
                        d2[(cmd1>>28) & 0x3],
                        d1[(cmd1>>30) & 0x3]
                        );

                mode_l &= ~0xfffffff8;
                cmd1   &=  0xfffffff8;
                mode_l |=  cmd1;
                break;

            case 0x16:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_L BLENDER")
                break;

            default:
                PRINT_RDP_NOT_IMPLEMENTED("SETOTHERMODE_L ?")
				break;

        } /* switch((cmd0 >> 8) & 0xff) */

} /* static void rdp_setothermode_l() */





static void rdp_enddl()
{

        if(rdp_reg.pc_i < 0)
        {
                PRINT_RDP_WARNING("DL STACK UNDERRUN (too much ENDDL calls)\n"
                                  "will not return from dl!!! - ignored")
                return;
        }

        if(rdp_reg.pc_i == 0)
        {
                PRINT_RDP_MNEMONIC("RDP stops here - RSP BREAK will be simulated\n");
                GL_FlushVisual(rdp_colorimg_addr + mem.rd_ram);

                rsp_reg.code = 0x0000000d;
                rsp_dis(0x0000000d);   /* RSP BREAK instruction */
                rdp_reg.halt = 1;
        }

        rdp_reg.pc_i--;

} /* static void rdp_enddl() */





static void rdp_setgeometrymode()
{
        PRINT_RDP_INFO("SETGEOMETRYMODE ")
        PRINT_RDP_MNEMONIC(
                "+$%08lx:\n"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s"
                "%s",
                cmd1,
                (cmd1 & 0x00000001) ? "        zbuffer\n" : "",
                (cmd1 & 0x00000002) ? "        texture\n" : "",
                (cmd1 & 0x00000004) ? "        shade\n" : "",
                (cmd1 & 0x00000200) ? "        shade smooth\n" : "",
                (cmd1 & 0x00001000) ? "        cull front\n" : "",
                (cmd1 & 0x00002000) ? "        cull back\n" : "",
                (cmd1 & 0x00010000) ? "        fog\n" : "",
                (cmd1 & 0x00020000) ? "        lightning\n" : "",
                (cmd1 & 0x00040000) ? "        texture gen\n" : "",
                (cmd1 & 0x00080000) ? "        texture gen lin\n" : "",
                (cmd1 & 0x00100000) ? "        lod\n" : ""
                );

        geometrymode |= cmd1;


} /* static void rdp_setgeometrymode() */





static void rdp_cleargeometrymode()
{
/*              (cmd1 & 0x00000001) ? "        zbuffer\n" : "",
                (cmd1 & 0x00000002) ? "        texture\n" : "",
                (cmd1 & 0x00000004) ? "        shade\n" : "",
                (cmd1 & 0x00000200) ? "        shade smooth\n" : "",
                (cmd1 & 0x00001000) ? "        cull front\n" : "",
                (cmd1 & 0x00002000) ? "        cull back\n" : "",
                (cmd1 & 0x00010000) ? "        fog\n" : "",
                (cmd1 & 0x00020000) ? "        lightning\n" : "",
                (cmd1 & 0x00040000) ? "        texture gen\n" : "",
                (cmd1 & 0x00080000) ? "        texture gen lin\n" : "",
                (cmd1 & 0x00100000) ? "        lod\n" : ""
*/
        geometrymode &= ~cmd1;

} /* static void rdp_cleargeometrymode() */





static void rdp_line3d()
{
        PRINT_RDP_NOT_IMPLEMENTED("LINE3D")

} /* static void rdp_line3d() */





static void rdp_rdphalf_1()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_1")
} /* static void rdp_rdphalf_1() */





static void rdp_rdphalf_2()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_2")
} /* static void rdp_rdphalf_2() */





static void rdp_rdphalf_cont()
{
        PRINT_RDP_NOT_IMPLEMENTED("RDPHALF_CONT")
} /* static void rdp_rdphalf_cont() */





static void rdp_noop()
{

} /* static void rdp_noop() */





static void rdp_texrect()
{
        uWORD    a = rdp_reg.pc[rdp_reg.pc_i];
        uWORD    cmd2 = ((uWORD *)mem.rd_ram)[(a>>2)+1];
        uWORD    cmd3 = ((uWORD *)mem.rd_ram)[(a>>2)+3];

        uHWORD   xh     = (cmd0 & 0x00fff000) >> 14;
        uHWORD   yh     = (cmd0 & 0x00000fff) >> 2;
        uHWORD   xl     = (cmd1 & 0x00fff000) >> 14;
        uHWORD   yl     = (cmd1 & 0x00000fff) >> 2;
        int     tile   = (cmd1 & 0x07000000) >> 24;
/*
        uHWORD   s      = (cmd2 & 0xffff0000) >> 16;
        uHWORD   t      = cmd2 & 0x0000ffff;
        float   f_s    = (float)((s & 0xffe0) + (s & 0x001f) / 31.f);
        float   f_t    = (float)((t & 0xffe0) + (t & 0x001f) / 31.f);
*/
        float   f_s    = ((float)((cmd2 & 0xffff0000) >> 16) / 31.f);
        float   f_t    = ((float)(cmd2 & 0x0000ffff) / 31.f);                  /* is this correct jeff? */
/*
        sHWORD  dsdx   = (cmd3 & 0xffff0000) >> 16;
        sHWORD  dsdy   = cmd3 & 0x0000ffff;
        float   f_dsdx = (float)((dsdx & 0xfc00) + (dsdx & 0x02ff) / 1023.f);
        float   f_dsdy = (float)((dsdy & 0xfc00) + (dsdy & 0x02ff) / 1023.f);
*/
        float   f_dsdx = (((float)((cmd3 & 0xffff0000) >> 16)) / 1023.f);
        float   f_dsdy = (((float)(cmd3 & 0x0000ffff)) / 1023.f);



        PRINT_RDP_INFO("TEXRECT ")
        PRINT_RDP_MNEMONIC("xh=%4hu; yh=%4hu; tile=%u; xl=%4hu; yl=%4hu; s=%+9.4f; t=%+9.4f; dsdx=%+7.4f; dtdy=%+7.4f\n",
               xh, yh, tile, xl, yl, f_s, f_t, f_dsdx, f_dsdy);
   //     printf("xh=%4hu; yh=%4hu; tile=%u; xl=%4hu; yl=%4hu; s=%+9.4f; t=%+9.4f; dsdx=%+7.4f; dtdy=%+7.4f\n",
   //            xh, yh, tile, xl, yl, f_s, f_t, f_dsdx, f_dsdy);
        PRINT_RDP_WARNING("some parameter are ignored")



//        draw2d(rdp_textureimg_addr, rdp_tlut_8_addr, rdp_td[tile].format, rdp_td[tile].size,
                //(uHWORD)((((float)(xh - xl)))*(f_dsdx)), (uHWORD)f_s, (uHWORD)f_t,
//                xh - xl, 0, 0,
//                xh - xl, yh - yl, 0, 0,
//               rdp_colorimg_addr, rdp_colorimg_fmt, rdp_colorimg_size, rdp_colorimg_width, xl, yl);

    /* so rdp_rdphalf_1 and rdp_rdphalf_2 are never reached */
        curr = ((uWORD *)mem.dpc_reg)[2] = a + 16;
        rdp_reg.pc[rdp_reg.pc_i] += 16;

} /* static void rdp_texrect() */





static void rdp_texrectflip()
{
        PRINT_RDP_NOT_IMPLEMENTED("TEXRECTFLIP")

} /* static void rdp_texrectflip() */





static void rdp_loadsync()
{
        PRINT_RDP_INFO("LOADSYNC\n")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_loadsync() */





static void rdp_pipesync()
{
} /* static void rdp_pipesync() */





static void rdp_tilesync()
{
        PRINT_RDP_INFO("TILESYNC\n")
        PRINT_RDP_WARNING("ignored")

} /* static void rdp_tilesync() */





static void rdp_fullsync()
{
        PRINT_RDP_INFO("FULLSYNC\n")

        reg.do_or_check_sthg = RS4300I_DO_DP_INTERRUPT;
        //((uWORD *)mem.dpc_reg)[2] = 0;
} /* static void rdp_fullsync() */





static void rdp_setkeygb()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYGB")
} /* static void rdp_setkeygb() */





static void rdp_setkeyr()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETKEYR")
} /* static void rdp_setkeyr() */





static void rdp_setconvert()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETCONVERT")
} /* static void rdp_setconvert() */





static void rdp_setscissor()
{
        static char     *interlace_modes[] = { "no", "?", "even", "odd" };


/*
        PRINT_RDP_MNEMONIC("x=%lu; y=%lu; interlace_mode=%s; width=%lu; height=%lu\n",
               (cmd0 & 0x00fff000) >> 14,
               (cmd0 & 0x00000fff) >> 2,
               interlace_modes[(cmd1 & 0x03000000) >> 24],
               (cmd1 & 0x00fff000) >> 14,
               (cmd1 & 0x00000fff) >> 2);
        PRINT_RDP_WARNING("ignored")

/*
        SetVisualClipRect((cmd0 & 0x00fff000) >> 14, (cmd0 & 0x00000fff) >> 2,
                          (cmd1 & 0x00fff000) >> 14, (cmd1 & 0x00000fff) >> 2);
*/ /* read below! */

/* You are mostly correct here, but this comes back to my origional gripe..
   I followed this function for the X11 handler...  You use X's scissor,
   that isnt correct.  You have to do scissors on all RDP commands to the frame
   buffer in RDRAM.  Eventually the vertical retrace causes this to make it to 
   the screen. */
/* Conclusion, you affect X-Window, not RDRAM frame buffer..
   It is a problem. */

} /* static void rdp_setscissor() */





static void rdp_setprimdepth()
{
        PRINT_RDP_NOT_IMPLEMENTED("SETPRIMDEPTH")
} /* static void rdp_setprimdepth() */





static void rdp_setothermode()
{
        PRINT_RDP_INFO("SETOTHERMODE\n")
        PRINT_RDP_WARNING("this should clear/init SETOTHERMODE_[HL] - is not done")

} /* static void rdp_setothermode() */





static void rdp_loadtlut()
{
        if( (cmd1 & 0x00ffc000) >> 14 == 255)  /* if 8 bit LUT */
        {
            /* Should change TLUT_8 to TLUT_4[*] slower but more accurate 
               It would only be a problem if either:
               1) a program loaded 4 bit LUTS then drew as though an 8 bit LUT was loaded
               2) a program loaded 8 bit LUT then drew as though 4 bit LUTS were loaded */
                
                rdp_tlut_8_fmt   = rdp_textureimg_fmt;
                rdp_tlut_8_size  = rdp_textureimg_size;

                rdp_tlut_8_addr = rdp_textureimg_addr;
  
                PRINT_RDP_MNEMONIC("tile=%lu, count=255\n", (cmd1 & 0x07000000) >> 24);
        }
        else  /* count = 15 and tile->palette is meaningfull */
        {
            /* only do this operation 1 time */
            /* (fix me with a structure pointer)  
               to prevent repettitve [CURR_TILE]'s */
          
                int CURR_TILE = (cmd1 & 0x07000000) >> 24;

                rdp_tlut_4_fmt[rdp_td[CURR_TILE].palette]   = rdp_textureimg_fmt; 
                rdp_tlut_4_size[rdp_td[CURR_TILE].palette]  = rdp_textureimg_size;
                rdp_tlut_4_addr[rdp_td[CURR_TILE].palette] = rdp_textureimg_addr;

                PRINT_RDP_MNEMONIC("tile=%u, count=%ld\n", 
                       CURR_TILE,
                       (cmd1 & 0x00ffc000) >> 14);

        }


} /* static void rdp_loadtlut() */





static void rdp_settilesize()
{
        rdp_td[(cmd1 & 0x07000000) >> 24].uls = (cmd0 & 0x00fff000) >> 12;
        rdp_td[(cmd1 & 0x07000000) >> 24].ult = cmd0 & 0x00000fff;
        rdp_td[(cmd1 & 0x07000000) >> 24].lrs = (cmd1 & 0x00fff000) >> 12;
        rdp_td[(cmd1 & 0x07000000) >> 24].lrt = cmd1 & 0x00000fff;

} /* static void rdp_settilesize() */





static void rdp_loadblock()
{
        int     CURR_TILE;
        int     i;

		uHWORD uls, ult, lrs, dxt;


        CURR_TILE = (cmd1 & 0x07000000) >> 24;
        rdp_td[CURR_TILE].uls = (cmd0 & 0x00fff000) >> 12;
        rdp_td[CURR_TILE].ult = cmd0 & 0x00000fff;
        rdp_td[CURR_TILE].lrs = (cmd1 & 0x00fff000) >> 12;
        rdp_td[CURR_TILE].dxt = cmd1 & 0x00000fff;
        

        PRINT_RDP_INFO("LOADBLOCK ")
        PRINT_RDP_MNEMONIC("uls=%lu; ult=%lu; tile=%lu; lrs=%lu; dxt=%lu\n",
                rdp_td[CURR_TILE].uls,
                rdp_td[CURR_TILE].ult,
                CURR_TILE,
                rdp_td[CURR_TILE].lrs,
                rdp_td[CURR_TILE].dxt);

/* BUG: here must be sthg wrong in my opinion (ult? dxt?) */

/* we draw it directly to the rdram */

		uls = rdp_td[CURR_TILE].uls;
		lrs = rdp_td[CURR_TILE].lrs;
        for(i=uls; i<=lrs; i++)
        {
                tmem[i + rdp_td[CURR_TILE].tmem] = mem.rd_ram[i + rdp_textureimg_addr];
        }


} /* static void rdp_loadblock() */





static void rdp_loadtile()
{
        PRINT_RDP_INFO("LOADTILE ")
        PRINT_RDP_MNEMONIC("uls=%lu, ult=%lu, tile=%lu; lrs=%lu, lrt=%lu\n", 
               (cmd0 & 0x00fff000) >> 12, 
               (cmd0 & 0x00000fff),
               (cmd1 & 0x07000000) >> 24,
               (cmd1 & 0x00fff000) >> 12, 
               (cmd1 & 0x00000fff) );
        PRINT_RDP_WARNING("ignored")

        /* We are going to try and ignore this command. */
        /* This command moves memory from RDRAM/CART into TMEM. */
        /* We are going to try and aviod this because it is an extra memcpy */
        /* This will get us into trouble only when the uls,ult,lrs,lrt of
           the LOADTILE are different from the uls,ult,lrs,lrt of the SETTILESIZE
           that occurs just before TEXRECT */

} /* static void rdp_loadtile() */





static void rdp_settile()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };
        static char *cm[]       = { "NOMIRROR/WARP", "MIRROR", "CLAMP", "(MIRROR&CLAMP)?" };



        /* only do this operation 1 time */
        /* (fix me with a structure pointer)  
           to prevent repettitve [CURR_TILE]'s */

        int CURR_TILE = (cmd1 & 0x07000000) >> 24;
        
        rdp_td[CURR_TILE].format  = (cmd0 & 0x00e00000) >> 21;
        rdp_td[CURR_TILE].size    = (cmd0 & 0x00180000) >> 19;

        /* The following is << 2, because line represents 64 bit entries.     */
        /* 1 lines = 64 bits ... 64 bits / 16 bits per pixel = 4 pixels.      */
        /* It will be easier for us to use number of pixels rather than lines */

        /* ??? (niki) what happens if we are on a 32 bit display (niki) ??? */
        /* ??? (niki) maybe we should use lines either           (niki) ??? */

        rdp_td[CURR_TILE].line    = ((cmd0 & 0x0003fe00) >> 9) << 2;
        rdp_td[CURR_TILE].tmem    = (cmd0 & 0x000001ff) ;
        rdp_td[CURR_TILE].palette = (cmd1 & 0x00f00000) >> 20;
        rdp_td[CURR_TILE].clampt  = (cmd1 & 0x00080000) >> 19;
        rdp_td[CURR_TILE].mirrort = (cmd1 & 0x00040000) >> 18;
        rdp_td[CURR_TILE].maskt   = (cmd1 & 0x0003c000) >> 14;
        rdp_td[CURR_TILE].shiftt  = (cmd1 & 0x00003c00) >> 10;
        rdp_td[CURR_TILE].clamps  = (cmd1 & 0x00000200) >> 9;
        rdp_td[CURR_TILE].mirrors = (cmd1 & 0x00000100) >> 8;
        rdp_td[CURR_TILE].masks   = (cmd1 & 0x000000f0) >> 4;
        rdp_td[CURR_TILE].shifts  = (cmd1 & 0x0000000f) ;


        PRINT_RDP_MNEMONIC("format=%s; size=%s; line=%lu (16bit pixel); tmem=%lu; tile=%lu; palette=$%lx; "
               "cmt=%s; maskt=%lu; shiftt=%lu; cms=%s; masks=%lu; shifts=%lu\n",
               format[rdp_td[CURR_TILE].format],
               size[rdp_td[CURR_TILE].size],
               rdp_td[CURR_TILE].line,
               rdp_td[CURR_TILE].tmem,
               CURR_TILE,
               rdp_td[CURR_TILE].palette,
               cm[rdp_td[CURR_TILE].clampt + rdp_td[CURR_TILE].mirrort],
               rdp_td[CURR_TILE].maskt,
               rdp_td[CURR_TILE].shiftt,
               cm[rdp_td[CURR_TILE].clamps + rdp_td[CURR_TILE].mirrors],
               rdp_td[CURR_TILE].masks,
               rdp_td[CURR_TILE].shifts);

} /* static void rdp_settile() */





static void rdp_fillrect()
{
        uWORD    ulx, uly, lrx, lry;
		float    fulx, fuly, flrx, flry;
        uWORD    start, y, x;


        lrx = (cmd0 & 0x00ffc000) >> 14;
        lry = (cmd0 & 0x00000ffc) >> 2;
        ulx = (cmd1 & 0x00ffc000) >> 14;
        uly = (cmd1 & 0x00000ffc) >> 2;

		flrx = lrx+1000;
		flry = lry+1000;
		fulx = ulx;
		fuly = uly;

        start   = ulx + uly * rdp_colorimg_width + rdp_colorimg_addr;
        start &= 0x003fffff;

        /* draw the pixel */
//		glDisable(GL_TEXTURE);
//		GL_SetColor(rdp_fillcolor);

//		glRasterPos2d(0,0);
//		glRectf(fulx, fuly, flrx, flry);
//		glEnable(GL_TEXTURE);

} /* static void rdp_fillrect() */


static void rdp_setfillcolor()
{
        rdp_fillcolor = cmd1;

} /* static void rdp_setfillcolor() */





static void rdp_setfogcolor()
{
        PRINT_RDP_INFO("SETFOGCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", cmd1);

        rdp_fogcolor = cmd1;

} /* static void rdp_setfogcolor() */





static void rdp_setblendcolor()
{
        PRINT_RDP_INFO("SETBLENDCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", cmd1);

        rdp_blendcolor = cmd1;

} /* static void rdp_setblendcolor() */





static void rdp_setprimcolor()
{
        PRINT_RDP_INFO("SETPRIMCOLOR ")
        PRINT_RDP_MNEMONIC("color=$%08lx\n", cmd1);

        rdp_primcolor = cmd1;

} /* static void rdp_setprimcolor() */





static void rdp_setenvcolor()
{
//        PRINT_RDP_MNEMONIC("color=$%08lx\n", cmd1);

        rdp_envcolor = cmd1;

} /* static void rdp_setenvcolor() */





static void rdp_setcombine()
{
/*        PRINT_RDP_MNEMONIC("mode0=$%06lx; mode1=$%08lx\n",
               (cmd0 & 0x00ffffff),
               (cmd1 & 0xffffffff));
        PRINT_RDP_WARNING("ignored")
*/
} /* static void rdp_setcombine() */





static void rdp_settextureimage()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };


        rdp_textureimg_fmt   = (cmd0 & 0x00e00000) >> 21;
        rdp_textureimg_size  = (cmd0 & 0x00180000) >> 19;
        rdp_textureimg_width = 1 + (cmd0 & 0x00000fff);
        rdp_textureimg_addr  = segoffset2addr(cmd1);


/*        PRINT_RDP_INFO("SETTEXTUREIMAGE ")
        PRINT_RDP_MNEMONIC("format=%s; size=%s; width=%hu; addr=$%08lx\n",
               format[rdp_textureimg_fmt],
               size[rdp_textureimg_size],
               rdp_textureimg_width,
               rdp_textureimg_addr);
*/
} /* static void rdp_settextureimage() */





static void rdp_setdepthimage()
{
//        PRINT_RDP_MNEMONIC("addr=$%lx\n", cmd1);
} /* static void rdp_setdepthimage() */





static void rdp_setcolorimage()
{
        static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
        static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };



        rdp_colorimg_fmt   = (cmd0 & 0x00e00000) >> 21;
        rdp_colorimg_size  = (cmd0 & 0x00180000) >> 19;
        rdp_colorimg_width = 1 + (cmd0 & 0x00000fff);
        rdp_colorimg_addr = cmd1;

/*        PRINT_RDP_MNEMONIC("format=%s; size=%s; width=%hu; addr=$%lx\n",
                           format[rdp_colorimg_fmt],
                           size[rdp_colorimg_size],
                           rdp_colorimg_width,
                           rdp_colorimg_addr);
*/
} /* static void rdp_setcolorimage() */





static void rdp_trifill()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILL")
} /* static void rdp_trifill() */





static void rdp_trishade()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADE")
} /* static void rdp_trishade() */





static void rdp_tritxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTR")
} /* static void rdp_tritxtr() */





static void rdp_trishadetxtr()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTR")
} /* static void rdp_trishadetxtr() */





static void rdp_trifillz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRIFILLZ")
} /* static void rdp_trifillz() */





static void rdp_trishadez()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADEZ")
} /* static void rdp_trishadez() */





static void rdp_tritxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRITXTRZ")
} /* static void rdp_tritxtrz() */





static void rdp_trishadetxtrz()
{
        PRINT_RDP_NOT_IMPLEMENTED("TRISHADETXTRZ")
} /* static void rdp_trishadetxtrz() */










