/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   FiRES,  Germany
**/

#include <windows.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "main.h"
#include "wingl.h"

#include "debug.h"


// ** extern c-progs  <-mhh... perhaps it is better to use only ANSI-C
extern "C" int Init_Emulator(char *RomName); // Emu_Init
extern "C" int Shut_Emulator(void);			 // Emu_Init
extern "C" int cpu_step(void);		// CPU
extern "C" int RePaint;				// RDP  - if a frame ready it is 1
extern "C" void PausedKey(void);	// Pause ???

extern "C"
{
	int iRenderWindowX = 360;
	int iRenderWindowY = 300;

	char rom_realname[1024]; // Zhe RealName of the ROM
	int done = 0;		     // Done = 1 -> the program ends	
	int Paused = 0;			 // DInput Unacquire()  (Acquire() = 0)
	int DrawBackground = 1;  // Draw Background - important for 2d Demos (fire ...)
	int FrameCounter = 0;
}
extern "C"
{
  //#include "n64/debug_wrt.h"
}

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	DWORD err = 0;
	int i;
	static char szAppName[] = "Win-TrueReality (by FiRES)" ;
	MSG         msg ;
	WNDCLASSEX  wndclass ;

	wndclass.cbSize        = sizeof (wndclass) ;
	wndclass.style         = CS_HREDRAW | CS_VREDRAW ;
	wndclass.lpfnWndProc   = WndProc ;
	wndclass.cbClsExtra    = 0 ;
	wndclass.cbWndExtra    = 0 ;
	wndclass.hInstance	   = hInstanceGlobal = hInstance ;
	wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
	wndclass.lpszMenuName  = NULL ;
	wndclass.lpszClassName = szAppName ;
	wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;
	wndclass.hIcon	 = NULL;
	wndclass.hIconSm = NULL;
	wndclass.hCursor = NULL;

	RegisterClassEx (&wndclass) ;
	hGraphics = CreateWindowEx (WS_EX_OVERLAPPEDWINDOW, szAppName, NULL,
				WS_OVERLAPPEDWINDOW | WS_VISIBLE | WS_CAPTION,
				390, 30, iRenderWindowX, iRenderWindowY,
				NULL, NULL, hInstance, NULL) ;

	ShowWindow (hGraphics, iCmdShow) ;
	UpdateWindow (hGraphics);
	SendMessage(hGraphics, WM_SETTEXT, 0, (LPARAM)(LPCTSTR)"Win-TrueReality (by FiRES)");

// ** Init Opengl Window
	WGL_Init(hGraphics);
	WGL_SetSize(iRenderWindowX, iRenderWindowY);
	WGL_InitGl();


// ** Load 'tp-r.ini' InitFile
//	if (!LoadIniFile() != 0) exit(-1);
	if (!LoadDialog() != 0) exit(-1);

// ** Init DirectInput
	if (!DI_Init()) exit(-1);

// ** Init_Emulator
	if (!Init_Emulator(rom_realname)) exit(-1);
	if (DebuggerStatus == Debugger_ON) StartDebugger();
// ** Mainloop 
	while(done == 0)
	{
	// ** Handle windows messages;
		while (PeekMessage (&msg, NULL, 0, 0,PM_REMOVE))
		{
			TranslateMessage (&msg);
			DispatchMessage (&msg);	
		}
    // ** Handle emulator 
		//** If Debugger = ON all cpu-handling is made by the Debugger
		if (DebuggerStatus == Debugger_ON) 
		{
			DebuggerAction(); 
			if (RePaint == 1) redraw();
		}
		//** every 10.000 CPU Steps we handle the WindowsMessages
		else 
		{		
			for (i=0; i < 100000; i++)
			{
				if (cpu_step() != 0 ) done = 1;
				if (RePaint == 1) redraw();				
			}						
		}
	//** Handle DirectInput
		DI_SyncAcquire();   
		PausedKey();

	}

// ** End of Emulation
	Shut_Emulator();

// ** Shutdown DirectInput
	DI_End();

// ** Shutdown OpenGL Window
	WGL_End(hGraphics);
	return msg.wParam ;
	
}

LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	int x,y;
	switch (iMsg)
    {
		case WM_QUERYDRAGICON:
			return (LRESULT) hIconGlobal;

		case WM_CREATE:									
			break;

		case WM_DESTROY:
			PostQuitMessage (0) ;
			done = 1;
			break;

		case WM_SIZE:
			x = LOWORD (lParam);
			y = HIWORD (lParam);

			if (WGL_GLRCValid()) {
				WGL_SetSize (x, y);
				WGL_SetViewport(x, y);

				iRenderWindowX = x;
				iRenderWindowY = y;
				return 0;
			}

			break;

		case WM_PALETTECHANGED:
			if (WGL_GLRCValid() && WGL_PaletteValid() && (HWND) wParam != hwnd) {

				WGL_RedoPalette ();
				break;
			}
			break;

		case WM_QUERYNEWPALETTE:

			if (WGL_GLRCValid() && WGL_PaletteValid()) {

				WGL_RedoPalette ();
				return TRUE;
			}
			break;

		case WM_PAINT:
        {
            PAINTSTRUCT ps;
            BeginPaint(hwnd, &ps);

            if (WGL_GLRCValid()) {
                redraw();
            }

            EndPaint(hwnd, &ps);
            return 0;
        }
        break;

		case WM_KEYDOWN :

			switch (wParam)
			{
				case VK_HOME :
					break;

				case VK_RIGHT:
					break;

				case VK_LEFT:
					break;

				case VK_UP:
					break;

				case VK_DOWN:
					break;

				case VK_ADD:
					break;

				case VK_SUBTRACT:
					DebuggerStatus = Debugger_OFF;
					Paused = 0;
					break;
			}
			return 0;

		default:
			break;
	}
	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}

void redraw ()
{	
	char FrameText[1024];
//**  Draw FrameNum  
	if (RePaint == 1) // RePaint only Set to 1 from RDP / Display.c
	{
		FrameCounter++;
		sprintf(FrameText, "Win-TrueReality (by FiRES) - Frames: %i", FrameCounter);
		SendMessage(hGraphics, WM_SETTEXT, 0, (LPARAM)(LPCTSTR)FrameText);
	}

//** Flush ScreenBuffer 
	WGL_FlushScreen();
	RePaint = 0;

}


int LoadIniFile()
{
	FILE *stream = NULL;
//** Load the Ini-File
	stream = fopen("tr-p.ini", "rt");
	if (stream == NULL) 
	{
		MessageBox(hGraphics, "Can't open Ini-FILE", "Error:", MB_OK);
		return 0;
	}

	fscanf(stream, "%s %i %i %i\n", rom_realname, &DrawBackground, 
									&Paused, &DebuggerStatus);
	fclose(stream);
//** Test if the Rom is present
	stream = fopen(rom_realname, "rt");
	if (stream == NULL)
	{
		MessageBox(hGraphics, "Can't open ROM", "Error:", MB_OK);
		return 0;
	}
	fclose(stream);
	return 1;
}

int LoadDialog()
{
	OPENFILENAME ofn;
	char szFileName[1024];
	char szFileTitle[1024];

	memset(&szFileName, 0,sizeof(szFileName));
	memset(&szFileTitle, 0, sizeof(szFileTitle));

    ofn.lStructSize			= sizeof(OPENFILENAME);
    ofn.hwndOwner			= hGraphics;
    ofn.lpstrFilter			= "N64 ROMs (*.V64, *.BIN)\0*.V64;*.BIN\0All Files (*.*)\0*.*\0";
	ofn.lpstrCustomFilter	= NULL;
    ofn.nMaxCustFilter		= 0;
    ofn.nFilterIndex		= 1;
    ofn.lpstrFile			= szFileName;
    ofn.nMaxFile			= 1024;
    ofn.lpstrInitialDir		= NULL;
    ofn.lpstrFileTitle		= szFileTitle;
    ofn.nMaxFileTitle		= 1024;
    ofn.lpstrTitle			= NULL;
    ofn.lpstrDefExt			= "TXT";
    ofn.Flags				= 0;

	if (!GetOpenFileName ((LPOPENFILENAME)&ofn))
		return 0;

	strcpy(rom_realname, szFileName);
	return 1;
}