/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   FiRES,  Germany
**/

#include <windows.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "wingl.h"

HDC WGL_hdc;
HGLRC WGL_hGLRC;
HPALETTE WGL_hPalette;

int WGL_sizeX;
int WGL_sizeY;


// ** GL WindowHandling Functions
void WGL_Init (HWND hWnd)
{
	WGL_hdc = GetDC(hWnd);
	WGL_SetupPixelFormat ();
	WGL_SetupPalette ();
	WGL_hGLRC = wglCreateContext (WGL_hdc);
	wglMakeCurrent(WGL_hdc, WGL_hGLRC);
}

void WGL_SetSize (int iX, int iY)
{
	WGL_sizeX = iX;
	WGL_sizeY = iY;
}

void WGL_GetSize (int *iX, int *iY)
{
	if (iX != NULL && iY != NULL)
	{
		WGL_sizeX = *iX;
		WGL_sizeY = *iY;
	}
}

void WGL_End (HWND hWnd)
{
	if (WGL_hGLRC) {
		wglMakeCurrent (NULL, NULL);
		wglDeleteContext(WGL_hGLRC);
	}

	if (WGL_hPalette) {
		DeleteObject (WGL_hPalette);
	}

	ReleaseDC(hWnd, WGL_hdc);
}

BOOL WGL_GLRCValid ()
{
	if (WGL_hGLRC)
		return TRUE;
	else
		return FALSE;
}

BOOL WGL_PaletteValid ()
{
	if (WGL_hPalette)
		return TRUE;
	else
		return FALSE;
}


void WGL_RedoPalette ()
{
	UnrealizeObject (WGL_hPalette);
	SelectPalette (WGL_hdc, WGL_hPalette, FALSE);
	RealizePalette (WGL_hdc);
}

void WGL_SetupPixelFormat()
{
	PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),  /* size */
		1,                              /* version */
		PFD_SUPPORT_OPENGL |
		PFD_DRAW_TO_WINDOW |
		PFD_DOUBLEBUFFER,               /* support double-buffering */
		PFD_TYPE_RGBA,                  /* color type */
		0,                              /* prefered color depth */
		0, 0, 0, 0, 0, 0,               /* color bits (ignored) */
		0,                              /* no alpha buffer */
		0,                              /* alpha bits (ignored) */
		0,                              /* no accumulation buffer */
		0, 0, 0, 0,                     /* accum bits (ignored) */
		16,                             /* depth buffer */
		0,                              /* no stencil buffer */
		0,                              /* no auxiliary buffers */
		PFD_MAIN_PLANE,                 /* main layer */
		0,                              /* reserved */
		0, 0, 0                         /* no layer, visible, damage masks */
	};
	int pixelFormat;

	pixelFormat = ChoosePixelFormat(WGL_hdc, &pfd);
	if (pixelFormat == 0) {
		MessageBox(WindowFromDC(WGL_hdc), "ChoosePixelFormat failed.", "ErrorWinGL",
			MB_ICONERROR | MB_OK);
		exit(-1);
	}

	if (SetPixelFormat(WGL_hdc, pixelFormat, &pfd) != TRUE) {
		MessageBox(WindowFromDC(WGL_hdc), "SetPixelFormat failed.", "Error WinGL",
			MB_ICONERROR | MB_OK);
		exit(-1);
	}
}

void WGL_SetupPalette ()
{
	int pixelFormat = GetPixelFormat(WGL_hdc);
	PIXELFORMATDESCRIPTOR pfd;
	LOGPALETTE* pPal;
	int paletteSize;

	DescribePixelFormat(WGL_hdc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	if (pfd.dwFlags & PFD_NEED_PALETTE) {
		paletteSize = 1 << pfd.cColorBits;
	} else {
		return;
	}

	pPal = (LOGPALETTE*)
		malloc(sizeof(LOGPALETTE) + paletteSize * sizeof(PALETTEENTRY));
	pPal->palVersion = 0x300;
	pPal->palNumEntries = paletteSize;

//** Create Palette RGB
	{
		int redMask = (1 << pfd.cRedBits) - 1;
		int greenMask = (1 << pfd.cGreenBits) - 1;
		int blueMask = (1 << pfd.cBlueBits) - 1;
		int i;

		for (i=0; i<paletteSize; ++i)
		{
			pPal->palPalEntry[i].peRed =
				(((i >> pfd.cRedShift) & redMask) * 255) / redMask;
			pPal->palPalEntry[i].peGreen =
				(((i >> pfd.cGreenShift) & greenMask) * 255) / greenMask;
			pPal->palPalEntry[i].peBlue =
				(((i >> pfd.cBlueShift) & blueMask) * 255) / blueMask;
			pPal->palPalEntry[i].peFlags = 0;
		}
	}

	WGL_hPalette = (HPALETTE)CreatePalette(pPal);
	free(pPal);

	if (WGL_hPalette) {
		SelectPalette(WGL_hdc, WGL_hPalette, FALSE);
		RealizePalette(WGL_hdc);
	}
}

//** GL Functions

//** Sets a new Viewport
void WGL_SetViewport(int x, int y)
{
	glViewport(0, 0, x, y);
}

//** Render the Scene and Blit it to Screen
void WGL_FlushScreen(void)
{
	glFlush();
    SwapBuffers(WGL_hdc);
}

//** Init OpenGL (all standard functions ...)
void WGL_InitGl(void)
{
	glViewport(0, 0, WGL_sizeX, WGL_sizeY);

    /* set viewing projection */
    glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-4.0f, 4.0f, -2.704225f, 2.704225f, 4.0f, 4096.0f);

    /* position viewer */
	glMatrixMode(GL_MODELVIEW);

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);
	glDisable(GL_CULL_FACE);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glClear(GL_DEPTH_BUFFER_BIT);	

	glShadeModel(GL_SMOOTH);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


	glEnable(GL_TEXTURE_2D);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

//	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
/*	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);

    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
    glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 0);*/

}
